//
//  EggCore.cpp
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#include "EggCore.h"

EggCore::EggCore()
{
    ReadGrammar rg;
    GraphGrammar gg;
    //cout<<system ("pwd")<<endl;
    string sFilePath = "tests/test1/sample_rule_1.txt";
    cout<<sFilePath<<endl;
    rg.ParseGrammarFile(sFilePath, gg);
    cout<<"EggCore Done!"<<endl;
}

EggCore::EggCore(map<string, string> mArguments)
//mArguments: input arguments
{
    string sEmptyStr = "";
    this->sWorkingDirectory = "." + sEmptyStr + this->cPathSeparator;
    this->sGrammarFilePath = "";
    this->sInputGraphFilePath = "";
    this->sInputGraphFolderPath = "";
    this->sOutputDirectory = "";
    this->bPartialParse = false;
    InitializeArguments(mArguments);
}

void EggCore::InitializeArguments(const map<string, string> & mArguments)
//mArguments: input arguments
{
    map<string, string >::const_iterator mssIterator;
    string sPartialParse;
    
    mssIterator = mArguments.find("working directory");
    if (mssIterator != mArguments.end())
        this->sWorkingDirectory = AddPathSeparator(mssIterator->second);
    mssIterator = mArguments.find("grammar file path");
    if (mssIterator != mArguments.end())
        this->sGrammarFilePath = mssIterator->second;
    mssIterator = mArguments.find("input graph");
    if (mssIterator != mArguments.end())
        this->sInputGraphFilePath = mssIterator->second;
    mssIterator = mArguments.find("input graph folder");
    if (mssIterator != mArguments.end())
        this->sInputGraphFolderPath = AddPathSeparator(mssIterator->second);
    mssIterator = mArguments.find("output directory");
    if (mssIterator != mArguments.end())
        this->sOutputDirectory = AddPathSeparator(mssIterator->second);
    if (this->sOutputDirectory == "")
        this->sOutputDirectory = this->sWorkingDirectory;
    if ((this->sInputGraphFolderPath == "") && (this->sInputGraphFilePath == ""))
        this->sInputGraphFolderPath = this->sWorkingDirectory;
    if (this->sGrammarFilePath == "")
    {
        cerr<<"grammar file path must be specified"<<endl;
        exit(1);
    }
    mssIterator = mArguments.find("partial parse");
    if (mssIterator != mArguments.end())
    {
        sPartialParse = mssIterator->second;
        if ((sPartialParse == "True") || (sPartialParse == "true") || (sPartialParse == "TRUE")
            || (sPartialParse == "T") || (sPartialParse == "t"))
            this->bPartialParse = true;
    }
    vsInputGraphFileNamesList = LocateAllInputGraphFiles();
}

string EggCore::AddPathSeparator(const string & sPath)
//sPath: input path
{
    string sProcessedPath = sPath;
    
    if (sProcessedPath == "")
        cerr<<"Warning: receive empty file path"<<endl;
    if (sProcessedPath.at(sProcessedPath.length()-1) != this->cPathSeparator)
        sProcessedPath += cPathSeparator;
    
    return sProcessedPath;
}

vector<string> EggCore::LocateAllInputGraphFiles()
//return the list of all input graph file names
{
    const string sGraphExt[] = {".gml", ".GML", ".Gml"};
    int i;
    vector<string> vsFileNamesFromCurSearch;
    vector<string> vsAllInputGraphFiles;
    
    vsAllInputGraphFiles.clear();
    //if sInputGraphFolderPath is specified, sInputGraphFilePath will be ignored.
    if (sInputGraphFolderPath != "")
        sInputGraphFilePath = "";
    if (sInputGraphFilePath != "")
        vsAllInputGraphFiles.push_back(sInputGraphFilePath);
    else if (sInputGraphFolderPath != "")
        for (i=0; i<3; i++)
        {
            vsFileNamesFromCurSearch.clear();
            vsFileNamesFromCurSearch = FilesOperation::LocateFilesWithPattern(sInputGraphFolderPath, sGraphExt[i]);
            vsAllInputGraphFiles.insert(vsAllInputGraphFiles.end(), vsFileNamesFromCurSearch.begin(),
                                        vsFileNamesFromCurSearch.end());
        }
    
    return vsAllInputGraphFiles;
}

void EggCore::LoadInputGraphFiles(const vector<string> & vsInputGraphPaths)
//vsInputGraphPaths: vector of input file paths
{
    vector< vector< RegularGraph *> >  vvpInputRegGraphs;
    ReadInputGraphs readInGraph;
    readInGraph.ParseAllInputGraphFiles(vsInputGraphPaths, vvpInputRegGraphs);
}

void EggCore::LoadInputGraphFiles(const vector<string> & vsInputGraphPaths,
                                  vector< vector< RegularGraph *> >  & vvpInputRegGraphs)
//vsInputGraphPaths: vector of input file paths
//vvpInputRegGraphs: vector of pointers of input regular graphs
{
    ReadInputGraphs readInGraph;
    readInGraph.ParseAllInputGraphFiles(vsInputGraphPaths, vvpInputRegGraphs);
}


void EggCore::ReadGraphGrammar()
{
    ReadGrammar readG;
    
    readG.ParseGrammarFile(this->sGrammarFilePath, gGrammar);
    cout<<"The graph grammar has been read!"<<endl;
}


bool EggCore::ParseRegGraph(RegularGraph *pInputRegGraph)
//pInputRegGraph: pointer of input regular graph
//return ture if the input regular graph can be parsed by the grammar; otherwise, return false.
{
    bool bParse = false;
    ParseRegularGraph ParseG(bPartialParse);
    
    bParse = ParseG.parseGraph(*pInputRegGraph, gGrammar);
    return bParse;
}