//
//  EggCore.h
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__EggCore__
#define __egg__EggCore__

#include <stdio.h>

#include "ReadGrammar.h"
#include "FilesOperation.h"
#include "ReadInputGraphs.h"
#include "ParseRegularGraph.h"

class EggCore
{
    string sWorkingDirectory; // path of the working directory
    string sGrammarFilePath;  // path of the file containing grammar
    string sInputGraphFilePath; // path of the input graph file
    string sInputGraphFolderPath; // path of the input graph folder
    string sOutputDirectory; // path of the output directory
    bool bPartialParse; //true: conduct partial parse (for searching specific subgraph), by default, it is false
    GraphGrammar gGrammar;
    
    const char cPathSeparator =       // path separator
                                #ifdef _WIN32
                                '\\';
                                #else
                                '/';
                                #endif

    //intialize member variables according to input arguments
    void InitializeArguments(const map<string, string> & mArguments);
    vector<string> LocateAllInputGraphFiles(); // locate all input graph files
    //if the path of folder doesn't end with a path separator, add the separator to the end
    string AddPathSeparator(const string & sPath);
    
public:
    vector<string> vsInputGraphFileNamesList; // list of input graph file names
    
    
    EggCore();
    EggCore(map<string, string> mArguments);
    //Load content of input graph files
    void LoadInputGraphFiles(const vector<string> & vsInputGraphPaths);
    //Load content of input graph files
    void LoadInputGraphFiles(const vector<string> & vsInputGraphPaths,
                             vector< vector< RegularGraph *> >  & vvpInputRegGraphs);
    //read grammar
    void ReadGraphGrammar();
    
    //parse regular graph
    bool ParseRegGraph(RegularGraph * pInputRegGraph);
};

#endif /* defined(__egg__EggCore__) */
