//
//  GML.cpp
//  egg
//
//  Created by Yingfeng Wang on 1/24/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#include "GML.h"

GML::GML(const GML &oriGML)
//oriGML: original GML treated as source
{
    Edge *pEdge;
    Node *pNode;
    size_t i;
    map<int, Edge *> mEdgeIdPointer;
    map<int, Node *> mNodeIdPointer;
    map<int, Node *>::const_iterator itCurNodeMap;
    
    this->eDirected = oriGML.eDirected;
    this->mNodeIdPos = oriGML.mNodeIdPos;
    this->mGMLInfo = oriGML.mGMLInfo;
    this->vpEdges.clear();
    for (i = 0; i < oriGML.vpEdges.size(); i++)
    {
        pEdge = new Edge();
        *pEdge = *(oriGML.vpEdges.at(i));
        pEdge->pSourceNode = nullptr;
        pEdge->pTargetNode = nullptr;
        mEdgeIdPointer.insert(pair<int, Edge *> (pEdge->iEdgeId, pEdge));
        this->vpEdges.push_back(pEdge);
    }
    for (i = 0; i < oriGML.vpNodes.size(); i++)
    {
        pNode = new Node();
        *pNode = *(oriGML.vpNodes.at(i));
        copyEdgeMap(oriGML.vpNodes.at(i)->mpSourceEdges, pNode->mpSourceEdges, mEdgeIdPointer);
        copyEdgeMap(oriGML.vpNodes.at(i)->mpTargetEdges, pNode->mpTargetEdges, mEdgeIdPointer);
        copyEdgeMap(oriGML.vpNodes.at(i)->mpUndirectedEdges, pNode->mpUndirectedEdges, mEdgeIdPointer);
        mNodeIdPointer.insert(pair<int, Node *> (pNode->iNodeId, pNode));
        this->vpNodes.push_back(pNode);
    }
    for (i = 0; i < oriGML.vpEdges.size(); i++)
    {
        if (oriGML.vpEdges.at(i)->pSourceNode != nullptr)
        {
            itCurNodeMap = mNodeIdPointer.find(oriGML.vpEdges.at(i)->pSourceNode->iNodeId);
            if (itCurNodeMap == mNodeIdPointer.end())
            {
                cerr<<"fail to find node with node id "<<oriGML.vpEdges.at(i)->pSourceNode->iNodeId<<endl;
                exit(1);
            }else
                this->vpEdges.at(i)->pSourceNode = itCurNodeMap->second;
        }
        if (oriGML.vpEdges.at(i)->pTargetNode != nullptr)
        {
            itCurNodeMap = mNodeIdPointer.find(oriGML.vpEdges.at(i)->pTargetNode->iNodeId);
            if (itCurNodeMap == mNodeIdPointer.end())
            {
                cerr<<"fail to find node with node id "<<oriGML.vpEdges.at(i)->pTargetNode->iNodeId<<endl;
                exit(1);
            }else
                this->vpEdges.at(i)->pTargetNode = itCurNodeMap->second;
        }
    }
    copyEdgeVector(oriGML.vpStartDangling, this->vpStartDangling, mEdgeIdPointer);
    copyEdgeVector(oriGML.vpEndDangling, this->vpEndDangling, mEdgeIdPointer);
    copyEdgeVector(oriGML.vpUndirectedStartDangling, this->vpUndirectedStartDangling, mEdgeIdPointer);
    copyEdgeVector(oriGML.vpUndirectedEndDangling, this->vpUndirectedEndDangling, mEdgeIdPointer);
}


GML::GML()
{
    eDirected = UNDIRECTED;
}

GML::~GML()
{
    size_t i;
    Node * pCurNode;
    Edge * pCurEdge;
    
    for (i = 0; i < vpNodes.size(); i++)
    {
        pCurNode = vpNodes.at(i);
        delete pCurNode;
        vpNodes.at(i) = nullptr;
    }
    for (i = 0; i < vpEdges.size(); i++)
    {
        pCurEdge = vpEdges.at(i);
        delete pCurEdge;
        vpEdges.at(i) = nullptr;
    }
}

void GML::copyEdgeMap(const map<int, Edge *> &mSource, map<int, Edge *> &mDestination,
                      const map<int, Edge *> &mEdgeIdPointer)
//mSource: source map; mDestination: destination map; mEdgeIdPointer: map between edge id and edge pointer
{
    map<int, Edge *>::const_iterator itEdgeMap;
    map<int, Edge *>::const_iterator itCurEdgeMap;
    
    mDestination.clear();
    
    for (itEdgeMap = mSource.begin(); itEdgeMap != mSource.end(); ++itEdgeMap)
    {
        itCurEdgeMap = mEdgeIdPointer.find(itEdgeMap->second->iEdgeId);
        if (itCurEdgeMap == mEdgeIdPointer.end())
        {
            cerr<<"fail to find edge with edge id "<<itEdgeMap->second->iEdgeId<<endl;
            exit(1);
        }else
        {
            mDestination.insert(make_pair(itEdgeMap->first, itCurEdgeMap->second));
        }
    }
}

void GML::copyEdgeVector(const vector<Edge *> &vSource, vector<Edge *> &vDestination,
                         const map<int, Edge *> &mEdgeIdPointer)
//vSource: source edge vector; vDestination: destination edge vector; mEdgeIdPointer: map between edge id and edge pointer
{
    map<int, Edge *>::const_iterator itCurEdgeMap;
    vector<Edge *>::const_iterator itpEdge;
    
    vDestination.clear();
    for (itpEdge = vSource.begin(); itpEdge != vSource.end(); ++itpEdge)
    {
        itCurEdgeMap = mEdgeIdPointer.find((*itpEdge)->iEdgeId);
        if (itCurEdgeMap == mEdgeIdPointer.end())
        {
            cerr<<"fail to find edge with edge id "<<(*itpEdge)->iEdgeId<<endl;
            exit(1);
        }else
        {
            vDestination.push_back(itCurEdgeMap->second);
        }
    }
}

void GML::setDirected(DirectionStatus eDirected)
//eDirected: directed status
{
    this->eDirected = eDirected;
}

DirectionStatus GML::getDirected()
//returen eDirected
{
    return eDirected;
}

void GML::dataOptimization()
{
    sortNodes();
}

void GML::sortNodes()
{
    sort(vpNodes.begin(), vpNodes.end(), Node::LessLabel);
}

bool GML::graphMatch(const vector<GML *> &vpHostContextGraph, const vector<GML *> &vpHostGraph,
                     const vector<GML *> &vpGuestContextGraph, const vector<GML *> &vpGuestGraph,
                     stack<map<Node *, Node *> > &smMatchGraph, bool bSimpleMode /* = false */)
//vpHostContextGraph: context part of the host graph; vpHostGraph: host graph without context
//vpGuestContextGraph: context part of the guest graph; vpGuestGraph: guest graph without context
//smMatchGraph stack of details of all matches
//bSimpleMode: if true, will switch to simple mode, in which, all edges are undirected, and there is no dangling edges.
//return true if the guest graph is a subgraph of the host graph; otherwise return false
{
    bool bMatch;
    vector <GML *> vpMergedHostGraph, vpMergedGuestGraph;
    
    while (!smMatchGraph.empty())
        smMatchGraph.pop();
    if (bSimpleMode)
        bMatch = graphMatchSimple(vpHostContextGraph, vpHostGraph, vpGuestContextGraph, vpGuestGraph, smMatchGraph);
    else
    {
        vpMergedHostGraph.reserve(vpHostContextGraph.size() + vpHostGraph.size());
        vpMergedHostGraph.insert(vpMergedHostGraph.end(), vpHostContextGraph.begin(), vpHostContextGraph.end());
        vpMergedHostGraph.insert(vpMergedHostGraph.end(), vpHostGraph.begin(), vpHostGraph.end());
        
        vpMergedGuestGraph.reserve(vpGuestContextGraph.size() + vpGuestGraph.size());
        vpMergedGuestGraph.insert(vpMergedGuestGraph.end(), vpGuestContextGraph.begin(), vpGuestContextGraph.end());
        vpMergedGuestGraph.insert(vpMergedGuestGraph.end(), vpGuestGraph.begin(), vpGuestGraph.end());
        
        bMatch = graphMatch(vpMergedHostGraph, vpMergedGuestGraph, smMatchGraph);
    }
    
    return bMatch;
}

bool GML::graphMatch(const vector<GML *> &vpHostGraph, const vector<GML *> &vpGuestGraph,
                     stack<map<Node *, Node *> > &smMatchGraph, bool bSimpleMode /* = false */)
//vpHostGraph: host graph; vpGuestGraph: guest graph; smMatchGraph: stack of details of all matches
//bSimpleMode: if true, will switch to simple mode, in which, all edges are undirected, and there is no dangling edges.
//return true if the guest graph is a subgraph of the host graph; otherwise return false
{
    bool bMatch;
    BigGML MergedHostBigGML, MergedGuestBigGML;
    
    while (!smMatchGraph.empty())
        smMatchGraph.pop();
    if (bSimpleMode)
        bMatch = graphMatchSimple(vpHostGraph, vpGuestGraph, smMatchGraph);
    else
    {
        mergeGMLs(vpHostGraph, MergedHostBigGML);
        mergeGMLs(vpGuestGraph, MergedGuestBigGML);
        bMatch = BigGMLMatch(MergedHostBigGML, MergedGuestBigGML, smMatchGraph);
    }

    return bMatch;
}

bool GML::graphMatchSimple(const vector <GML *> & vpHostContextGraph, const vector <GML *> & vpHostGraph,
                           const vector <GML *> & vpGuestContextGraph, const vector <GML *> & vpGuestGraph,
                           stack < map<Node *, Node *> > & smMatchGraph)
//vpHostContextGraph: context part of the host graph; vpHostGraph: host graph without context
//vpvpGuestContextGraph: context part of the guest graph; vpGuestGraph: guest graph without context
//smMatchGraph: stack of details of all matche
//return true if the guest graph is a subgraph of the host graph; otherwise return false
{
    bool bMatch = false;
    //need more codes ...
    return bMatch;
}

bool GML::graphMatchSimple(const vector <GML *> & vpHostGraph, const vector <GML *> & vpGuestGraph,
                           stack < map<Node *, Node *> > & smMatchGraph)
//vpHostGraph: host graph; vpGuestGraph: guest graph; smMatchGraph: stack of details of all matches
//return true if the guest graph is a subgraph of the host graph; otherwise return false
{
    bool bMatch;
    vector <GML *> vpEmptyContextGraph;
    
    vpEmptyContextGraph.clear();
    bMatch = graphMatchSimple(vpEmptyContextGraph, vpHostGraph, vpEmptyContextGraph, vpGuestGraph, smMatchGraph);
    
    return bMatch;
}

void GML::mergeGMLs(const vector<GML *> &vpGMLs, BigGML & mergedBigGML)
//vpGMLs: vector of pointers of GMLs
//mergedBigGML: a BigGML containing all nodes and edges of input GMLs. The current version doesn't merge mGMLInfo
{
    size_t i, j;
    for (i=0; i<vpGMLs.size(); i++)
    {
        for (j=0; j<vpGMLs.at(i)->vpNodes.size(); j++)
            mergedBigGML.vpNodes.push_back(vpGMLs.at(i)->vpNodes.at(j));
        for (j=0; j<vpGMLs.at(i)->vpEdges.size(); j++)
            mergedBigGML.vpEdges.push_back(vpGMLs.at(i)->vpEdges.at(j));
        for (j=0; j<vpGMLs.at(i)->vpStartDangling.size(); j++)
            mergedBigGML.vpStartDangling.push_back(vpGMLs.at(i)->vpStartDangling.at(j));
        for (j=0; j<vpGMLs.at(i)->vpEndDangling.size(); j++)
            mergedBigGML.vpEndDangling.push_back(vpGMLs.at(i)->vpEndDangling.at(j));
        for (j=0; j<vpGMLs.at(i)->vpUndirectedStartDangling.size(); j++)
            mergedBigGML.vpUndirectedStartDangling.push_back(vpGMLs.at(i)->vpUndirectedStartDangling.at(j));
        for (j=0; j<vpGMLs.at(i)->vpUndirectedEndDangling.size(); j++)
            mergedBigGML.vpUndirectedEndDangling.push_back(vpGMLs.at(i)->vpUndirectedEndDangling.at(j));
    }
    sort(mergedBigGML.vpNodes.begin(), mergedBigGML.vpNodes.end(), Node::LessLabel);
    for (i=0; i<mergedBigGML.vpNodes.size(); i++)
        mergedBigGML.mNodeIdPos.insert(pair<size_t,size_t>(mergedBigGML.vpNodes.at(i)->iNodeId, i));
}

bool GML::BigGMLMatch(const BigGML &HostBigGML, const BigGML &GuestBigGML, stack<map<Node *, Node *> > &smMatchGraph)
//HostBigGML: host BigGML; GuestBigGML: guest BigGML; smMatchGraph: stack of details of all matches
//return true if the guest BigGML is a subgraph of the host BigGML; otherwise return false
{
    bool bMatch = false;
    vector<vector<Node *> > vvpNodeCandidates;
    
    while (!smMatchGraph.empty())
        smMatchGraph.pop();
    //buildNodesMap(GuestBigGML.vpNodes, mpGuestNodes);
    vvpNodeCandidates = findAllMatchedNodesPairs(HostBigGML, GuestBigGML);
    bMatch = generateMatchedSubgraphs(vvpNodeCandidates, HostBigGML, GuestBigGML, smMatchGraph);
    
    return bMatch;
}

bool GML::GMLMatch(const GML & HostGML, const GML & GuestGML, stack < map<Node *, Node *> > & smMatchGraph)
//HostGML: host BigGML; GuestGML: guest GML; smMatchGraph: stack of details of all matches
//return true if the guest GML is a subgraph of the host GML; otherwise return false

{
    bool bMatch;
    BigGML HostBigGML, GuestBigGML;
    
    HostBigGML  = GML2BigGML(HostGML);
    GuestBigGML = GML2BigGML(GuestGML);
    bMatch = BigGMLMatch(HostBigGML, GuestBigGML, smMatchGraph);
    
    return bMatch;
}

BigGML GML::GML2BigGML(const GML &inputGML)
//inputGML: input GML
//return a transferred BigGML
{
    BigGML outputBigGML;
    
    outputBigGML.vpNodes = inputGML.vpNodes;
    outputBigGML.mNodeIdPos = inputGML.mNodeIdPos;
    outputBigGML.vpEdges = inputGML.vpEdges;
    outputBigGML.vpStartDangling = inputGML.vpStartDangling;
    outputBigGML.vpEndDangling = inputGML.vpEndDangling;
    outputBigGML.vpUndirectedStartDangling = inputGML.vpUndirectedStartDangling;
    outputBigGML.vpUndirectedEndDangling = inputGML.vpUndirectedEndDangling;
    
    return outputBigGML;
}

bool GML::isFlexible(Node *pCurrentNode)
//pCurrentNode: pointer to the current node; return true if the node referred by pCurrentNode is flexible
{
    bool bFlexible = false;
    map<string, string>::const_iterator mssIt;
    
    mssIt = pCurrentNode->mNodeInfo.find("flexible");
    if ((mssIt != pCurrentNode->mNodeInfo.end()) && (mssIt->second == "true"))
        bFlexible = true;
    
    return bFlexible;
}

vector<vector<Node *> > GML::findAllMatchedNodesPairs(const BigGML &HostBigGML, const BigGML &GuestBigGML)
//HostBigGML: host BigGML; GuestBigGML: guest BigGML;
//return all candidates of nodes GuestBigGML in HostBigGML
{
    size_t i, j;
    Node *pCurrentGuestNode, *pCurrentHostNode;
    vector<Node *> vpCurrentCandidates;
    vector<vector<Node *> > vvpAllCandidates;
    bool bFlexible;
    
    vvpAllCandidates.clear();
    for (i=0; i< GuestBigGML.vpNodes.size(); i++)
    {
        pCurrentGuestNode = GuestBigGML.vpNodes.at(i);
        vpCurrentCandidates.clear();
        bFlexible = isFlexible(pCurrentGuestNode);
        for (j=0; j< HostBigGML.vpNodes.size(); j++)
        {
            pCurrentHostNode = HostBigGML.vpNodes.at(j);
            if ((pCurrentGuestNode == nullptr)||(pCurrentHostNode == nullptr))
            {
                cerr<<"empty node pointer"<<endl;
                exit(1);
            }
            if ((pCurrentGuestNode->sLabel < pCurrentHostNode->sLabel)&&
                (!(compareAnnotationNode( pCurrentHostNode->sLabel, pCurrentGuestNode->sLabel))))
                break; // all nodes are sorted in ascending order of labels
            else if (compareAnnotationNode(pCurrentHostNode->sLabel, pCurrentGuestNode->sLabel))
            {
                if (bFlexible)
                {
                    if ((pCurrentGuestNode->mpSourceEdges.size() <= pCurrentHostNode->mpSourceEdges.size())
                        && (pCurrentGuestNode->mpTargetEdges.size() <= pCurrentHostNode->mpTargetEdges.size())
                        && (pCurrentGuestNode->mpUndirectedEdges.size() <= pCurrentHostNode->mpUndirectedEdges.size()))
                        vpCurrentCandidates.push_back(pCurrentHostNode);
                }else
                    if ((pCurrentGuestNode->mpSourceEdges.size() == pCurrentHostNode->mpSourceEdges.size())
                        && (pCurrentGuestNode->mpTargetEdges.size() == pCurrentHostNode->mpTargetEdges.size())
                        && (pCurrentGuestNode->mpUndirectedEdges.size() == pCurrentHostNode->mpUndirectedEdges.size()))
                        vpCurrentCandidates.push_back(pCurrentHostNode);
            }
        }
        vvpAllCandidates.push_back(vpCurrentCandidates);
    }
    return vvpAllCandidates;
}

bool GML::generateMatchedSubgraphs(const vector<vector<Node *> > & vvpNodeCandidates, const BigGML & HostBigGML,
                                   const BigGML & GuestBigGML, stack<map<Node *,Node *> > &smMatchGraph)
//vvpNodeCandidates: all candidates of nodes of guest graph in host graph; HostBigGML: BigGML of the host graph
//GuestBigGML: BigGML of the guest graph
//smMatchGraph: stack of details of all matches; vpGuestNodes: vector of node pointers of guest graph
//return true if at least one whole match, otehrwise, return false
{
    map<Node *,Node *> mMatchGraph;
    vector<size_t> vPositionRecord, vSize;
    bool bHasOneMatch = true, bCurrentMatch, bNextComb;
    size_t i;
    int iPos1st = -1, iPos2nd = -1;
    
    for (i=0; i<vvpNodeCandidates.size(); i++)
        if (vvpNodeCandidates.at(i).empty())
        {
            bHasOneMatch = false;
            break;
        }else
            vSize.push_back(vvpNodeCandidates.at(i).size());
    if (bHasOneMatch)
    {
        bHasOneMatch = false;
        mMatchGraph.clear();
        bNextComb = generateNextCombination(vPositionRecord, vSize, iPos1st);
        while (bNextComb)
        {
            bCurrentMatch = checkOneSubgraphPair(vvpNodeCandidates, HostBigGML, GuestBigGML, vPositionRecord,
                                                 iPos1st, iPos2nd);
            if (bCurrentMatch)
            {
                bHasOneMatch = true;
                buildHostGuestMap(vvpNodeCandidates, vPositionRecord, GuestBigGML.vpNodes, mMatchGraph);
                smMatchGraph.push(mMatchGraph);
            }
            bNextComb = generateNextCombination(vPositionRecord, vSize, iPos1st);
        }
    }    
    return bHasOneMatch;
}

bool GML::generateNextCombination(vector<size_t> &vPositionRecord,const vector<size_t> &vSize,int movePosition/* = -1 */)
//vPositionRecord: vector of current positions; vSize: vector of sizes of each positions
//movePosition: the position should be changed
//return true if successfully generate the next combination, otherwise false
{
    bool bHasNext = true;
    size_t i, startPosition;
    
    startPosition = max(movePosition, 0);
    for (i=0; i<vSize.size(); i++)
        if (vSize.at(i) == 0)
            bHasNext = false;
    if (bHasNext)
    {
        if (vPositionRecord.empty())
        {
            for (i=0; i<vSize.size(); i++)
                vPositionRecord.push_back(0);
        }else
        {
            for (i=startPosition; i<vPositionRecord.size(); i++)
                if (vPositionRecord.at(i) < (vSize.at(i) - 1))
                {
                    vPositionRecord.at(i) += 1;
                    break;
                }
                else
                    vPositionRecord.at(i) = 0;
            if (i == vPositionRecord.size())
                bHasNext = false;
        }
    }
    
    return bHasNext;
}

bool GML::storeNodesInMap(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                          map<int, Node *> & mpNodes, int & iPosSameNode1st)
//vvpNodes: vector of vector of pointers of node candidates;vPositionRecord: vector of node candidate choices
//mpNodes: map of pointer of chosen nodes; iPosSameNode1st: position of chosen node, which has been chosen before
//return true if all chosen nodes are unique; otherwise, false
{
    bool bNoSameNodes = true;
    vector<size_t>::const_reverse_iterator reviIterator;
    map<int, Node *>::const_iterator mpIter;
    size_t i;
    
    mpNodes.clear();
    i = vPositionRecord.size();
    for (reviIterator = vPositionRecord.rbegin(); reviIterator != vPositionRecord.rend(); ++reviIterator)
    {
        i--;
        mpIter = mpNodes.find(vvpNodes.at(i).at(*reviIterator)->iNodeId);
        if (mpIter == mpNodes.end())
            mpNodes.insert(make_pair (vvpNodes.at(i).at(*reviIterator)->iNodeId, vvpNodes.at(i).at(*reviIterator)));
        else
        {
            bNoSameNodes = false;
            iPosSameNode1st = (int)i;
            break;
        }
    }
    
    return bNoSameNodes;
}

bool GML::checkOneSubgraphPair(const vector<vector<Node *> > & vvpNodeCandidates, const BigGML & HostBigGML,
                               const BigGML & GuestBigGML, const vector<size_t> &vPositionRecord,
                               int &iPos1st, int &iPos2nd)
//vvpNodeCandidates: all candidates of nodes of guest graph in host graph; HostBigGML: BigGML of the host graph
//GuestBigGML: BigGML of the guest graph
//vPositionRecord: vector of current positions; vSize: vector of sizes of each positions;
//if return false, at most two positions are involved: iPos1st and iPos2nd
//return true if this pair matches, otherwise false
{
    bool bMatch;
    map<int, Node *> mpSubHostNodes;
    vector<map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > > > vmDanglingHost;
    vector<map<pair<string, NodeSide>, int> > vmDanglingHostCount;
    
    iPos1st = -1;
    iPos2nd = -1;
    bMatch = storeNodesInMap(vvpNodeCandidates, vPositionRecord, mpSubHostNodes, iPos1st);
    if (bMatch)
    {
        bMatch = compareRegularEdges(vvpNodeCandidates, vPositionRecord, GuestBigGML.vpNodes, GuestBigGML.mNodeIdPos,
                                     GuestBigGML.vpEdges, iPos1st, iPos2nd);
        if (bMatch)
        {
            vmDanglingHost.clear();
            vmDanglingHostCount.clear();
            buildDanglingEdgesMaps(vvpNodeCandidates, vPositionRecord, HostBigGML.vpNodes, HostBigGML.mNodeIdPos,
                                   mpSubHostNodes, vmDanglingHost, vmDanglingHostCount);
            bMatch = compareDanglingEdges(vvpNodeCandidates, vPositionRecord, GuestBigGML.mNodeIdPos,
                                          vmDanglingHost, vmDanglingHostCount, GuestBigGML.vpStartDangling,
                                          GuestBigGML.vpEndDangling, GuestBigGML.vpUndirectedStartDangling,
                                          GuestBigGML.vpUndirectedEndDangling);
        }
    }
    
    return bMatch;
}

bool GML::compareDanglingEdges(const vector<vector<Node *> > &vvpNodeCandidates, const vector<size_t> &vPositionRecord,
                               const map <int, size_t> & mGuestNodeIdPos,
                               const vector<map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > > &vmDanglingHost,
                               vector<map<pair<string,NodeSide>,int> > &vmDanglingHostCount,
                               const vector<Edge *> &vpGuestStartDandgling, const vector<Edge *> &vpGuestEndDangling,
                               const vector<Edge *> &vpGuestUndirectedStartDangling,
                               const vector<Edge *> &vpGuestUndirectedEndDangling)
//vvpNodeCandidates: all candidates of nodes of guest graph in host graph;
//vPositionRecord: vector of current positions; vSize: vector of sizes of each positions;
//mGuestNodeIdPos: map between node id and node position
//vmDangling and vmDanglingCount: vector of maps for dangling edges of host subgraph
//vpGuestStartDandgling, vpGuestEndDangling: vectors of directed dangling edges with start or end nodes in the guest graph
//vpGuestUndirectedStartDangling,vpGuestUndirectedEndDangling: vectors of undirected dangling edges in the guest graph
{
    bool bMatch = true;
    Node *pGuestNode;
    Edge *pGuestEdge;
    size_t i;
    
    for (i=0; i<vpGuestStartDandgling.size(); i++)
    {
        pGuestEdge = vpGuestStartDandgling.at(i);
        pGuestNode = pGuestEdge->pTargetNode;
        bMatch = compareDanglingEdge(pGuestNode, pGuestEdge, vvpNodeCandidates, vPositionRecord, mGuestNodeIdPos,
                                     vmDanglingHost, vmDanglingHostCount, END);
        if (!bMatch)
            break;
    }
    if (bMatch)
        for (i=0; i<vpGuestEndDangling.size(); i++)
        {
            pGuestEdge = vpGuestEndDangling.at(i);
            pGuestNode = pGuestEdge->pSourceNode;
            bMatch = compareDanglingEdge(pGuestNode, pGuestEdge, vvpNodeCandidates, vPositionRecord, mGuestNodeIdPos,
                                         vmDanglingHost, vmDanglingHostCount, START);
            if (!bMatch)
                break;
        }
    if (bMatch)
        for (i=0; i<vpGuestUndirectedStartDangling.size(); i++)
        {
            pGuestEdge = vpGuestUndirectedStartDangling.at(i);
            pGuestNode = pGuestEdge->pTargetNode;
            bMatch = compareDanglingEdge(pGuestNode, pGuestEdge, vvpNodeCandidates, vPositionRecord, mGuestNodeIdPos,
                                         vmDanglingHost, vmDanglingHostCount, UND);
            if (!bMatch)
                break;
        }
    if (bMatch)
        for (i=0; i<vpGuestUndirectedEndDangling.size(); i++)
        {
            pGuestEdge = vpGuestUndirectedEndDangling.at(i);
            pGuestNode = pGuestEdge->pSourceNode;
            bMatch = compareDanglingEdge(pGuestNode, pGuestEdge, vvpNodeCandidates, vPositionRecord, mGuestNodeIdPos,
                                         vmDanglingHost, vmDanglingHostCount, UND);
            if (!bMatch)
                break;
        }
    return bMatch;
}

bool GML::compareDanglingEdge(Node *pGuestNode, Edge *pGuestEdge, const vector<vector<Node *> > &vvpNodeCandidates,
                              const vector<size_t> &vPositionRecord, const map<int, size_t> &mGuestNodeIdPos,
                              const vector<map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > > > &vmDanglingHost,
                              vector<map<pair<string, NodeSide>, int> > &vmDanglingHostCount,
                              const NodeSide & curNodeSide)
//pGuestNode: pointer of the guest node; pGuestEdge: pointer of the guest edge;
//vvpNodeCandidates: all candidates of nodes of guest graph in host graph;
//vPositionRecord: vector of current positions; vSize: vector of sizes of each positions;
//mGuestNodeIdPos: map between node id and node position
//vmDanglingHost and vmDanglingCount: vector of maps for dangling edges of host subgraph
//vpGuestStartDandgling, vmDanglingHostCount: vectors of directed dangling edges with start or end nodes in the guest graph
//curNodeSide: indicate the node is on starting side, ending site, or on an undirected edge
{
    bool bMatch = false;
    size_t iGuestNodePos;
    map<pair<string, NodeSide>, int>::iterator mDanglingHostCountIt;
    map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > >::const_iterator mDanglingHostIt;
    
    iGuestNodePos = mGuestNodeIdPos.at(pGuestNode->iNodeId);
    //mDanglingHostCountIt = vmDanglingHostCount.at(iGuestNodePos).find(make_pair(pGuestEdge->sLabel, curNodeSide));
    mDanglingHostCountIt = searchDanglingCountMap(vmDanglingHostCount.at(iGuestNodePos), pGuestEdge->sLabel, curNodeSide);
    if ((mDanglingHostCountIt != vmDanglingHostCount.at(iGuestNodePos).end())&&(mDanglingHostCountIt->second > 0))
    {
        mDanglingHostIt = searchDanglingMap(vmDanglingHost.at(iGuestNodePos), pGuestEdge->sLabel, curNodeSide);
        if (compareAnnotationEdge(mDanglingHostIt->second.at(mDanglingHostCountIt->second  - 1).first->sLabel, pGuestEdge->sLabel))
        //if (compareAnnotationEdge(vmDanglingHost.at(iGuestNodePos).at(make_pair(pGuestEdge->sLabel, curNodeSide))
        //                          .at(mDanglingHostCountIt->second  - 1).first->sLabel, pGuestEdge->sLabel))
        {
            mDanglingHostCountIt->second -= 1;
            bMatch = true;
        }
    }
    return bMatch;
}

map<pair<string, NodeSide>, int>::iterator GML::searchDanglingCountMap(map<pair<string, NodeSide>, int> &mCount, const string &sLabel,
                                                                      const NodeSide & curNodeSide)
//mCount: map from edges with specific label and direction to number count; sLabel: specific sLabel for search
//curNodeSide: edge direction for search
{
    string sExactLabel;
    map<pair<string, NodeSide>, int>::iterator mDanglingCountIt;
    
    if (sLabel.find("*ANY*") == string::npos)
        mDanglingCountIt = mCount.find(make_pair(sLabel, curNodeSide));
    else
    {
        for (mDanglingCountIt = mCount.begin(); mDanglingCountIt != mCount.end(); ++mDanglingCountIt)
        {
            sExactLabel = mDanglingCountIt->first.first;
            if ((mDanglingCountIt->first.second == curNodeSide) && (compareAnnotationEdge(sExactLabel, sLabel)))
                break;
        }
    }
    
    return mDanglingCountIt;
}

map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > >::const_iterator GML::searchDanglingMap(const map<pair<string, NodeSide>,
                                                                                             vector<pair<Edge *, Node *> > > &mDangling,
                                                                                             const string &sLabel, const NodeSide & curNodeSide)
//mDangling: map from edges with specific label and direction to dangling edges; sLabel: specific sLabel for search
//curNodeSide: edge direction for search
{
    string sExactLabel;
    map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > >::const_iterator mDanglingIt;
    
    if (sLabel.find("*ANY*") == string::npos)
        mDanglingIt = mDangling.find(make_pair(sLabel, curNodeSide));
    else
    {
        for (mDanglingIt = mDangling.begin(); mDanglingIt != mDangling.end(); ++mDanglingIt)
        {
            sExactLabel = mDanglingIt->first.first;
            if ((mDanglingIt->first.second == curNodeSide) && (compareAnnotationEdge(sExactLabel, sLabel)))
                break;
        }
    }
    
    return mDanglingIt;
}



void GML::buildDanglingEdgesMaps(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                                 const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                 const map<int, Node *> &mpSubNodes,
                                 vector<map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > > &vmDangling,
                                 vector<map<pair<string,NodeSide>, int> > &vmDanglingCount)
//vvpNodes: vector of vector of pointers of node candidates; vPositionRecord: vector of node candidate choices
//vpNodes: vector of pointers of nodes in host; mNodeIdPos: map of node id and position in vpNodes,
//mpSubNodes: map of pointer of chosen nodes; vmDangling and vmDanglingCount: vector of maps for dangling edges
{
    size_t i;
    Node * pCurrentNode;
    map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > mDangling;
    map<pair<string,NodeSide>, int> mDanglingCount;
    
    vmDangling.clear();
    vmDanglingCount.clear();
    for (i=0; i<vPositionRecord.size(); i++)
    {
        mDangling.clear();
        mDanglingCount.clear();
        pCurrentNode = vvpNodes.at(i).at(vPositionRecord.at(i));
        buildDanglingMapsForOneNode(vpNodes, mNodeIdPos, mpSubNodes, pCurrentNode, mDangling, mDanglingCount);
        vmDangling.push_back(mDangling);
        vmDanglingCount.push_back(mDanglingCount);
    }
}

void GML::buildDanglingMapsForOneNode(const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                      const map<int, Node *> &mpSubNodes, Node *pNode,
                                      map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > & mDangling,
                                      map<pair<string,NodeSide>, int> & mDanglingCount)
//vpNodes: vector of pointers of nodes in host; mNodeIdPos: map of node id and position in vpNodes
//mpSubNodes: map of nodes in the subgraph; pNode: fixed node; mDangling and mDanglingCount : maps for dangling edges
{
    int iAnotherNode;
    map<int, Edge *>::const_iterator mEdgeIt;
    
    mDangling.clear();
    mDanglingCount.clear();
    for (mEdgeIt = pNode->mpSourceEdges.begin(); mEdgeIt != pNode->mpSourceEdges.end(); ++mEdgeIt)
    {
        iAnotherNode = mEdgeIt->first;
        if ((iAnotherNode < 0) || (mpSubNodes.find(iAnotherNode) == mpSubNodes.end()))
            addNodeEdgeToDanglingMaps(vpNodes, mNodeIdPos, iAnotherNode, mEdgeIt->second,mDangling,mDanglingCount,START);
    }
    for (mEdgeIt = pNode->mpTargetEdges.begin(); mEdgeIt != pNode->mpTargetEdges.end(); ++mEdgeIt)
    {
        iAnotherNode = mEdgeIt->first;
        if ((iAnotherNode < 0) || (mpSubNodes.find(iAnotherNode) == mpSubNodes.end()))
            addNodeEdgeToDanglingMaps(vpNodes, mNodeIdPos, iAnotherNode, mEdgeIt->second, mDangling, mDanglingCount, END);
    }
    for (mEdgeIt = pNode->mpUndirectedEdges.begin(); mEdgeIt != pNode->mpUndirectedEdges.end(); ++mEdgeIt)
    {
        iAnotherNode = mEdgeIt->first;
        if ((iAnotherNode < 0) || (mpSubNodes.find(iAnotherNode) == mpSubNodes.end()))
            addNodeEdgeToDanglingMaps(vpNodes, mNodeIdPos, iAnotherNode, mEdgeIt->second, mDangling, mDanglingCount, UND);
    }
}

void GML::addNodeEdgeToDanglingMaps(const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                    const int & iNodeId, Edge *pEdge,
                                    map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > &mDangling,
                                    map<pair<string,NodeSide>, int> &mDanglingCount, const NodeSide & CurNodeside)
//vpNodes: vector of pointers of nodes in host; mNodeIdPos: map of node id and position in vpNodes
//iNodeid: >= 0 nodeid <0 null; pEdge: pointer of related dangling edge
//mDangling and mDanglingCount : maps for dangling edges;
//CurNodeSide: the node side (start, end or undirected) of the current node
{
    Node *pNode;
    map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > >::iterator mDanglingIt;
    vector<pair<Edge *, Node *> > vpInit ;
    if (iNodeId < 0)
        pNode = nullptr;
    else
        pNode = vpNodes.at(mNodeIdPos.at(iNodeId));
    mDanglingIt = mDangling.find(make_pair(pEdge->sLabel, CurNodeside));
    if (mDanglingIt == mDangling.end())
    {
        vpInit.clear();
        vpInit.push_back(pair<Edge *, Node *>(pEdge, pNode));
        mDangling[make_pair(pEdge->sLabel, CurNodeside)] = vpInit;
    }else
        mDanglingIt->second.push_back(pair<Edge *, Node *>(pEdge, pNode));
    for (mDanglingIt = mDangling.begin(); mDanglingIt != mDangling.end(); ++mDanglingIt)
        mDanglingCount[mDanglingIt->first] = (int)(mDanglingIt->second.size());
}

bool GML::compareRegularEdges(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                              const vector<Node *> &vpGuestNodes, const map<int, size_t> &mGuestNodesIdPos,
                              const vector<Edge *> &vpAllGuestEdges, int &iPosBadNode1st, int &iPosBadNode2nd)
//vvpNodes: vector of vector of pointers of node candidates;vPositionRecord: vector of node candidate choices
//vpGuestNodes: vector of pointers of nodes of the guest graph; mGuestNodesIdPos: map between guest node id and position
//vpAllGuestEdges: vector of pointers of all edges of guest graph;
//iPosBadNode1st and iPosBadNode2nd, poistion of nodes of the unmatched edge.
{
    bool bMatch= true;
    Edge *pGuestEdge;
    Node *pHostSourceNode;
    size_t i, iGuestSourceNodePos, iGuestTargetNodePos, iHostTargetNodeId;

    for (i=0; i<vpAllGuestEdges.size(); i++)
    {
        pGuestEdge = vpAllGuestEdges.at(i);
        if ((pGuestEdge->pSourceNode != nullptr) && (pGuestEdge->pTargetNode != nullptr))
        {
            iGuestSourceNodePos = mGuestNodesIdPos.at(pGuestEdge->pSourceNode->iNodeId);
            iGuestTargetNodePos = mGuestNodesIdPos.at(pGuestEdge->pTargetNode->iNodeId);
            pHostSourceNode = vvpNodes.at(iGuestSourceNodePos).at(vPositionRecord.at(iGuestSourceNodePos));
            iHostTargetNodeId = vvpNodes.at(iGuestTargetNodePos).at(vPositionRecord.at(iGuestTargetNodePos))->iNodeId;
            if (pGuestEdge->getDirected())
                bMatch = compareRegularEdge(pHostSourceNode->mpSourceEdges, pGuestEdge, iHostTargetNodeId);
            else
                bMatch = compareRegularEdge(pHostSourceNode->mpUndirectedEdges, pGuestEdge, iHostTargetNodeId);
            if (!bMatch)
            {
                iPosBadNode1st = (int)min(iGuestSourceNodePos, iGuestTargetNodePos);
                iPosBadNode2nd = (int)max(iGuestSourceNodePos, iGuestTargetNodePos);
                break;
            }
        }
    }
    return bMatch;
}

bool GML::compareRegularEdge(const map<int, Edge *> &mEdges, const Edge* pGuestEdge, const size_t &iHostTargetNodeId)
//mEdges: map of edges of host graph; pGuestEdge: pointer of the guest edge
//iHostTargetNodeId: Id of target node of the edge on the host graph
{
    bool bMatch = false;
    Edge *pHostEdge;
    map<int, Edge *>::const_iterator itHostEdge;
    
    itHostEdge = mEdges.find((int)iHostTargetNodeId);
    if (itHostEdge != mEdges.end())
    {
        pHostEdge = itHostEdge->second;
        if (compareAnnotationEdge(pHostEdge->sLabel, pGuestEdge->sLabel))
            bMatch = true;
    }
    
    return bMatch;
}

void GML::buildHostGuestMap(const vector<vector<Node *> > &vvpNodeCandidates, const vector<size_t> &vPositionRecord,
                            const vector<Node *> vpGuestNodes, map<Node *, Node *> &mMatchGraph)
//vvpNodeCandidates: vector of vector of pointers of node candidates;vPositionRecord: vector of node candidate choices
//vpGuestNodes: vector of points of guest nodes; mMatchGraph: map of host nodes and guest nodes

{
    size_t i;
    Node *pHostNode, *pGuestNode;
    
    mMatchGraph.clear();
    if ((vpGuestNodes.size() != vPositionRecord.size()) || (vpGuestNodes.size() != vvpNodeCandidates.size()))
    {
        cerr<<"sizes of vectors don't match"<<endl;
        exit(1);
    }
    for (i=0; i<vpGuestNodes.size(); i++)
    {
        pGuestNode = vpGuestNodes.at(i);
        pHostNode  = vvpNodeCandidates.at(i).at(vPositionRecord.at(i));
        mMatchGraph[pHostNode] = pGuestNode;
    }
}

bool GML::compareAnnotationEdge(const string &sHostAnnotation, const string &sGuestAnnotation)
//sHostAnnotation: annotation of host edge; sGuestAnnotation: annotation of guest edge
//if annotations match, return true; otherwise, return false
{
    bool bLabelMatch;
    
    bLabelMatch = compareAnnotation(sHostAnnotation, sGuestAnnotation);
    
    return bLabelMatch;
}

bool GML::compareAnnotationNode(const string &sHostAnnotation, const string &sGuestAnnotation)
//sHostAnnotation: annotation of host node; sGuestAnnotation: annotation of guest node
//if annotations match, return true; otherwise, return false
{
    bool bLabelMatch;
    
    bLabelMatch = compareAnnotation(sHostAnnotation, sGuestAnnotation);
    
    return bLabelMatch;
}

bool GML::compareAnnotation(const string &sHostAnnotation, const string &sGuestAnnotation)
//sHostAnnotation: annotation from host graph; sGuestAnnotation: annotation from guest graph
//if annotations match, return true; otherwise, return false
{
    bool bLabelMatch = false;
    size_t i;
    vector<string> vsHost, vsGuest;
    
    if ((sGuestAnnotation == "*ANY*") || (sGuestAnnotation == "\"*ANY*\""))
    {
        bLabelMatch = true;
    }
    else
        if ((sHostAnnotation == "*ANY*") || (sHostAnnotation == "\"*ANY*\"") || (sHostAnnotation == sGuestAnnotation))
        {
            bLabelMatch = true;
        }
        else
        {
            annotationSeparation(sHostAnnotation, vsHost);
            annotationSeparation(sGuestAnnotation, vsGuest);
            if (vsHost.size() == vsGuest.size())
            {
                for (i = 0; i < vsHost.size(); i++)
                {
                    if ((vsHost.at(i)!=vsGuest.at(i))&&(vsHost.at(i)!="*ANY*")&&(vsGuest.at(i)!="*ANY*"))
                        break;
                }
                if (i == vsHost.size())
                    bLabelMatch = true;
            }
        }
    
    return bLabelMatch;
}

bool GML::IsInitial(bool bPartialParse)
//bPartialParse : true if partial parse mode, by default it is false
//return true if the current GML contains initial status 
{
    vector<Node *>::const_iterator itpNode;
    string sNodeLabel;
    bool bInitital = false;
    
    if (bPartialParse)
    {
        if (vpNodes.size() > 0)
            for (itpNode = vpNodes.begin(); itpNode != vpNodes.end(); ++itpNode)
            {
                sNodeLabel = (*itpNode)->sLabel;
                if ((sNodeLabel == "Initial")||(sNodeLabel == "initial")||(sNodeLabel == "INITIAL")
                    ||(sNodeLabel == "\"Initial\"")||(sNodeLabel == "\"initial\"")||(sNodeLabel == "\"INITIAL\""))
                {
                    bInitital = true;
                    break;
                }
            }
    }else
    {
        if (vpNodes.size() == 1)
        {
            sNodeLabel = vpNodes.at(0)->sLabel;
            if ((sNodeLabel == "Initial")||(sNodeLabel == "initial")||(sNodeLabel == "INITIAL")
                ||(sNodeLabel == "\"Initial\"")||(sNodeLabel == "\"initial\"")||(sNodeLabel == "\"INITIAL\""))
            {
                bInitital = true;
            }
        }
    }
    
    return true;
}

void GML::annotationSeparation(const string &sOriginalStr, vector<string> &vsSegments)
//sOriginalStr: original annotation string; vsSegments: vector of separated strings
{
    istringstream ss(sOriginalStr);
    string token;
    
    vsSegments.clear();
    while(getline(ss, token, '|'))
    {
        vsSegments.push_back(token);
    }
}
