//
//  GML.h
//  egg
//
//  Created by Yingfeng Wang on 1/24/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__GML__
#define __egg__GML__

#include <stdio.h>
#include <stack>
#include <algorithm>
#include <sstream>

#include "GraphBasics.h"

using namespace std;

struct BigGML
//for merged GMLs. Destructor of GML will free all memory all nodes and edges, while BigGML doesn't have this problem.
{
    vector<Node *>  vpNodes; // pointers of all nodes of the GML
    map<int, size_t> mNodeIdPos; // map between node id and node position in vpNodes;
    vector<Edge *>  vpEdges; // pointers of all edges of the GML;
    vector<Edge *>  vpStartDangling; // pointers of all dangling edges without specifying starting nodes
    vector<Edge *>  vpEndDangling; // pointers of all dangling edges without specifying ending nodes
    vector<Edge *>  vpUndirectedStartDangling; // pointers of all undirected dangling edges dangling on starting node
    vector<Edge *>  vpUndirectedEndDangling; // pointers of all undirected dangling edges dangling on starting node
};

class GML
// A GML is a graph, which has no guarantee to be a connected graph, while there is no edge between two GMLs.
{
    void copyEdgeMap(const map<int, Edge *> &mSource, map<int, Edge *> &mDestination,
                     const map<int, Edge *> &mEdgeIdPointer); //copy edge map
    //copy edge vector
    void copyEdgeVector(const vector<Edge *> &vSource, vector<Edge *> &vDestination, const map<int, Edge *> &mEdgeIdPointer);
    void sortNodes(); // sort nodes of the GML;
    //find all candidates of nodes GuestBigGML in HostBigGML
    vector<vector<Node *> > findAllMatchedNodesPairs(const BigGML & HostBigGML, const BigGML & GuestBigGML);
    //generate all matched subgraphs
    bool generateMatchedSubgraphs(const vector<vector<Node *> > &vvpNodeCandidates, const BigGML & HostBigGML,
                                  const BigGML & GuestBigGML, stack<map<Node *,Node *> > &smMatchGraph);
    //check whether two subgraphs match
    bool checkOneSubgraphPair(const vector<vector<Node *> > &vvpNodeCandidates, const BigGML & HostBigGML,
                              const BigGML & GuestBigGML, const vector<size_t> & vPositionRecord,
                              int & iPos1st, int & iPos2nd);
    //store all node pointers in the map
    bool storeNodesInMap(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                         map<int, Node *> & mpNodes, int & iPosSameNode1st);
    //compare all regular edges between guest graph and the subgraph of the host graph
    bool compareRegularEdges(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                             const vector<Node *> & vpGuestNodes, const map<int, size_t> &mGuestNodesIdPos,
                             const vector<Edge *> & vpAllGuestEdges, int & iPosBadNode1st, int & iPosBadNode2nd);
    //compare one regular edge
    bool compareRegularEdge(const map<int, Edge *> & mEdges, const Edge * pGuestEdge, const size_t & iHostTargetNodeId);
    //build the vector of maps of dangling edges for the subgraph
    void buildDanglingEdgesMaps(const vector<vector<Node *> > &vvpNodes, const vector<size_t> &vPositionRecord,
                                const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                const map<int, Node *> & mpSubNodes,
                                vector<map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > > > & vmDangling,
                                vector<map<pair<string, NodeSide>, int> > & vmDanglingCount);
    //build the vector of maps of dangling edges related to one node
    void buildDanglingMapsForOneNode(const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                     const map<int, Node *> &mpSubNodes,
                                     Node *pNode,map<pair<string,NodeSide>, vector<pair<Edge *,Node *> > > &mDangling,
                                     map<pair<string, NodeSide>, int> & mDanglingCount);
    //add the node and the edge to dangling maps
    void addNodeEdgeToDanglingMaps(const vector<Node *> & vpNodes, const map <int, size_t> & mNodeIdPos,
                                   const int & iNodeId, Edge * pEdge,
                                   map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > & mDangling,
                                   map<pair<string,NodeSide>, int> & mDanglingCount, const NodeSide & CurNodeside);
    //compare all dangling edges between guest graph and the subgraph of the host graph
    bool compareDanglingEdges(const vector<vector<Node *> > &vvpNodeCandidates, const vector<size_t> &vPositionRecord,
                              const map <int, size_t> & mGuestNodeIdPos,
                              const vector<map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > > & vmDanglingHost,
                              vector<map<pair<string,NodeSide>, int> > &vmDanglingHostCount,
                              const vector<Edge *> &vpGuestStartDandgling, const vector<Edge *> &vpGuestEndDangling,
                              const vector<Edge *> &vpGuestUndirectedStartDangling,
                              const vector<Edge *> & vpGuestUndirectedEndDangling);
    //compare one dangling edge
    bool compareDanglingEdge(Node *pGuestNode, Edge *pGuestEdge, const vector<vector<Node *> > &vvpNodeCandidates,
                             const vector<size_t> &vPositionRecord, const map <int, size_t> & mGuestNodeIdPos,
                             const vector<map<pair<string,NodeSide>, vector<pair<Edge *, Node *> > > > &vmDanglingHost,
                             vector<map<pair<string,NodeSide>, int> > &vmDanglingHostCount, const NodeSide & curNodeSide);
    //build the map between nodes of host and that of guest
    void buildHostGuestMap(const vector<vector<Node *> > &vvpNodeCandidates, const vector<size_t> &vPositionRecord,
                           const vector<Node *> vpGuestNodes, map<Node *,Node *> & mMatchGraph);
    //compare annotations of edges
    bool compareAnnotationEdge(const string &sHostAnnotation, const string &sGuestAnnotation);
    //compare annotations of nodes
    bool compareAnnotationNode(const string &sHostAnnotation, const string &sGuestAnnotation);
    //compare annotations 
    bool compareAnnotation(const string &sHostAnnotation, const string &sGuestAnnotation);
    //search dangling edge count map 
    map<pair<string, NodeSide>, int>::iterator searchDanglingCountMap(map<pair<string, NodeSide>, int> &mCount, const string &sLabel,
                                                                      const NodeSide & curNodeSide);
    //search dangling edge map
    map<pair<string, NodeSide>, vector<pair<Edge *, Node *> > >::const_iterator searchDanglingMap(const map<pair<string, NodeSide>,
                                                                                                  vector<pair<Edge *, Node *> > > &mDangling,
                                                                                                  const string &sLabel,
                                                                                                  const NodeSide & curNodeSide);
    //return flexible status of node
    bool isFlexible(Node * pCurrentNode);
public:

    DirectionStatus eDirected; //directed status of the subgraph. eDirected of production has higher priority
    vector<Node *>  vpNodes; // pointers of all nodes of the GML
    map<int, size_t> mNodeIdPos; // map between node id and node position in vpNodes;
    vector<Edge *>  vpEdges; // pointers of all edges of the GML;
    vector<Edge *>  vpStartDangling; // pointers of all dangling edges without specifying starting nodes
    vector<Edge *>  vpEndDangling; // pointers of all dangling edges without specifying ending nodes
    vector<Edge *>  vpUndirectedStartDangling; // pointers of all undirected dangling edges dangling on starting node
    vector<Edge *>  vpUndirectedEndDangling; // pointers of all undirected dangling edges dangling on starting node
    map <string, string> mGMLInfo; // other information of GML;
    
    GML(const GML &oriGML); //copy GML
    GML();
    ~GML();
    void setDirected(DirectionStatus eDirected); // set value of eDirected
    DirectionStatus getDirected(); //get value of eDirected
    void dataOptimization(); // optimize data in GML by reorganizing them;
    // compare two graphs
    bool graphMatch(const vector <GML *> & vpHostContextGraph, const vector <GML *> & vpHostGraph,
                    const vector <GML *> & vpGuestContextGraph, const vector <GML *> & vpGuestGraph,
                    stack < map<Node *, Node *> > & smMatchGraph, bool bSimpleMode = false);
    //compare two graphs
    bool graphMatch(const vector <GML *> & vpHostGraph, const vector <GML *> & vpGuestGraph,
                    stack < map<Node *, Node *> > & smMatchGraph, bool bSimpleMode = false);
    //compare two BigGMLs
    bool BigGMLMatch(const BigGML & HostBigGML, const BigGML & GuestBigGML, stack < map<Node *,Node *> > & smMatchGraph);
    //compare two GMLs
    bool GMLMatch(const GML & HostGML, const GML & GuestGML, stack < map<Node *, Node *> > & smMatchGraph);
    //compare two graphs under simple mode, which assumes graphs are undirected, and have no dangling edges
    bool graphMatchSimple(const vector <GML *> & vpHostContextGraph, const vector <GML *> & vpHostGraph,
                          const vector <GML *> & vpGuestContextGraph, const vector <GML *> & vpGuestGraph,
                          stack < map<Node *, Node *> > & smMatchGraph);
    //compare two graphs under simple mode, which assumes graphs are undirected, and have no dangling edges
    bool graphMatchSimple(const vector <GML *> & vpHostGraph, const vector <GML *> & vpGuestGraph,
                          stack < map<Node *, Node *> > & smMatchGraph);
    //merge GMLs nodes and edges into one BigGML
    static void mergeGMLs(const vector<GML *> & vpGMLs, BigGML & mergedGML);
    //transfer a GML to a BigGML
    BigGML GML2BigGML(const GML & inputGML);
    //generate the next position combination based on the current record
    static bool generateNextCombination(vector<size_t> & vPositionRecord, const vector<size_t> & vSize,
                                        int movePosition = -1);
    bool IsInitial(bool bPartialParse = false);//check whether initial status is in the GML
    //separate a string with separator "|"
    void annotationSeparation(const string &sOriginalStr, vector<string> &vsSegments);
    // checking content of stack for debugging
    template <typename BasicDataType>
    void debugStack(stack<BasicDataType> sInputStack)
    {//sInputStack: the stack for debugging
        vector<BasicDataType> vDebugVector;
        vDebugVector.clear();
        while (!sInputStack.empty())
        {
            vDebugVector.push_back(sInputStack.top());
            sInputStack.pop();
        }
    }
};



#endif /* defined(__egg__GML__) */
