//
//  GraphBasics.cpp
//  egg
//
//  Created by Yingfeng Wang on 1/1/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#include "GraphBasics.h"

Edge::Edge()
{
    bDirected = false;
    sLabel = "";
    iEdgeId = -1;
    iSource = -1;
    iTarget = -1;
    pSourceNode = nullptr;
    pTargetNode = nullptr;
}

Edge::~Edge()
{
    RemoveEdgeLinksWithNodes(this);
}

void Edge::setDirected(bool bDirected)
// true: directed edge; false: undirected edge; the bDirected of GML has high priority
{
    this->bDirected = bDirected;
}

bool Edge::getDirected() const
//returen bDirected
{
    return bDirected;
}

void Edge::FailToFindEdgeOnNode(int iCurrentEdgeId, int iCurrentNodeId)
//iCurrentEdgeId: id of edge which is searched
//iCurrentNodeId: id of node, on which, the edge is searched
{
    cerr<<"edge "<<iCurrentEdgeId<<" is not connected to node "<<iCurrentNodeId<<endl;
    exit(1);
}

void Edge::RemoveEdgeLinksWithNodes(Edge * pCurrentEdge)
//pCurrentEdge: pointer of current edge
{
    Node *pSourceNode, *pTargetNode;
    map<int, Edge *>::iterator mIpEdgeIt;

    pSourceNode = pCurrentEdge->pSourceNode;
    pTargetNode = pCurrentEdge->pTargetNode;
    if (pSourceNode != nullptr)
    {
        if (pCurrentEdge->getDirected())
        {
            mIpEdgeIt = findEdge(pSourceNode->mpSourceEdges, pCurrentEdge->iEdgeId);
            if ( mIpEdgeIt != pSourceNode->mpSourceEdges.end())
                mIpEdgeIt->second = nullptr;
            else
                FailToFindEdgeOnNode(pCurrentEdge->iEdgeId, pSourceNode->iNodeId);
        }
        else
        {
            mIpEdgeIt = findEdge(pSourceNode->mpUndirectedEdges, pCurrentEdge->iEdgeId);
            if ( mIpEdgeIt !=pSourceNode->mpUndirectedEdges.end())
                mIpEdgeIt->second = nullptr;
            else
                FailToFindEdgeOnNode(pCurrentEdge->iEdgeId, pSourceNode->iNodeId);
        }
        pCurrentEdge->pSourceNode = nullptr;
    }
    if (pTargetNode != nullptr)
    {
        if (pCurrentEdge->getDirected())
        {
            mIpEdgeIt = findEdge(pTargetNode->mpTargetEdges, pCurrentEdge->iEdgeId);
            if (mIpEdgeIt != pTargetNode->mpTargetEdges.end())
                mIpEdgeIt->second = nullptr;
            else
                FailToFindEdgeOnNode(pCurrentEdge->iEdgeId, pTargetNode->iNodeId);
        }
        else
        {
            mIpEdgeIt = findEdge(pTargetNode->mpUndirectedEdges, pCurrentEdge->iEdgeId);
            if (mIpEdgeIt != pTargetNode->mpUndirectedEdges.end())
                mIpEdgeIt->second = nullptr;
            else
                FailToFindEdgeOnNode(pCurrentEdge->iEdgeId, pTargetNode->iNodeId);
        }
        pCurrentEdge->pTargetNode = nullptr;
    }
}

int Edge::getSourceNodeId()
//should be executed after pSourceNode and pTargetNode are specified
{
    int iReSourceId;
    
    if (this->iSource >= 0)
        iReSourceId = this->iSource;
    else if (this->pTargetNode == nullptr)
    {
        cerr << "no valid pointer to target node"<<endl;
        exit(1);
    }else if (bDirected)
    {
        if (this->pTargetNode->mpTargetEdges.empty())
            iReSourceId = -1;
        else
            iReSourceId = this->pTargetNode->mpTargetEdges.begin()->first - 1;
    }else
    {
        if (this->pTargetNode->mpUndirectedEdges.empty())
            iReSourceId = -1;
        else
            iReSourceId = this->pTargetNode->mpUndirectedEdges.begin()->first - 1;
    }
    
    return iReSourceId;
}

int Edge::getTargetNodeId()
//should be executed after pSourceNode and pTargetNode are specified
{
    int iReTargetId;
    
    if (this->iTarget >= 0)
        iReTargetId = this->iTarget;
    else if (this->pSourceNode == nullptr)
    {
        cerr << "no valid pointer to source node"<<endl;
        exit(1);
    }else if (bDirected)
    {
        if (this->pSourceNode->mpSourceEdges.empty())
            iReTargetId = -1;
        else
            iReTargetId = this->pSourceNode->mpSourceEdges.begin()->first - 1;
    }else
    {
        if (this->pSourceNode->mpUndirectedEdges.empty())
            iReTargetId = -1;
        else
            iReTargetId = this->pSourceNode->mpUndirectedEdges.begin()->first - 1;
    }
    
    return iReTargetId;
}

map<int, Edge *>::iterator Edge::findEdge(map<int, Edge *> & mpEdges, int iEdgeId)
//mpEdges: map of edges, first is the node id of one related node,
//iEdgeId: the edge id for search
//return iterator of edge
{
    map<int, Edge *>::iterator mpEdgesItor;
    
    for (mpEdgesItor = mpEdges.begin(); mpEdgesItor != mpEdges.end(); ++mpEdgesItor)
        if ((mpEdgesItor->second != nullptr) && (mpEdgesItor->second->iEdgeId == iEdgeId))
            break;
    
    return mpEdgesItor;
}

Node::Node()
{
    sLabel = "";
    iNodeId = -1;
}

Node::~Node()
{
    RemoveNodeLinksWithDirectedEdges(this);
    RemoveNodeLinksWithUndirectedEdges(this);
}

void Node::RemoveNodeLinksWithUndirectedEdges(Node *pCurrentNode)
//pCurrentNode: pointer of current node
{
    map<int, Edge *>::iterator mIpEdgeIt;
    
    for (mIpEdgeIt = pCurrentNode->mpUndirectedEdges.begin(); mIpEdgeIt!=pCurrentNode->mpUndirectedEdges.end(); ++mIpEdgeIt)
    {
        if (mIpEdgeIt->second != nullptr)
        {
            if (pCurrentNode == mIpEdgeIt->second->pSourceNode)
                mIpEdgeIt->second->pSourceNode = nullptr;
            else if (pCurrentNode == mIpEdgeIt->second->pTargetNode)
                mIpEdgeIt->second->pTargetNode = nullptr;
            else
            {
                cerr<<"node with id "<<pCurrentNode->iNodeId<<" is not connected to edge with id "<<mIpEdgeIt->first<<
                    " "<<mIpEdgeIt->second->iEdgeId<<endl;
                exit(1);
            }
            mIpEdgeIt->second = nullptr;
        }
    }
}

void Node::RemoveNodeLinksWithDirectedEdges(Node *pCurrentNode)
//pCurrentNode: pointer of current node
{
    map<int, Edge *>::iterator mIpEdgeIter;
    
    for (mIpEdgeIter = pCurrentNode->mpSourceEdges.begin(); mIpEdgeIter != pCurrentNode->mpSourceEdges.end(); ++mIpEdgeIter)
        if (mIpEdgeIter->second != nullptr)
        {
            if (pCurrentNode == mIpEdgeIter->second->pSourceNode)
                mIpEdgeIter->second->pSourceNode = nullptr;
            else
            {
                cerr<<"node with id "<<pCurrentNode->iNodeId<<" is not connected to edge with id "<<mIpEdgeIter->first<<
                    " "<<mIpEdgeIter->second->iEdgeId<<" as source node"<<endl;
                exit(1);
            }
            mIpEdgeIter->second = nullptr;
        }
    for (mIpEdgeIter = pCurrentNode->mpTargetEdges.begin(); mIpEdgeIter != pCurrentNode->mpTargetEdges.end(); ++mIpEdgeIter)
        if (mIpEdgeIter->second != nullptr)
        {
            if (pCurrentNode == mIpEdgeIter->second->pTargetNode)
                mIpEdgeIter->second->pTargetNode = nullptr;
            else
            {
                cerr<<"node with id "<<pCurrentNode->iNodeId<<" is not connected to edge with id "<<mIpEdgeIter->first<<
                    " "<<mIpEdgeIter->second->iEdgeId<<" as target node"<<endl;
                exit(1);
            }
            mIpEdgeIter->second = nullptr;
        }
}


bool Node::LessLabel(const Node *pNodeA, const Node *pNodeB)
//pNodeA: pointer of Node A
//pNodeB: pointer of Node B
//if sLabel of NodeA is less than that of Node B (string compare), return true, otherwise, return false
{
    bool bLess;
    //if both of pNodeA and pNodeB are nullptr, bLess is false
    if (pNodeB == nullptr)
        bLess = false;
    else if (pNodeA == nullptr)
        bLess = true;
    else if ((pNodeA->sLabel == "*ANY*") && (pNodeB->sLabel != "*ANY*"))
        bLess = true;
    else if ((pNodeA->sLabel != "*ANY*") && (pNodeB->sLabel == "*ANY*"))
        bLess = false;
    else
        bLess = ((pNodeA->sLabel) < (pNodeB->sLabel));
    return bLess;
}

