//
//  GraphBasics.h
//  egg
//
//  Created by Yingfeng Wang on 1/1/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__GraphBasics__
#define __egg__GraphBasics__

#include <stdio.h>
#include <map>
#include <string>
#include <vector>
#include <iostream>

using namespace std;

//DIRECTED: all edges are directed
//HYBRID: some edges are directed, some not
//UNDIRECTED: all edges are undirected
enum DirectionStatus {DIRECTED, HYBRID, UNDIRECTED};

//START: node is on the staring part of the directed edge
//END: node is on the ending part of the directed edge
//UND: node is on an undirected edge
enum NodeSide {START, END, UND};

class Node;

class Edge
{
    bool bDirected; // true: directed edge; false: undirected edge; the eDirected of GML has high priority
    
    //set nullptr to nodes info of edge *pCurrentEdge
    void RemoveEdgeLinksWithNodes(Edge * pCurrentEdge);
    //handle the case if the edge can't be found on the node
    void FailToFindEdgeOnNode(int iCurrentEdgeId, int iCurrentNodeId);
    map<int, Edge *>::iterator findEdge(map<int, Edge *> & mpEdges, int iEdgeId); //search edge in a map
    
public:
    int iEdgeId; // id of edge, unique
    int iSource; // if directed edeg, node id of starting node; if undirected, node id of one node; -1: not specified
    int iTarget; // if directed edeg, node id of ending node; if undirected, node id of one node; -1: not specified
    string sLabel; // label of the edge
    Node * pSourceNode; // pointer of the source node
    Node * pTargetNode; // pointer of the target node
    map <string, string> mEdgeInfo; //map of other information

    Edge();
    ~Edge();
    void setDirected(bool bDirected); // set value of bDirected
    bool getDirected() const; //get value of bDirected
    int getSourceNodeId();// return source node id, get a suitable negative id if no source node id is specified
    int getTargetNodeId();// return target node id, get a suitable negative id if no target node id is specified
};
    
class Node
{
    //set nullptr to source node or target node info of directed edges with node *pCurrentNode
    void RemoveNodeLinksWithDirectedEdges(Node * pCurrentNode);
    //set nullptr to source node or target node info of undirected edges with node *pCurrentNode
    void RemoveNodeLinksWithUndirectedEdges(Node * pCurrentNode);
    
public:
    int iNodeId; //id of node, unique
    string sLabel; //label of the node
    map <string, string> mNodeInfo; //map of other information
    map<int, Edge *> mpSourceEdges; // map of pointers of edges whose source node is the current node, key is target node id
    map<int, Edge *> mpTargetEdges; // map of pointers of edges whose target node is the current node, key is source node id
    //map of pointers of undirected edges connected with the current node, key is another node id
    map<int, Edge *> mpUndirectedEdges;

    Node();
    ~Node();
    static bool LessLabel(const Node * pNodeA, const Node * pNodeB); //compare two nodes according to their labels
};

#endif /* defined(__egg__GraphBasics__) */
