//
//  ParseRegularGraph.cpp
//  egg
//
//  Created by Yingfeng Wang on 2/23/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#include "ParseRegularGraph.h"
ParseRegularGraph::ParseRegularGraph(bool bPartialParse)
{
    this->bPartialParse = bPartialParse;
}



bool ParseRegularGraph::parseGraph(const RegularGraph &curRegularGraph, const GraphGrammar &curGraphGrammar)
//curRegularGraph: current regular graph; curGraphGrammar: current graph grammar
//return true if curGraphGrammar can parse curRegularGraph
{
    bool bParse = false;
    GML compareGML;
    vector <GML *> vpEmptyGML;
    Production * pCurProduction;
    stack<map<Node *, Node *> > smCurMatchGraph;
    map<Node *, Node *> mCurMatchGraph;
    RegularGraph replacedRegularGraph;
    size_t i;
    bool bValidUpdate;//false: if invalid, e.g. the transformed graph contains a pair of nodes connected by 1 more edges.
    
    vpEmptyGML.clear();
    if (isInitialStatus(curRegularGraph, bPartialParse))
        bParse = true;
    else
        for (i=0; i<curGraphGrammar.vpProductions.size(); i++)
        {
            pCurProduction = curGraphGrammar.vpProductions.at(i);
            while(!smCurMatchGraph.empty())
                smCurMatchGraph.pop();
            compareGML.graphMatch(vpEmptyGML, curRegularGraph.vpSubgraphs, pCurProduction->vpContextGraph,
                                           pCurProduction->vpRightHandSide, smCurMatchGraph);
            while(!smCurMatchGraph.empty())
            {
                bValidUpdate = true;
                mCurMatchGraph = smCurMatchGraph.top();
                smCurMatchGraph.pop();
                replacedRegularGraph.clear();
                updateRegularGraph((*pCurProduction), curRegularGraph, mCurMatchGraph, replacedRegularGraph,bValidUpdate);
                if (bValidUpdate)
                {
                    bParse = parseGraph(replacedRegularGraph, curGraphGrammar);
                    if (bParse)
                        break;
                }
            }
            if (bParse)
                break;
        }
    return bParse;
}


bool ParseRegularGraph::parseGraphSimple(const RegularGraph &curRegularGraph, const GraphGrammar &curGraphGrammar)
//This function is for stochastic context free graph grammar (metabolomics)
//curRegularGraph: current regular graph; curGraphGrammar: current graph grammar
//return true if curGraphGrammar can parse curRegularGraph
{
    bool bParse = false;
    // need more codes ...
    return bParse;
}

bool ParseRegularGraph::isInitialStatus(const RegularGraph &curRegularGraph, bool bPartialParse)
//curRegularGraph : input regular graph
//bPartialParse : true: once an initial is reached, stop even if there are other nodes and edges.
//return true, if curRegularGraph is with inital status, otherwise, return false
{
    bool bInitial = false;
    BigGML mergedGML;
    string sNodeLabel;
    vector<Node *>::const_iterator vpNodeIterator;
    
    GML::mergeGMLs(curRegularGraph.vpSubgraphs, mergedGML);
    if ((bPartialParse) && (mergedGML.vpNodes.size() > 0))
    {
        for (vpNodeIterator = mergedGML.vpNodes.begin(); vpNodeIterator != mergedGML.vpNodes.end(); ++vpNodeIterator)
        {
            sNodeLabel = (*vpNodeIterator)->sLabel;
            bInitial = isInitialAnnotation(sNodeLabel);
            if (bInitial)
                break;
        }
    }else
        if ((mergedGML.vpEdges.size() == 0) && (mergedGML.vpNodes.size() == 1))
        {
            sNodeLabel = mergedGML.vpNodes.at(0)->sLabel;
            bInitial = isInitialAnnotation(sNodeLabel);
        }
    
    return bInitial;
}

bool ParseRegularGraph::isInitialAnnotation(const string &sLabel)
//sLabel: annotation label
{
    bool bInitial = false;
    
    if ((sLabel == "Initial")||(sLabel == "initial")||(sLabel == "INITIAL")
        ||(sLabel == "\"Initial\"")||(sLabel == "\"initial\"")||(sLabel == "\"INITIAL\""))
        bInitial = true;
    
    return bInitial;
}

void ParseRegularGraph::updateRegularGraph(const Production &curProduction, const RegularGraph &curRegularGraph,
                                           const map<Node *, Node *> & mCurMatchGraph, RegularGraph &updatedRegularGraph,
                                           bool &bValid)
//curProduction: the current production; curRegularGraph: original regular graph
//mCurMatchGraph: map between nodes of the host graph and that of right hand side
//updatedRegularGraph: the updated regular graph;
//bValid: false: if there is a pair of nodes connected by more than one edges
{
    map <Node *, Node *> mpHostToLeftDangling;
    set <Node *> spReplacedHostNodes;
    int iMaxProductionNodeId = -1, iMaxProductionEdgeId = -1;
    GML *pGML;
    
    organizeHostLeftNodes(curProduction, mCurMatchGraph, mpHostToLeftDangling, spReplacedHostNodes);
    
    pGML = new GML();
    updatedRegularGraph.vpSubgraphs.push_back(pGML);
    addElesOnLeftToGraph(curProduction, updatedRegularGraph, iMaxProductionNodeId, iMaxProductionEdgeId);
    
    addOriginalGraphToNew(curProduction, mpHostToLeftDangling, spReplacedHostNodes, curRegularGraph, updatedRegularGraph,
                          iMaxProductionNodeId+1, iMaxProductionEdgeId+1, bValid);
}

void ParseRegularGraph::addOriginalGraphToNew(const Production &curProduction,
                                              const map <Node *, Node *> &mpHostToLeftDangling,
                                              const set<Node *> &spReplacedHostNodes, const RegularGraph &curRegularGraph,
                                              RegularGraph &updatedRegularGraph, int iMinNodeId, int iMinEdgeId,
                                              bool &bValid)
//curProduction: the current production;mpHostToLeftDangling: map of dangling related nodes between host and left hands
//spReplacedHostNodes: set of nodes of host graph, which will be replaced.
//curRegularGraph: original regular graph; updatedRegularGraph: the updated regular graph;
//iMinNodeId: smallest node id for kept nodes; iMinEdgeId: smallest edge id for kept edges
//bValid: false: if there is a pair of nodes connected by more than one edges
{
    vector<GML *>::const_iterator vpGMLItor;
    
    for (vpGMLItor = curRegularGraph.vpSubgraphs.begin(); vpGMLItor != curRegularGraph.vpSubgraphs.end(); ++vpGMLItor)
    {
        if ((*vpGMLItor)->getDirected() != UNDIRECTED)
            updatedRegularGraph.vpSubgraphs.at(0)->setDirected((*vpGMLItor)->getDirected());
        updatedRegularGraph.vpSubgraphs.at(0)->mGMLInfo.insert((*vpGMLItor)->mGMLInfo.begin(),
                                                               (*vpGMLItor)->mGMLInfo.end());
        addNodesOfOriginalToNew(curProduction, mpHostToLeftDangling, spReplacedHostNodes, *(*vpGMLItor),
                                updatedRegularGraph, iMinNodeId);
        addEdgesOfOriginalToNew(curProduction, mpHostToLeftDangling, spReplacedHostNodes, *(*vpGMLItor),
                                updatedRegularGraph, iMinNodeId, iMinEdgeId, bValid);
        if (!bValid)
            break;
    }
    
    
}

void ParseRegularGraph::addNodesOfOriginalToNew(const Production &curProduction,
                                                const map<Node *, Node *> &mpHostToLeftDangling,
                                                const set<Node *> &spReplacedHostNodes,
                                                const GML &curGML, RegularGraph &updatedRegularGraph,
                                                int iMinNodeId)
//curProduction: the current production;mpHostToLeftDangling: map of dangling related nodes between host and left hands
//spReplacedHostNodes: set of nodes of host graph, which will be replaced.
//curGML: original GML (subgraph); updatedRegularGraph: the updated regular graph;
//iMinNodeId: smallest node id for kept nodes;
{
    vector<Node *>::const_iterator vpNodeItor;
    Node *pCurrentNode;
    
    for (vpNodeItor = curGML.vpNodes.begin(); vpNodeItor != curGML.vpNodes.end(); ++vpNodeItor)
        if (spReplacedHostNodes.find(*vpNodeItor) == spReplacedHostNodes.end())
        {
            pCurrentNode = new Node();
            *pCurrentNode = *(*vpNodeItor);
            pCurrentNode->iNodeId = iMinNodeId + pCurrentNode->iNodeId;
            pCurrentNode->mpSourceEdges.clear();
            pCurrentNode->mpTargetEdges.clear();
            pCurrentNode->mpUndirectedEdges.clear();
            updatedRegularGraph.vpSubgraphs.at(0)->vpNodes.push_back(pCurrentNode);
        }
}

void ParseRegularGraph::addEdgesOfOriginalToNew(const Production &curProduction,
                                                const map<Node *, Node *> &mpHostToLeftDangling,
                                                const set<Node *> &spReplacedHostNodes,
                                                const GML &curGML, RegularGraph &updatedRegularGraph, int iMinNodeId,
                                                int iMinEdgeId, bool &bValid)
//curProduction: the current production;mpHostToLeftDangling: map of dangling related nodes between host and left hands
//spReplacedHostNodes: set of nodes of host graph, which will be replaced.
//curGML: original GML (subgraph); updatedRegularGraph: the updated regular graph;
//iMinNodeId: smallest node id for kept nodes;
//iMinEdgeId: smallest edge id for kept edges;
//bValid: false: if there is a pair of nodes connected by more than one edges
{
    vector<Edge *>::const_iterator pEdgeItor;
    set<pair<int, int> > setEdgeNodesPairs;
    Edge *pCurrentEdge;
    for (pEdgeItor = curGML.vpEdges.begin(); pEdgeItor != curGML.vpEdges.end(); ++ pEdgeItor)
        if ((spReplacedHostNodes.find((*pEdgeItor)->pSourceNode) == spReplacedHostNodes.end())
            || (spReplacedHostNodes.find((*pEdgeItor)->pTargetNode) == spReplacedHostNodes.end()))
        {
            pCurrentEdge = new Edge();
            (*pCurrentEdge) = *(*pEdgeItor);
            pCurrentEdge->iEdgeId = iMinEdgeId + pCurrentEdge->iEdgeId;
            if (spReplacedHostNodes.find((*pEdgeItor)->pSourceNode) == spReplacedHostNodes.end())
                pCurrentEdge->iSource = iMinNodeId + (*pEdgeItor)->iSource;
            else
                pCurrentEdge->iSource = mpHostToLeftDangling.find((*pEdgeItor)->pSourceNode)->second->iNodeId;
            if (spReplacedHostNodes.find((*pEdgeItor)->pTargetNode) == spReplacedHostNodes.end())
                pCurrentEdge->iTarget = iMinNodeId + (*pEdgeItor)->iTarget;
            else
                pCurrentEdge->iTarget = mpHostToLeftDangling.find((*pEdgeItor)->pTargetNode)->second->iNodeId;
            bValid = NewEdgeCheck(setEdgeNodesPairs, pCurrentEdge->iSource, pCurrentEdge->iTarget);
            if (bValid)
            {
                pCurrentEdge->pSourceNode = Production::SearchNode(updatedRegularGraph.vpSubgraphs.at(0)->vpNodes,
                                                                   pCurrentEdge->iSource, true);
                pCurrentEdge->pTargetNode = Production::SearchNode(updatedRegularGraph.vpSubgraphs.at(0)->vpNodes,
                                                                   pCurrentEdge->iTarget, true);
                setEdgeNodesPairs.insert(make_pair(pCurrentEdge->iSource, pCurrentEdge->iTarget));
                setEdgeNodesPairs.insert(make_pair(pCurrentEdge->iTarget, pCurrentEdge->iSource));
                updatedRegularGraph.vpSubgraphs.at(0)->vpEdges.push_back(pCurrentEdge);
                if (pCurrentEdge->getDirected())
                {
                    pCurrentEdge->pSourceNode->mpSourceEdges.insert(make_pair(pCurrentEdge->pTargetNode->iNodeId,
                                                                              pCurrentEdge));
                    pCurrentEdge->pTargetNode->mpTargetEdges.insert(make_pair(pCurrentEdge->pSourceNode->iNodeId,
                                                                              pCurrentEdge));
                }else
                {
                    pCurrentEdge->pSourceNode->mpUndirectedEdges.insert(make_pair(pCurrentEdge->pTargetNode->iNodeId,
                                                                                  pCurrentEdge));
                    pCurrentEdge->pTargetNode->mpUndirectedEdges.insert(make_pair(pCurrentEdge->pSourceNode->iNodeId,
                                                                                  pCurrentEdge));
                }
            }else
            {
                pCurrentEdge->pSourceNode = nullptr;
                pCurrentEdge->pTargetNode = nullptr;
                delete pCurrentEdge;
                break;
            }
        }
}

bool ParseRegularGraph::NewEdgeCheck(const set<pair<int, int> > &setEdgeNodesPairs, int &iSource, int &iTarget)
//setEdgeNodesPairs: set of node id pairs, each pair is a pair of nodes of an edge
//iSource: source node id; iTarget: target node id
//return false the new edge connects two nodes that have been connected
{
    bool bValid = false;
    
    if (setEdgeNodesPairs.find(make_pair(iSource, iTarget)) == setEdgeNodesPairs.end())
        bValid = true;
    
    return bValid;
}

void ParseRegularGraph::addElesOnLeftToGraph(const Production &curProduction, RegularGraph &updatedRegularGraph,
                                             int &iMaxNodeId,int &iMaxEdgeId)
//curProduction: the current production; updatedRegularGraph: the updated regular graph
//iMaxNodeId: the maximum node id on the left side; iMaxEdgeId: the maximum edge id on the left side
{
    vector<GML *>::const_iterator pGMLItor;
    
    for (pGMLItor = curProduction.vpContextGraph.begin(); pGMLItor != curProduction.vpContextGraph.end(); ++pGMLItor)
        addElesOnGMLToGraph( (*(*pGMLItor)), updatedRegularGraph, iMaxNodeId, iMaxEdgeId);
    
    for (pGMLItor = curProduction.vpLeftHandSide.begin(); pGMLItor != curProduction.vpLeftHandSide.end(); ++pGMLItor)
        addElesOnGMLToGraph( (*(*pGMLItor)), updatedRegularGraph, iMaxNodeId, iMaxEdgeId);
    
    Production::AssociateEdgesAndNodes(updatedRegularGraph.vpSubgraphs.at(0));
}

void ParseRegularGraph::addElesOnGMLToGraph(const GML &curGML, RegularGraph &updatedRegularGraph, int &iMaxNodeId, int &iMaxEdgeId)
//curGML: current GML; updatedRegularGraph: the updated regular graph;
//iMaxNodeId: maximum node id in curGML; iMaxEdgeId: maximum edge id in curGML
{
    vector<Node *>::const_iterator pNodeItor;
    Node *pCurrentNode;
    
    for (pNodeItor = curGML.vpNodes.begin(); pNodeItor != curGML.vpNodes.end(); ++pNodeItor)
    {
        if (iMaxNodeId < (*pNodeItor)->iNodeId)
            iMaxNodeId = (*pNodeItor)->iNodeId;
        pCurrentNode = new Node();
        *pCurrentNode = *(*pNodeItor);
        pCurrentNode->mpSourceEdges.clear();
        pCurrentNode->mpTargetEdges.clear();
        pCurrentNode->mpUndirectedEdges.clear();
        updatedRegularGraph.vpSubgraphs.at(0)->vpNodes.push_back(pCurrentNode);
    }
    addEdgesOnGMLToGraph(curGML, updatedRegularGraph, iMaxEdgeId);
    updatedRegularGraph.vpSubgraphs.at(0)->mNodeIdPos = curGML.mNodeIdPos;
}

void ParseRegularGraph::addEdgesOnGMLToGraph(const GML &curGML, RegularGraph &updatedRegularGraph, int &iMaxEdgeId)
//curGML: current GML; updatedRegularGraph: the updated regular graph;
//iMaxEdgeId: maximum edge id in curGML
{
    vector<Edge *>::const_iterator pEdgeItor;
    Edge *pCurrentEdge;
    
    for (pEdgeItor = curGML.vpEdges.begin(); pEdgeItor != curGML.vpEdges.end(); ++ pEdgeItor)
    {
        if (((*pEdgeItor)->pSourceNode != nullptr) && ((*pEdgeItor)->pTargetNode != nullptr))
        {
            if (iMaxEdgeId < (*pEdgeItor)->iEdgeId)
                iMaxEdgeId = (*pEdgeItor)->iEdgeId;
            pCurrentEdge = new Edge();
            *pCurrentEdge = *(*pEdgeItor);
            pCurrentEdge->pSourceNode = Production::SearchNode(updatedRegularGraph.vpSubgraphs.at(0)->vpNodes,
                                                               (*pEdgeItor)->pSourceNode->iNodeId, true);
            pCurrentEdge->pTargetNode = Production::SearchNode(updatedRegularGraph.vpSubgraphs.at(0)->vpNodes,
                                                               (*pEdgeItor)->pTargetNode->iNodeId, true);
            updatedRegularGraph.vpSubgraphs.at(0)->vpEdges.push_back(pCurrentEdge);
            if (pCurrentEdge->getDirected())
            {
                pCurrentEdge->pSourceNode->mpSourceEdges.insert(make_pair(pCurrentEdge->pTargetNode->iNodeId,
                                                                          pCurrentEdge));
                pCurrentEdge->pTargetNode->mpTargetEdges.insert(make_pair(pCurrentEdge->pSourceNode->iNodeId,
                                                                          pCurrentEdge));
            }else
            {
                pCurrentEdge->pSourceNode->mpUndirectedEdges.insert(make_pair(pCurrentEdge->pTargetNode->iNodeId,
                                                                              pCurrentEdge));
                pCurrentEdge->pTargetNode->mpUndirectedEdges.insert(make_pair(pCurrentEdge->pSourceNode->iNodeId,
                                                                              pCurrentEdge));
            }
        }
    }
}

void ParseRegularGraph::organizeHostLeftNodes(const Production &curProduction,const map<Node *,Node *> &mCurMatchGraph,
                                              map<Node *,Node *> &mpHostToLeftDangling,set<Node *> &spReplacedHostNodes)
//curProduction: the current production; mCurMatchGraph: map between nodes of the host graph and that of right hand side
//mpHostToLeftDangling: map between to-be replaced dangling related nodes of the host graph and nodes on the left side
//spReplacedHostNodes: set of nodes of host graph, which will be replaced.
{
    map<Node *, Node *>::const_iterator mpNodesItor, mpDanglingNodesItor;
    Node *pCurrentRightNode, *pCurrentLeftNode, *pCurrentHostNode;
    
    for (mpNodesItor = mCurMatchGraph.begin(); mpNodesItor != mCurMatchGraph.end(); ++mpNodesItor)
    {
        pCurrentHostNode  = mpNodesItor->first;
        pCurrentRightNode = mpNodesItor->second;
        spReplacedHostNodes.insert(pCurrentHostNode);
        mpDanglingNodesItor  = curProduction.mRightToLeftDangling.find(pCurrentRightNode);
        if (mpDanglingNodesItor != curProduction.mRightToLeftDangling.end())
        {
            pCurrentLeftNode = mpDanglingNodesItor->second;
            mpHostToLeftDangling.insert(make_pair(pCurrentHostNode, pCurrentLeftNode));
        }
    }
}