//
//  ParseRegularGraph.h
//  egg
//
//  Created by Yingfeng Wang on 2/23/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__ParseRegularGraph__
#define __egg__ParseRegularGraph__

#include <stdio.h>
#include <set>

#include "GraphGrammar.h"
#include "RegularGraph.h"

using namespace std;

class ParseRegularGraph
{
    bool bPartialParse; // true: once an initial is reached, stop even if there are other nodes and edges.
    
    //replace subgraph of the regular graph
    void updateRegularGraph (const Production &curProduction, const RegularGraph &curRegularGraph,
                             const map<Node *, Node *> &mCurMatchGraph, RegularGraph &replacedRegularGraph, bool &bValid);
    //add nodes and edges of GML to the graph (this function is usually used to add elements on left side to new graph)
    static void addElesOnGMLToGraph(const GML &curGML, RegularGraph &updatedRegularGraph,int &iMaxNodeId,int &iMaxEdgeId);
    //add edges of GML to the graph (this function is usually used to add elements on left side to new graph)
    static void addEdgesOnGMLToGraph(const GML &curGML, RegularGraph &updatedRegularGraph, int &iMaxEdgeId);
    //add nodes of original graph to the new regular graph
    static void addNodesOfOriginalToNew(const Production &curProduction, const map <Node *, Node *> &mpHostToLeftDangling,
                                        const set<Node *> &spReplacedHostNodes, const GML &curGML,
                                        RegularGraph &updatedRegularGraph, int iMinNodeId);
    //add edges of original graph to the new regular graph
    static void addEdgesOfOriginalToNew(const Production &curProduction, const map <Node *, Node *> &mpHostToLeftDangling,
                                        const set<Node *> &spReplacedHostNodes, const GML &curGML,
                                        RegularGraph &updatedRegularGraph, int iMinNodeId, int iMinEdgeId, bool &bValid);
    //check whether the new edge connects two nodes that have been connected.
    static bool NewEdgeCheck(const set<pair<int, int> > &setEdgeNodesPairs, int &iSource, int &iTarget);

public:
    ParseRegularGraph(bool bPartialParse=false);
    
    //parse the given regular graph with given graph grammar
    bool parseGraph(const RegularGraph &curRegularGraph, const GraphGrammar &curGraphGrammar);
    //parse the given regular graph with given stochastic context free graph grammar (for metabolomics) 
    bool parseGraphSimple(const RegularGraph &curRegularGraph, const GraphGrammar &curGraphGrammar);
    //judge whether the regular graph is in the initial status
    static bool isInitialStatus(const RegularGraph &curRegularGraph, bool bPartialParse=false);
    //build a map between to-be replaced dangling related nodes of the host graph and nodes on the left hand side
    static void organizeHostLeftNodes(const Production &curProduction, const map<Node *, Node *> &mCurMatchGraph,
                                      map<Node *, Node *> &mpHostToLeftDangling, set<Node *> &spReplacedHostNodes);
    //add elements on the left nodes to the graph
    static void addElesOnLeftToGraph(const Production &curProduction, RegularGraph &updatedRegularGraph, int &iMaxNodeId,
                                     int &iMaxEdgeId);
    //add elements on the original graph to the new regular graph
    static void addOriginalGraphToNew(const Production &curProduction, const map <Node *, Node *> &mpHostToLeftDangling,
                                      const set<Node *> &spReplacedHostNodes, const RegularGraph &curRegularGraph,
                                      RegularGraph &updatedRegularGraph, int iMinNodeId, int iMinEdgeId, bool &bValid);
    //check whether the annotation is "initial"
    static bool isInitialAnnotation(const string &sLabel);


};


#endif /* defined(__egg__ParseRegularGraph__) */
