//
//  Production.cpp
//  egg
//
//  Created by Yingfeng Wang on 1/25/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#include "Production.h"

Production::Production(const Production &oriProduction)
//oriProduction: original production treated as source
{
    GML *pGML;
    size_t i, j;
    map<int, Node *> mNodeIdPointerLeft, mNodeIdPointerRight;
    Node *pNode;
    
    this->dProductionProb = oriProduction.dProductionProb;
    this->mProductionInfo = oriProduction.mProductionInfo;
    this->eDirected = oriProduction.eDirected;
    this->sId = oriProduction.sId;
    this->vpLeftHandSide.clear();
    for (i = 0; i < oriProduction.vpLeftHandSide.size(); i++)
    {
        pGML = new GML(*(oriProduction.vpLeftHandSide.at(i)));
        this->vpLeftHandSide.push_back(pGML);
        for (j = 0; j < this->vpLeftHandSide.at(i)->vpNodes.size(); j++)
        {
            pNode = this->vpLeftHandSide.at(i)->vpNodes.at(j);
            mNodeIdPointerLeft.insert(make_pair(pNode->iNodeId, pNode));
        }
    }
    this->vpRightHandSide.clear();
    for (i = 0; i < oriProduction.vpRightHandSide.size(); i++)
    {
        pGML = new GML(*(oriProduction.vpRightHandSide.at(i)));
        this->vpRightHandSide.push_back(pGML);
        for (j = 0; j < this->vpRightHandSide.at(i)->vpNodes.size(); j++)
        {
            pNode = this->vpRightHandSide.at(i)->vpNodes.at(j);
            mNodeIdPointerRight.insert(make_pair(pNode->iNodeId, pNode));
        }
    }
    this->vpContextGraph.clear();
    for (i = 0; i < oriProduction.vpContextGraph.size(); i++)
    {
        pGML = new GML(*(oriProduction.vpContextGraph.at(i)));
        this->vpContextGraph.push_back(pGML);
        for (j = 0; j < this->vpContextGraph.at(i)->vpNodes.size(); j++)
        {
            pNode = oriProduction.vpContextGraph.at(i)->vpNodes.at(j);
            mNodeIdPointerLeft.insert(make_pair(pNode->iNodeId, pNode));
            mNodeIdPointerRight.insert(make_pair(pNode->iNodeId, pNode));
        }
    }
    
    //mRightToLeftDangling
    copyMapPairNodes(oriProduction.mRightToLeftDangling, this->mRightToLeftDangling, mNodeIdPointerLeft,
                     mNodeIdPointerRight);

}


Production::Production()
{
    sId = "";
    dProductionProb = 1.0;
    eDirected = UNDIRECTED;
}

Production::~Production()
{
    size_t i;
    for (i=0; i<vpContextGraph.size(); i++)
    {
        delete vpContextGraph.at(i);
        vpContextGraph.at(i) = nullptr;
    }
    for (i=0; i<vpLeftHandSide.size(); i++)
    {
        delete vpLeftHandSide.at(i);
        vpLeftHandSide.at(i) = nullptr;
    }
    for (i=0; i<vpRightHandSide.size(); i++)
    {
        delete vpRightHandSide.at(i);
        vpRightHandSide.at(i) = nullptr;
    }
}

void Production::copyMapPairNodes(const map<Node *,Node *> &mSource, map<Node *,Node *> &mDest,
                                  const map<int, Node *> &mNodeIdPointerLeft, const map<int, Node *> &mNodeIdPointerRight)
//mSource: source of the right to left map;
//mDest: destination, right to left on new production;
//mNodeIdPointerLeft: map between node id and node pointer (left+context)
//mNodeIdPointerRight: map between node id and node pointer (right+context)
{
    map<Node *,Node *>::const_iterator itmpNode;
    map<int, Node *>::const_iterator itCurmpNodeOne, itCurmpNodeTwo;
    
    mDest.clear();
    for (itmpNode = mSource.begin(); itmpNode != mSource.end(); ++itmpNode)
    {
        itCurmpNodeOne = mNodeIdPointerRight.find(itmpNode->first->iNodeId);
        itCurmpNodeTwo = mNodeIdPointerLeft.find(itmpNode->second->iNodeId);
        if ((itCurmpNodeOne == mNodeIdPointerRight.end()) || (itCurmpNodeTwo == mNodeIdPointerLeft.end()))
        {
            cerr<<"fail to find node with id "<<itmpNode->first->iNodeId<<" or "<<itmpNode->second->iNodeId<<endl;
            exit(1);
        }else
        {
            mDest.insert(make_pair(itCurmpNodeOne->second, itCurmpNodeTwo->second));
        }
    }
}

Node * Production::SearchNode(const vector<Node *> vpCurrentNodes, int iSearchId, bool bMustFind /*=false*/)
//vpCurrentNodes: vector of pointers of nodes
//iSearchId: the node id we are searching
//bMustFind: if true and can't find the node, report error
//the pointer of Node we are searching, or nullptr if we can't find it.
{
    Node * pFoundNode = nullptr;
    vector<Node *>::const_iterator citvpNode;
    
    for (citvpNode = vpCurrentNodes.begin(); citvpNode != vpCurrentNodes.end(); ++citvpNode)
        if ((*citvpNode)->iNodeId == iSearchId)
        {
            pFoundNode = *citvpNode;
            break; // iSearchId is unique
        }
    if ((bMustFind) && (pFoundNode == nullptr))
    {
        cerr<<"can't find node "<<iSearchId<<endl;
        exit(1);
    }
    return pFoundNode;
}

Node * Production::SearchNode(const vector<GML *> vpCurrentSubgraphs, int iSearchId, bool bMustFind /*=false*/)
//vpCurrentSubgraphs: vector of pointers of subgraphs (GMLs)
//iSearchId: the node id we are searching
//bMustFind: if true and can't find the node, report error
//return the pointer of Node we are searching, or nullptr if we can't find it.
{
    Node * pFoundNode = nullptr;
    vector<GML *>::const_iterator citvpGML;
    
    for (citvpGML = vpCurrentSubgraphs.begin(); citvpGML != vpCurrentSubgraphs.end(); ++citvpGML)
    {
        pFoundNode = SearchNode((*citvpGML)->vpNodes, iSearchId);
        if (pFoundNode != nullptr)
            break;
    }
    if ((bMustFind) && (pFoundNode == nullptr))
    {
        cerr<<"can't find node "<<iSearchId<<endl;
        exit(1);
    }
    
    return pFoundNode;
}

void Production::AssociateEdgesAndNodes(GML * pCurrentSubGraph)
//pCurrentSubGraph: pointer of current subgraph
{
    vector<GML *> vpEmptyGraph;
    vpEmptyGraph.clear();
    AssociateEdgesAndNodes(pCurrentSubGraph, vpEmptyGraph);
}

void Production::AssociateEdgesAndNodes(GML * pCurrentSubGraph, const vector<GML *> vpCurContextGraph)
//pCurrentSubGraph: pointer of current subgraph
//vpCurContextGraph: context graph, in which no node or edge will be replaced in the production
{
    vector<Edge *>::const_iterator citvpEdge;
    int iSourceNodeId, iTargetNodeId, iUniqueSourceId, iUniqueTargetId;
    
    for (citvpEdge = pCurrentSubGraph->vpEdges.begin(); citvpEdge != pCurrentSubGraph->vpEdges.end(); ++citvpEdge)
    {
        iSourceNodeId = (*citvpEdge)->iSource;
        iTargetNodeId = (*citvpEdge)->iTarget;
        if (iSourceNodeId >= 0)
        {
            (*citvpEdge)->pSourceNode = SearchNode(pCurrentSubGraph->vpNodes, iSourceNodeId);
            if ((*citvpEdge)->pSourceNode == nullptr)
                (*citvpEdge)->pSourceNode = SearchNode(vpCurContextGraph, iSourceNodeId, true);
        }
        if (iTargetNodeId >= 0)
        {
            (*citvpEdge)->pTargetNode = SearchNode(pCurrentSubGraph->vpNodes, iTargetNodeId);
            if ((*citvpEdge)->pTargetNode == nullptr) 
                (*citvpEdge)->pTargetNode = SearchNode(vpCurContextGraph,iTargetNodeId, true);
        }
        if((((*citvpEdge)->pSourceNode == nullptr)&&(iSourceNodeId >= 0))||(((*citvpEdge)->pTargetNode == nullptr)&&(iTargetNodeId >= 0)))
        {
            cerr<<"Can't find node "<<iSourceNodeId<<" or "<<iTargetNodeId<<endl;
            exit(1);
        }
        iUniqueSourceId = (*citvpEdge)->getSourceNodeId(); // could be -1, -2, -3 ...
        iUniqueTargetId = (*citvpEdge)->getTargetNodeId(); // could be -1, -2, -3 ...
        if ((*citvpEdge)->getDirected())
        {// if edge is directed
            if (iSourceNodeId >= 0)
                (*citvpEdge)->pSourceNode->mpSourceEdges.insert(pair<int, Edge *>(iUniqueTargetId, (*citvpEdge)));
            if (iTargetNodeId >= 0)
                (*citvpEdge)->pTargetNode->mpTargetEdges.insert(pair<int, Edge *>(iUniqueSourceId, (*citvpEdge)));
        }else
        {// if edge is undirected
            if (iSourceNodeId >= 0)
                (*citvpEdge)->pSourceNode->mpUndirectedEdges.insert(pair<int, Edge *>(iUniqueTargetId, (*citvpEdge)));
            if (iTargetNodeId >= 0)
                (*citvpEdge)->pTargetNode->mpUndirectedEdges.insert(pair<int, Edge *>(iUniqueSourceId, (*citvpEdge)));
        }
    }
}

void Production::AssociateProductionEdgesAndNodes()
{
    size_t i;
    if (!(vpContextGraph.empty()))
    {
        for (i=0; i<vpContextGraph.size(); i++)
            AssociateEdgesAndNodes(vpContextGraph.at(i));
        if (!(vpLeftHandSide.empty()))
            for (i=0; i<vpLeftHandSide.size(); i++)
                AssociateEdgesAndNodes(vpLeftHandSide.at(i), vpContextGraph);
        if (!(vpRightHandSide.empty()))
            for (i=0; i<vpRightHandSide.size(); i++)
                AssociateEdgesAndNodes(vpRightHandSide.at(i), vpContextGraph);
    }else
    {
        if (!(vpLeftHandSide.empty()))
            for (i=0; i<vpLeftHandSide.size(); i++)
                AssociateEdgesAndNodes(vpLeftHandSide.at(i));
        if (!(vpRightHandSide.empty()))
            for (i=0; i<vpRightHandSide.size(); i++)
                AssociateEdgesAndNodes(vpRightHandSide.at(i));
    }
    buildDanglingNodesMapSides(vpLeftHandSide, vpRightHandSide, vpContextGraph, mRightToLeftDangling);
}

void Production::buildDanglingNodesMapSides(const vector<GML *> &vpLeft, const vector<GML *> &vpRight,
                                            const vector<GML *> &vpContext, map<Node *, Node *> &mRightLeft)
//vpLeft: subgraphs on the left side; vpRight: subgraphs on the right side; vpContext: context subgraphs
//mRightLeft:1tomany map beween dangling edges of nodes on right side (right+context) and that on left side (left+context)
{
    BigGML leftBigGML, rightBigGML;
    size_t i, iNodePos;
    int iCorrespondingNodeId;
    Node *pCurrentNode, *pLeftNode;
    map<string, string>::const_iterator mssIt;
    ReadGML curReadGML;
    
    mRightLeft.clear();
    GML::mergeGMLs(vpLeft, leftBigGML);
    GML::mergeGMLs(vpContext, leftBigGML);
    GML::mergeGMLs(vpRight, rightBigGML);
    GML::mergeGMLs(vpContext, rightBigGML);
    for (i=0; i<leftBigGML.vpStartDangling.size(); i++)
        mapPairNodes(leftBigGML.vpStartDangling.at(i), rightBigGML.vpStartDangling, mRightLeft, false);
    for (i=0; i<leftBigGML.vpEndDangling.size(); i++)
        mapPairNodes(leftBigGML.vpEndDangling.at(i), rightBigGML.vpEndDangling, mRightLeft, true);
    for (i=0; i<leftBigGML.vpUndirectedStartDangling.size(); i++)
        mapPairNodes(leftBigGML.vpUndirectedStartDangling.at(i), rightBigGML.vpUndirectedStartDangling,mRightLeft, false);
    for (i=0; i<leftBigGML.vpUndirectedEndDangling.size(); i++)
        mapPairNodes(leftBigGML.vpUndirectedEndDangling.at(i), rightBigGML.vpUndirectedEndDangling, mRightLeft, true);
    for (i=0; i<rightBigGML.vpNodes.size(); i++)
    {
        pCurrentNode = rightBigGML.vpNodes.at(i);
        mssIt = pCurrentNode->mNodeInfo.find("left");
        if (mssIt != pCurrentNode->mNodeInfo.end())
        {
            iCorrespondingNodeId = curReadGML.String2Number<int>(mssIt->second);
            iNodePos = leftBigGML.mNodeIdPos[iCorrespondingNodeId];
            pLeftNode = leftBigGML.vpNodes.at(iNodePos);
            mRightLeft.insert(make_pair(pCurrentNode, pLeftNode));
        }
    }
}

void Production::mapPairNodes(Edge * pLeftEdge, const vector<Edge *> &vpRightEdges,
                                             map<Node *, Node *> &mpNodesRightLeft, bool bSource)
//pLeftEdge: the pointer of an dangling edge on the left side;
//vpRightEdges: vector of pointers of related dangling edges on the right side
//bSource: true if the source node of dangling edge is known; false: if not
//mpNodesRightLeft: a map between nodes from right side to left side
{
    Node *pLeftNode, *pRightNode;
    size_t i;
    int iCurrentEdgeId;
    
    pLeftNode = bSource ? pLeftEdge->pSourceNode : pLeftEdge->pTargetNode;
    iCurrentEdgeId = pLeftEdge->iEdgeId;
    for (i=0; i<vpRightEdges.size(); i++)
        if (vpRightEdges.at(i)->iEdgeId == iCurrentEdgeId)
        {
            pRightNode = bSource ? vpRightEdges.at(i)->pSourceNode : vpRightEdges.at(i)->pTargetNode;
            if ((pLeftNode == nullptr) || (pRightNode == nullptr))
            {
                cerr<<"no node informaiton on edge "<<iCurrentEdgeId<<endl;
                exit(1);
            }
            mpNodesRightLeft.insert(make_pair(pRightNode, pLeftNode));
            break;
        }
    
    //return mpNodesRightLeft;
}

bool Production::IsInitial(bool bPartialParse)
//bPartialParse : true if partial parse mode, by default it is false
//return true if there is initial status is on the left hand side of the production
{
    bool bIsInitial = false;
    vector<GML *>::const_iterator itpGML;
    
    if (bPartialParse)
    {
        if (vpLeftHandSide.size() > 0)
            for(itpGML = vpLeftHandSide.begin(); itpGML != vpLeftHandSide.end(); ++itpGML)
            {
                if ((*itpGML)->IsInitial(bPartialParse))
                {
                    bIsInitial = true;
                    break;
                }
            }
    }else
    {
        if (vpLeftHandSide.size() == 1)
            if (vpLeftHandSide.at(0)->IsInitial(bPartialParse))
                bIsInitial = true;
    }
    
    return bIsInitial;
}


