//
//  Production.h
//  egg
//
//  Created by Yingfeng Wang on 1/25/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__Production__
#define __egg__Production__

#include <stdio.h>

#include <vector>
#include <map>
#include "ReadGML.h"
#include <iostream>
#include <fstream>

using namespace std;

class Production
{
    //map dangling edge related nodes on the right side to that on the left side
    void mapPairNodes(Edge * pLeftEdge, const vector<Edge *> &vpRightEdges,
                                     map<Node *, Node *> &mpNodesRightLeft, bool bSource);
    void copyMapPairNodes(const map<Node *,Node *> &mSource, map<Node *,Node *> &mDest,
                          const map<int, Node *> &mNodeIdPointerLeft, const map<int, Node *> &mNodeIdPointerRight);

public:
    //left hand side of the production; if left hand side and context are empty (get "[ ]"), the left side is initial graph
    vector <GML *> vpLeftHandSide;
    vector <GML *> vpRightHandSide; // right hand side of the production
    vector <GML *> vpContextGraph; //context graph, in which no node or edge will be replaced in the production
    double dProductionProb; // probability of the production. default value is -1.0, if not specified
    map <string, string> mProductionInfo; // other production information
    //directed status of the subgraph. eDirected of production has higher priority the priority > that of GML
    DirectionStatus eDirected;
    string sId; // productionID
    //map beween dangling edges of nodes on right side (right+context) and that on left side (left+context); one : many
    map<Node *, Node *> mRightToLeftDangling;
    
    Production(const Production &oriProduction);//copy Production
    Production(); //constructor
    ~Production(); //destructor
    void AssociateProductionEdgesAndNodes(); // associate edges and nodes of the production
    // associate edges and nodes of GML without considering context
    static void AssociateEdgesAndNodes(GML * pCurrentSubGraph);
    // associate edges and nodes of GML with considering context
    static void AssociateEdgesAndNodes(GML * pCurrentSubGraph, const vector<GML *> vpCurContextGraph);
    // search node with the given id, id is unique
    static Node * SearchNode(const vector<Node *> vpCurrentNodes, int iSearchId, bool bMustFind = false);
    // search node in a vector of subgraphs with the given id, id is unique
    static Node * SearchNode(const vector<GML *> vpCurrentSubgraphs, int iSearchId, bool bMustFind = false);
    //build 1tomany map beween dangling edges of nodes on right side (right+context) and that on left side (left+context)
    void buildDanglingNodesMapSides(const vector <GML *> &vpLeft, const vector <GML *> &vpRight,
                                    const vector <GML *> &vpContext, map<Node *, Node *> &mRightLeft);
    //check whether initial status is on the left hand side of the production
    bool IsInitial(bool bPartialParse = false);
};


#endif /* defined(__egg__Production__) */
