//
//  ReadGML.cpp
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#include "ReadGML.h"


ReadGML::ReadGML()
{
    //cout<<"Read GML"<<endl;
}

DirectionStatus ReadGML::ParseDirectionStatus(const string & sDirectionStauts)
//sDirectionStauts: string of direction status
// return parsed direction status as DirectionStatus type
{
    DirectionStatus eCurrentDirected;
    
    if ((sDirectionStauts == "Directed") || (sDirectionStauts == "directed") || (sDirectionStauts == "DIRECTED"))
        eCurrentDirected = DIRECTED;
    else if ((sDirectionStauts == "Hybrid") || (sDirectionStauts == "hybrid") || (sDirectionStauts == "HYBRID"))
        eCurrentDirected = HYBRID;
    else if ((sDirectionStauts == "Undirected") || (sDirectionStauts == "undirected") || (sDirectionStauts == "UNDIRECTED"))
        eCurrentDirected = UNDIRECTED;
    else
    {
        cerr<<sDirectionStauts<<" is not a valid direction status"<<endl;
        exit(1);
    }
    
    return eCurrentDirected;
}

bool ReadGML::ParseBool(const string & sBool) // read true or false from the input string
//sBool: string of boolean value
//turn Parsed boolean value
{
    bool bBoolValue;
    
    if ((sBool == "True") || (sBool == "true") || (sBool == "TRUE"))
        bBoolValue = true;
    else if ((sBool == "False") || (sBool == "false") || (sBool == "FALSE"))
        bBoolValue = false;
    else
    {
        cerr << "Expect true or false, but can't recognize it in " <<sBool<<endl;
        exit(1);
    }
    
    return bBoolValue;
}

Node * ReadGML::ParseNode (const string & sNode)
// sNode: string contains information of node, which is consisted of id and label
// return pointer of organized node as a Node object
{
    Node * pCurrentNode = new Node();
    vector < pair <string, string> > vpssCurrentNode;
    string sCurrentFeatureName, sCurrentFeatureValue;
    size_t i;
    
    vpssCurrentNode = ParseLayer(sNode);
    for (i=0; i<vpssCurrentNode.size(); i++)
    {
        sCurrentFeatureName  = vpssCurrentNode.at(i).first;
        sCurrentFeatureValue = vpssCurrentNode.at(i).second;
        if ((sCurrentFeatureName == "id")||(sCurrentFeatureName == "Id")||(sCurrentFeatureName == "ID"))
            pCurrentNode->iNodeId = String2Number<int>(sCurrentFeatureValue);
        else if ((sCurrentFeatureName == "label")||(sCurrentFeatureName == "Label")||(sCurrentFeatureName == "LABEL"))
            pCurrentNode->sLabel = sCurrentFeatureValue;
        else if ((sCurrentFeatureName == "flexible")||(sCurrentFeatureName == "Flexible")||(sCurrentFeatureName == "FLEXIBLE"))
            pCurrentNode->mNodeInfo.insert(pair<string, string> ("flexible", sCurrentFeatureValue));
        else if ((sCurrentFeatureName == "left")||(sCurrentFeatureName == "Left")||(sCurrentFeatureName == "LEFT"))
            pCurrentNode->mNodeInfo.insert(pair<string, string> ("left", sCurrentFeatureValue));
        else
            pCurrentNode->mNodeInfo.insert(pair<string, string> (sCurrentFeatureName, sCurrentFeatureValue));
    }
    if (pCurrentNode->iNodeId < 0)
    {
        cerr<<"no id information in "<<sNode<<endl;
        exit(1);
    }
    
    return pCurrentNode;
}

Edge * ReadGML::ParseEdge (const string & sEdge, DirectionStatus eSubgraphDirected)
// sEdge: string contains information of edge, which is consisted of source node id, target node id, label, and directed
// eSubgraphDirected: eGraphDirected: the directed status of the subgraph
// return pointer of organized edge as a Edge object
{
    Edge *pCurrentEdge = new Edge();
    vector < pair <string, string> > vpssCurrentEdge;
    string sCurrentFeatureName, sCurrentFeatureValue;
    size_t i;
    
    vpssCurrentEdge = ParseLayer(sEdge);
    for (i=0; i<vpssCurrentEdge.size(); i++)
    {
        sCurrentFeatureName  = vpssCurrentEdge.at(i).first;
        sCurrentFeatureValue = vpssCurrentEdge.at(i).second;
        if ((sCurrentFeatureName == "id")||(sCurrentFeatureName == "Id")||(sCurrentFeatureName == "ID"))
            pCurrentEdge->iEdgeId = String2Number<int>(sCurrentFeatureValue);
        else if ((sCurrentFeatureName == "source")||(sCurrentFeatureName == "Source")||(sCurrentFeatureName == "SOURCE"))
            pCurrentEdge->iSource = String2Number<int>(sCurrentFeatureValue);
        else if ((sCurrentFeatureName == "target")||(sCurrentFeatureName == "Target")||(sCurrentFeatureName == "TARGET"))
            pCurrentEdge->iTarget = String2Number<int>(sCurrentFeatureValue);
        else if ((sCurrentFeatureName == "label")||(sCurrentFeatureName == "Label")||(sCurrentFeatureName == "LABEL"))
            pCurrentEdge->sLabel = sCurrentFeatureValue;
        else if ((sCurrentFeatureName == "directed")||(sCurrentFeatureName == "Directed")
                 ||(sCurrentFeatureName == "DIRECTED"))
        {
            if (eSubgraphDirected == HYBRID)
            {
                if (ParseBool(sCurrentFeatureValue))
                    pCurrentEdge->setDirected(true);
                else
                    pCurrentEdge->setDirected(false);
            }
        }else
            pCurrentEdge->mEdgeInfo.insert(pair<string, string> (sCurrentFeatureName, sCurrentFeatureName));
    }
    if (eSubgraphDirected == UNDIRECTED)
        pCurrentEdge->setDirected(false);
    else if (eSubgraphDirected == DIRECTED)
        pCurrentEdge->setDirected(true);
    if ((pCurrentEdge->iEdgeId < 0) || ((pCurrentEdge->iSource < 0) && (pCurrentEdge->iTarget < 0)))
    {
        cerr<<"miss id information or both nodes' information "<<sEdge<<endl;
        exit(1);
    }
    return pCurrentEdge;
}

void ReadGML::ParseSubgraph (const string & sSubgraph, GML * pCurrentGML, DirectionStatus eGraphDirected)
//sSubgraph: string contains inforamtion of subgraph, which is consisted of nodes and edges. Each subgraph is a GML object
//pCurrentGML: pointer of organized subgraph as a GML object
//eGraphDirected: the directed status of the whole graph
{
    Node * pCurrentNode;
    vector < pair <string, string> > vpssCurSubgraph;
    string sCurrentFeatureName, sCurrentFeatureValue;
    vector <string> vsEdges;
    size_t i;
    
    vpssCurSubgraph = ParseLayer(sSubgraph);
    for (i=0; i<vpssCurSubgraph.size(); i++)
    {
        sCurrentFeatureName  = vpssCurSubgraph.at(i).first;
        sCurrentFeatureValue = vpssCurSubgraph.at(i).second;
        if ((sCurrentFeatureName == "node")||(sCurrentFeatureName == "Node")||(sCurrentFeatureName == "NODE"))
        {
            pCurrentNode = ParseNode (sCurrentFeatureValue);
            pCurrentGML->vpNodes.push_back(pCurrentNode);
            pCurrentGML->mNodeIdPos.insert(pair<size_t, size_t>(pCurrentNode->iNodeId, pCurrentGML->vpNodes.size()-1));
        }else if ((sCurrentFeatureName == "edge")||(sCurrentFeatureName == "Edge")||(sCurrentFeatureName == "EDGE"))
            vsEdges.push_back(sCurrentFeatureValue);
        else if ((sCurrentFeatureName == "directionstatus")||(sCurrentFeatureName == "DirectionStatus")
                  ||(sCurrentFeatureName == "DIRECTIONSTATUS")||(sCurrentFeatureName == "directionStatus"))
        {
            if (eGraphDirected == HYBRID)
                pCurrentGML->setDirected(ParseDirectionStatus(sCurrentFeatureValue));
        } else
            pCurrentGML->mGMLInfo.insert(pair<string, string> (sCurrentFeatureName, sCurrentFeatureName));
    }
    if (eGraphDirected == UNDIRECTED)
        pCurrentGML->setDirected(UNDIRECTED);
    else if (eGraphDirected == DIRECTED)
        pCurrentGML->setDirected(DIRECTED);
    for (i = 0; i < vsEdges.size(); i++)
        ProcessEdgeInfo(vsEdges.at(i), pCurrentGML->getDirected(), pCurrentGML);
    pCurrentGML->dataOptimization();
}

void ReadGML::ProcessEdgeInfo (const string & sEdge, DirectionStatus eSubgraphDirected, GML * pCurrentGML)
// sEdge: string contains information of edge, which is consisted of source node id, target node id, label, and directed
// eSubgraphDirected: eGraphDirected: the directed status of the subgraph
//pCurrentGML: pointer of organized subgraph as a GML object
{
    Edge * pCurrentEdge;
    
    pCurrentEdge = ParseEdge(sEdge, pCurrentGML->getDirected());
    pCurrentGML->vpEdges.push_back(pCurrentEdge);
    if ((pCurrentEdge->iSource < 0) && (pCurrentEdge->getDirected()))
        pCurrentGML->vpStartDangling.push_back(pCurrentEdge);
    else if ((pCurrentEdge->iTarget < 0) && (pCurrentEdge->getDirected()))
        pCurrentGML->vpEndDangling.push_back(pCurrentEdge);
    else if (pCurrentEdge->iSource < 0)
        pCurrentGML->vpUndirectedStartDangling.push_back(pCurrentEdge);
    else if (pCurrentEdge->iTarget < 0)
        pCurrentGML->vpUndirectedEndDangling.push_back(pCurrentEdge);
}

void ReadGML::ParseGraph(vector <GML *> & vpCurrentGraph, const string & sGraphInfo,
                         DirectionStatus eGraphDirected)
// vpCurrentGraph: vector of pointers of subgraphs, each subgraph is a GML. There should be no edge between two subgraphs
// sGraphInfo: string containing information of all related subgraphs
// eGraphDirected: the directed status of the whole graph;
{
    vector < pair <string, string> > vpssAllSubgraphs;
    string sCurFeatureName, sCurFeatureValue;
    size_t i;
    GML * pCurGML;
    
    vpssAllSubgraphs = ParseLayer (sGraphInfo);
    for (i=0; i<vpssAllSubgraphs.size(); i++)
    {
        sCurFeatureName = vpssAllSubgraphs.at(i).first;
        sCurFeatureValue = vpssAllSubgraphs.at(i).second;
        if ((sCurFeatureName != "subgraph")&&(sCurFeatureName != "Subgraph")&&(sCurFeatureName != "SUBGRAPH"))
        {
            cerr<<"expect feature name subgraph in "<<sGraphInfo<<endl;
            exit(1);
        }
        pCurGML = new GML();
        ParseSubgraph(sCurFeatureValue, pCurGML, eGraphDirected);
        vpCurrentGraph.push_back(pCurGML);
    }
}

size_t ReadGML::LocateEndFeatureEndPos(const string & sWholeLine, const size_t & beginPos)
//sWholeLine : the whole string of features
//beginPos : the beginning position of search
//return: the ending position of the first feature, whose value starts with beginPos
{
    size_t endPos, firstWhitePos;
    int iUnmatchedLeftBracketNum;
    if (sWholeLine.at(beginPos) == '"')
        // if meet '"', get the next '"'
        endPos = sWholeLine.find('"', beginPos+1);
    else if ( sWholeLine.at(beginPos) != '[' )
    {
        firstWhitePos = sWholeLine.find_first_of(WHITESPACES, beginPos);
        if (firstWhitePos == string::npos)
            endPos = sWholeLine.length() - 1;
        else
            endPos = firstWhitePos - 1;
    }
    else
    {
        endPos = beginPos + 1;
        iUnmatchedLeftBracketNum = 1;
        while (endPos < sWholeLine.length())
        {
            if (sWholeLine.at(endPos) == '[')
                iUnmatchedLeftBracketNum++;
            else if (sWholeLine.at(endPos) == ']')
                iUnmatchedLeftBracketNum--;
            if (iUnmatchedLeftBracketNum == 0)
                break;
            endPos++;
        }
    }
    
    if ((endPos == string::npos) || (endPos == sWholeLine.length()))
    {
        cerr << "begin with position " <<beginPos<< " : "<<endl<< sWholeLine <<endl<< "has some problem."<<endl;
        exit(1);
    }
    return endPos;
}

vector < pair <string, string> > ReadGML::ParseLayer (const string & sCurrentNestedInfo)
//a string like "a [ b c d [ ] e [ ] f g ] " or "[ b c d [ ] e [ ] f g ]"
//return <b, c>, <d, []>, <e, []>, <f, g>
{
    vector < pair <string, string> > vpssLayerInfo;
    string sNestedInfo, sFeatureName, sFeatureValue;
    size_t featureBeginPos = 0, sepBeginPos = 0, sepEndPos = 0, featureEndPos = 0;
    
    sNestedInfo = RemoveOutsideWrapper(sCurrentNestedInfo);
    RemoveBothEndsWhiteSpace(sNestedInfo);
    if (sNestedInfo != "")
    {
        do {
            sepBeginPos = sNestedInfo.find_first_of(WHITESPACES, featureBeginPos);
            if (sepBeginPos == string::npos)
            {
                cerr<<"begin with position "<<featureBeginPos<<" : "<<endl<<sNestedInfo<<endl<<"has some problem."<<endl;
                exit(1);
            }
            sFeatureName = sNestedInfo.substr(featureBeginPos, sepBeginPos - featureBeginPos);
            sepEndPos = sNestedInfo.find_first_not_of(WHITESPACES, sepBeginPos) - 1;
            if (sepEndPos == string::npos)
            {
                cerr << "begin with position " <<sepBeginPos<< " : "<<endl<< sNestedInfo <<endl<< "has some problem."<<endl;
                exit(1);
            }
            sFeatureName = sNestedInfo.substr(featureBeginPos, sepBeginPos - featureBeginPos);
            featureEndPos = LocateEndFeatureEndPos(sNestedInfo, sepEndPos + 1);
            sFeatureValue = sNestedInfo.substr(sepEndPos + 1, featureEndPos - sepEndPos);
            vpssLayerInfo.push_back( make_pair(sFeatureName, sFeatureValue) );
            //locate the beginning postion of next feature;
            featureBeginPos = sNestedInfo.find_first_not_of(WHITESPACES, featureEndPos + 1);
        }while(featureBeginPos != string::npos);
    }
    return vpssLayerInfo;
}

string ReadGML::RemoveOutsideWrapper(const string & sOriginalData)
//sOriginalData: input string with format "a [ content ]" or "[ content ]"
// return content;
{
    string sCurrentData = sOriginalData;
    size_t currentLength, beginPos, endPos;
    RemoveBothEndsWhiteSpace(sCurrentData);
    currentLength = sCurrentData.length();
    if (currentLength == 0)
    {
        cerr<< "empty string for rmoving outside [ ]"<<endl;
        exit(1);
    }
    if (sCurrentData.at(currentLength-1) != ']')
    {
        cerr<< sCurrentData << " :last character is not ]"<<endl;
        exit(1);
    }else
    {
        endPos = currentLength-1;
        beginPos = sCurrentData.find("[");
        if (beginPos == string::npos)
        {
            cerr << sCurrentData << " : doesn't have ]"<<endl;
            exit(1);
        }else if (sCurrentData.substr(0, beginPos).find("]") != string::npos)
        {
            cerr << sCurrentData << " : [] don't match"<<endl;
            exit(1);
        }else
            sCurrentData = sCurrentData.substr(beginPos+1, endPos - beginPos - 1);
    }
    
    return sCurrentData;
}

void ReadGML::RemoveBothEndsWhiteSpace(string &sLine)
//sLine: the sLine this function manipulates
{
    size_t whitespacePos, nonWhiteSpacePos;
    
    nonWhiteSpacePos = sLine.find_last_not_of(WHITESPACES);
    if (nonWhiteSpacePos != string::npos)
        sLine.erase(nonWhiteSpacePos+1);
    else
        sLine.clear();
    whitespacePos = sLine.find_first_not_of(WHITESPACES);
    if ((whitespacePos != string::npos) && (whitespacePos != 0))
        sLine.erase(0, whitespacePos);
}