//
//  ReadGML.h
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#ifndef __egg__ReadGML__
#define __egg__ReadGML__


#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
#include <sstream>

#include "GraphBasics.h"
#include "GML.h"

#define WHITESPACES " \t\f\v\n\r"

using namespace std;


class ReadGML
{
    string RemoveOutsideWrapper (const string & sOriginalData); // Remove the most outside "[" and "]" 
    size_t LocateEndFeatureEndPos(const string & sWholeLine, const size_t & beginPos); // loacte the end position of feature
    Node * ParseNode (const string & sNode); // parse node consisted of id and label
    // parse edge consisted of source, target, label, and directed
    Edge * ParseEdge (const string & sEdge, DirectionStatus eSubgraphDirected);
    // process edge information
    void ProcessEdgeInfo (const string & sEdge, DirectionStatus eSubgraphDirected, GML * pCurrentGML);

public:
    ReadGML();
    void RemoveBothEndsWhiteSpace(string & sLine); //remove white spaces on both ends of a string
    vector < pair <string, string> > ParseLayer (const string & sCurrentNestedInfo);//parse one layer with nested GML format
    // parse subgraph consisted of nodes and edges. Each subgraph is a GML
    void ParseSubgraph (const string & sSubgraph, GML * pCurrentGML, DirectionStatus eGraphDirected);
    // parse graph consisted of subgraphes. Each subgraph is a GML
    void ParseGraph (vector <GML *> & vpCurrentGraph, const string & sGraphInfo, DirectionStatus eGraphDirected);
    DirectionStatus ParseDirectionStatus (const string & sDirectionStauts); //Parse direction status from its string format
    bool ParseBool(const string & sBool); // read true or false from the input string
    
    // transfer string to Number (e.g. int and double)
    template <typename NumberType>
    NumberType String2Number(const string & sInputStr)
    //sInputStr: the input string
    //return: the tansferred value, e.g. int, float, or double
    {
        NumberType transferredValue;
        istringstream istrTransfer;
        
        istrTransfer.clear();
        istrTransfer.str(sInputStr);
        istrTransfer >> transferredValue;
        istrTransfer.clear();
        return transferredValue;
    }
};

#endif /* defined(__egg__ReadGML__) */
