//
//  ReadGrammar.cpp
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#include "ReadGrammar.h"

ReadGrammar::ReadGrammar()
{
    //cout<<"Read Grammar"<<endl;
}

void ReadGrammar::ExtractProductionInfo(Production * pCurProduction,
                                        const vector< pair<string, string> > & vpssProductionContent)
// vpssProductionContent: store information of production
// pCurProduction: pointer of production store extracted information from vpssProductionContent
{
    string sCurrentFeatureName, sCurrentFeatureValue, sContextValue = "", sLeftHandSideValue, sRightHandSideValue;
    size_t i;
    for (i=0; i<vpssProductionContent.size(); i++)
    {
        sCurrentFeatureName = vpssProductionContent.at(i).first;
        sCurrentFeatureValue = vpssProductionContent.at(i).second;
        if (sCurrentFeatureName == "productionID")
            pCurProduction->sId = sCurrentFeatureValue; // '"'and '"' are not removed
        else if ((sCurrentFeatureName == "directionstatus")||(sCurrentFeatureName == "DirectionStatus")
                 ||(sCurrentFeatureName == "DIRECTIONSTATUS")||(sCurrentFeatureName == "directionStatus"))
            pCurProduction->eDirected = ParseDirectionStatus(sCurrentFeatureValue);
        else if (sCurrentFeatureName == "productionProbability")
            pCurProduction->dProductionProb = String2Number<double>(sCurrentFeatureValue) ;
        else if ((sCurrentFeatureName == "context")||(sCurrentFeatureName == "Context")||(sCurrentFeatureName == "CONTEXT"))
            sContextValue = sCurrentFeatureValue;
        else if ((sCurrentFeatureName == "left")||(sCurrentFeatureName == "Left")||(sCurrentFeatureName == "LEFT"))
            sLeftHandSideValue = sCurrentFeatureValue;
        else if ((sCurrentFeatureName == "right")||(sCurrentFeatureName == "Right")||(sCurrentFeatureName == "RIGHT"))
            sRightHandSideValue = sCurrentFeatureValue;
        else
            pCurProduction->mProductionInfo.insert(pair<string, string> (sCurrentFeatureName, sCurrentFeatureName));
    }
    if (sContextValue != "")
        ParseGraph (pCurProduction->vpContextGraph, sContextValue, pCurProduction->eDirected);
    if ((sLeftHandSideValue == "") || (sRightHandSideValue == ""))
    {
        cerr << "Miss left hand side or right hand side" <<endl;
        exit(1);
    }else
    {
        ParseGraph (pCurProduction->vpLeftHandSide, sLeftHandSideValue, pCurProduction->eDirected);
        ParseGraph (pCurProduction->vpRightHandSide, sRightHandSideValue, pCurProduction->eDirected);
    }
}

void ReadGrammar::ParseProduction(vector<string> & vsProduction, GraphGrammar & currentGraphGrammar)
//vsProduction: vector of strings of the production
//currentGraphGrammar: graph grammar contained in the file
{
    Production * pCurrentProduction = new Production();
    string sProduction;
    vector < pair <string, string> > vpssProductionContent;
    size_t i;
    
    sProduction = vsProduction.at(0);
    for (i = 1; i < vsProduction.size(); i++)
        sProduction += " " + vsProduction.at(i);
    vpssProductionContent = ParseLayer(sProduction);
    ExtractProductionInfo(pCurrentProduction, vpssProductionContent);
    pCurrentProduction->AssociateProductionEdgesAndNodes();
    currentGraphGrammar.vpProductions.push_back(pCurrentProduction);
    vsProduction.clear();
}

void ReadGrammar::ParseGrammar(ifstream & currentGrammarFile_stream, GraphGrammar & currentGraphGrammar)
//currentGrammarFile_stream: file stream of the graph grammar file
//currentGraphGrammar: graph grammar contained in the file
{
    string sCurrentLine;
    vector <string> vsCurrentProduction;
    size_t poundPos;
    
    currentGraphGrammar.vpProductions.clear();
    vsCurrentProduction.clear();
    
    while (!currentGrammarFile_stream.eof())
    {
        sCurrentLine.clear();
        getline(currentGrammarFile_stream, sCurrentLine);
        poundPos = sCurrentLine.find("#");
        if (poundPos != string::npos)
            sCurrentLine.erase(poundPos);
        RemoveBothEndsWhiteSpace(sCurrentLine);
        if (sCurrentLine.length() > 0)
        {
            if ((sCurrentLine.substr(0, 10) == "production") && (sCurrentLine.find_first_of(WHITESPACES) == 10))
                if (vsCurrentProduction.size() > 0)
                    ParseProduction(vsCurrentProduction, currentGraphGrammar);
            vsCurrentProduction.push_back(sCurrentLine);
        }
    }
    
    if (vsCurrentProduction.size() > 0)
        ParseProduction(vsCurrentProduction, currentGraphGrammar);
}

void ReadGrammar::ParseGrammarFile(string sGrammarFilePath, GraphGrammar & graphGrammar)
//sGrammarFilePath: the path of the file containing the graph grammar
//graphGrammar: the graph grammar
{
    ifstream grammarFile_stream;
    bool bGrammarFileExist;
    
    grammarFile_stream.open(sGrammarFilePath.c_str(), ifstream::in);
    bGrammarFileExist = grammarFile_stream.is_open();
    if (bGrammarFileExist)
    {
        ParseGrammar(grammarFile_stream, graphGrammar);
    }else
    {
        cerr<<"Can't open file "<<sGrammarFilePath<<" !"<<endl;
        exit(1);
    }
    grammarFile_stream.clear();
    grammarFile_stream.close();
}





