//
//  ReadInputGraphs.cpp
//  egg
//
//  Created by Yingfeng Wang on 1/24/15.
//  Copyright (c) 2015 Yingfeng Wang. All rights reserved.
//

#include "ReadInputGraphs.h"

void ReadInputGraphs::ExtractRegularGraphInfo(RegularGraph *pRegG, const string & sGraph,
                                              const vector< pair<string, string> > & vpssRegularGraphContent)
//pRegG: pointer of regular graph storing extracted information from vpssProductionContent
//sGraph: original string containing graph information
//vpssRegularGraphContent: store information of regular graph
{
    //ParseGraph (pRegG->vpSubgraphs, sGraph, HYBRID);
    string sCurrentFeatureName, sCurrentFeatureValue;
    GML * pCurGML;
    DirectionStatus eGraphDirected = HYBRID;
    size_t i;
    
    for (i=0; i<vpssRegularGraphContent.size(); i++)
    {
        sCurrentFeatureName  = vpssRegularGraphContent.at(i).first;
        sCurrentFeatureValue = vpssRegularGraphContent.at(i).second;
        if (sCurrentFeatureName == "graphID")
            pRegG->sGraphId = sCurrentFeatureValue;
        else if ((sCurrentFeatureName == "directionstatus")||(sCurrentFeatureName == "DirectionStatus")
                 ||(sCurrentFeatureName == "DIRECTIONSTATUS")||(sCurrentFeatureName == "directionStatus"))
            eGraphDirected = ParseDirectionStatus(sCurrentFeatureValue);
        else if ((sCurrentFeatureName=="subgraph")||(sCurrentFeatureName=="Subgraph")||(sCurrentFeatureName=="SUBGRAPH"))
        {
            pCurGML = new GML();
            ParseSubgraph(sCurrentFeatureValue, pCurGML, eGraphDirected);
            pRegG->vpSubgraphs.push_back(pCurGML);
        }else
            pRegG->mRegularGraphInfo.insert(pair<string,string>(sCurrentFeatureName, sCurrentFeatureValue));
    }
}

void ReadInputGraphs::ParseCurrentRegularGraph(vector<string> &vsCurrentSubgraphs, vector<RegularGraph *> &vpRegG)
//vsCurrentSubgraphs: vector of strings representing subgraphs
//vpRegG: vector of regular graphs in the current file
{
    RegularGraph * pRegG = new RegularGraph();
    string sGraph;
    vector< pair <string, string> > vpssRegularGraphContent;
    size_t i;
    
    sGraph = vsCurrentSubgraphs.at(0);
    for (i=1; i< vsCurrentSubgraphs.size(); i++)
        sGraph += " " + vsCurrentSubgraphs.at(i);
    vpssRegularGraphContent = ParseLayer(sGraph);
    ExtractRegularGraphInfo(pRegG, sGraph, vpssRegularGraphContent);
    for (i=0; i< pRegG->vpSubgraphs.size(); i++)
        Production::AssociateEdgesAndNodes(pRegG->vpSubgraphs.at(i));
    vpRegG.push_back(pRegG);
    vsCurrentSubgraphs.clear();
}

void ReadInputGraphs::ParseRegularGraphs(ifstream &currentGraphFile_stream, vector<RegularGraph *> &vpRegG)
//currentGraphFile_stream: file stream of the graph file
//vpRegG: vector of regular graphs in the current file
{
    string sCurrentLine;
    vector <string> vsCurrentSubgraphs;
    size_t poundPos;
    
    vpRegG.clear();
    vsCurrentSubgraphs.clear();
    
    while (!currentGraphFile_stream.eof())
    {
        sCurrentLine.clear();
        getline(currentGraphFile_stream, sCurrentLine);
        poundPos = sCurrentLine.find("#");
        if (poundPos != string::npos)
            sCurrentLine.erase(poundPos);
        RemoveBothEndsWhiteSpace(sCurrentLine);
        if (sCurrentLine.length() > 0)
        {
            if ((sCurrentLine.substr(0, 5) == "graph") && (sCurrentLine.find_first_of(WHITESPACES) == 5))
                if (vsCurrentSubgraphs.size() > 0)
                    ParseCurrentRegularGraph(vsCurrentSubgraphs, vpRegG);
            vsCurrentSubgraphs.push_back(sCurrentLine);
        }
    }
    
    if (vsCurrentSubgraphs.size() > 0)
        ParseCurrentRegularGraph(vsCurrentSubgraphs, vpRegG);
}

void ReadInputGraphs::ParseRegularGraphFile(const string & sCurrentRegGraphInputFilePath, vector<RegularGraph *> & vpRegG)
//sCurrentRegGraphInputFilePath: path of current file containing regular graphs
//vpRegG: vector of regular graphs in the current file
{
    ifstream graphFile_stream;
    bool bGraphFileExist;
    
    graphFile_stream.open(sCurrentRegGraphInputFilePath.c_str(), ifstream::in);
    bGraphFileExist = graphFile_stream.is_open();
    if (bGraphFileExist)
    {
        ParseRegularGraphs(graphFile_stream, vpRegG);
    }else
    {
        cerr<<"Can't open file "<<sCurrentRegGraphInputFilePath<<" !"<<endl;
        exit(1);
    }
    graphFile_stream.clear();
    graphFile_stream.close();
    
}

void ReadInputGraphs::ParseAllInputGraphFiles(const vector<string> & vsGraphFilePaths,
                                              vector< vector < RegularGraph *> > & vvpRegGraphs)
//vsGraphFilePaths: vector of graph file paths;
//vvpRegGraphs: all input graphs
{
    size_t i;
    vector< RegularGraph * > vpRegGraphs;
    
    vvpRegGraphs.clear();
    for (i=0; i<vsGraphFilePaths.size(); i++)
    {
        vpRegGraphs.clear();
        ParseRegularGraphFile(vsGraphFilePaths.at(i), vpRegGraphs);
        vvpRegGraphs.push_back(vpRegGraphs);
    }
}