//
//  main.cpp
//  egg
//
//  Created by Yingfeng Wang on 12/26/14.
//  Copyright (c) 2014 Yingfeng Wang. All rights reserved.
//

#include <iostream>

#include "EggCore.h"

using namespace std;

void ParseArguments(int argc, const char * argv[], map<string, string> & mArguments)
{
    vector<string> vsArguments;
    size_t i;

    while(argc--)
        vsArguments.push_back(*argv++);
    for (i = 0; i < vsArguments.size(); i++)
    {
        if ((vsArguments[i] == "-g") || (vsArguments[i] == "--grammar"))
            mArguments.insert(pair<string, string>("grammar file path", vsArguments[++i]));
        else if ((vsArguments[i] == "-w") || (vsArguments[i] == "--working_directory"))
            mArguments.insert(pair<string, string>("working directory", vsArguments[++i]));
        else if ((vsArguments[i] == "-f") || (vsArguments[i] == "--input_graph"))
            mArguments.insert(pair<string, string>("input graph", vsArguments[++i]));
        else if ((vsArguments[i] == "-i") || (vsArguments[i] == "--input_graph_folder"))
            mArguments.insert(pair<string, string>("input graph folder", vsArguments[++i]));
        else if ((vsArguments[i] == "-o") || (vsArguments[i] == "--output_directory"))
            mArguments.insert(pair<string, string>("output directory", vsArguments[++i]));
        else if ((vsArguments[i] == "-p") || (vsArguments[i] == "--partial_parse"))
            mArguments.insert(pair<string, string>("partial parse", vsArguments[++i]));
        else if ((vsArguments[i] == "-h") || (vsArguments[i] == "--help"))
        {
            cout<<"Usage: -g grammar_file_path -w working_directory -p is_partial_parse(optional)"<<endl;
            cout<<"-f input_graph_file_path -i path_of_folder_of_input_graphs -o output_directory"<<endl;
            cout<<"If working director is not specified, the current directory will be used as working directory."<<endl;
            cout<<"If neither -f nor -i is specified, all .GML files in the working directory are input graphs."<<endl;
            cout<<"If both of -f and -i are specified, -i will be ignored."<<endl;
            cout<<"If output directory is not specified, the working directory is the output directory."<<endl;
            cout<<"the current version doesn't search subdirectory"<<endl;
            cout<<"is_partial_parse should be true or false, while, by default, it is false"<<endl;
            exit(0);
        }
    }
}

int main(int argc, const char * argv[]) {
    map<string, string> mArgumentsInfo;
    ParseArguments(argc, argv, mArgumentsInfo);
    bool bParsed;
    size_t i, j;
    
    EggCore ec(mArgumentsInfo);
    vector< vector< RegularGraph *> > vvpInputRegGraphs;
    
    cout << "Hello, this is EGG project" << endl;
    
    //ec.vsInputGraphFileNamesList can be replace for other file paths list,
    //which has potential for multi-threading (e.g. OpenMP) and multi-processing (e.g. MPI)
    
    //ec.LoadInputGraphFiles(ec.vsInputGraphFileNamesList);
    ec.LoadInputGraphFiles(ec.vsInputGraphFileNamesList, vvpInputRegGraphs);
    ec.ReadGraphGrammar();
    
    for (i=0; i<vvpInputRegGraphs.size(); i++)
        for (j=0; j<vvpInputRegGraphs.at(i).size(); j++)
        {
            cout<<"parsing "<<i<<" "<<j<<" regular graph"<<endl;
            bParsed = ec.ParseRegGraph(vvpInputRegGraphs.at(i).at(j));
            if (bParsed)
                cout<<"input graph "<<i<<" "<<j<<" is accepted."<<endl;
            else
                cout<<"input graph "<<i<<" "<<j<<" is NOT accepted."<<endl;
        }
    return 0;
}
