//
//  Database.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/18/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include "Database.h"

Database::Database(const SettingInfo * pSettingInfo)
//pSettingInfo: pointer to setting information
{
    this->pSettingInfo = pSettingInfo;
}

Database::~Database()
{
    clear();
}

void Database::clear()
{
    vector<dbRecordType>::iterator itDBRecord;
    for (itDBRecord = vCompoundDB.begin(); itDBRecord != vCompoundDB.end(); ++itDBRecord)
    {
        delete itDBRecord->pCompound->pRWMol;
        delete itDBRecord->pCompound;
    }
    vector<dbRecordType>().swap(vCompoundDB);
}

void Database::readTopologies()
{
    int i, iDBsize;
    
    iDBsize = (int) vCompoundDB.size();
    //openmp chance
    for (i = 0; i < iDBsize; i++)
    {
        readOneTopology(i);
    }
}

void Database::readOneTopology(int iEntryID)
//iEntryID: the index of database entry;
{
    GML* pCurGraph;

    //pCurGraph = MolGraph2GML(vCompoundDB.at(iEntryID).pCompound->molTopology);
    pCurGraph = ROMol2GML(vCompoundDB.at(iEntryID).pCompound->pRWMol);
    
    vCompoundDB.at(iEntryID).pCompound->rgTopology.vpSubgraphs.push_back(pCurGraph);
    //cout<<vCompoundDB.at(iEntryID).pCompound->rgTopology.vpSubgraphs.size()<<endl;
}

string Database::BondType2Str(Bond::BondType bondtype)
{
    string sBondType;
    
    switch (bondtype)
    {
        case (Bond::BondType::UNSPECIFIED) : sBondType = "UNSPECIFIED"; break;
        case (Bond::BondType::SINGLE) : sBondType = "SINGLE"; break;
        case (Bond::BondType::DOUBLE) : sBondType = "DOUBLE"; break;
        case (Bond::BondType::TRIPLE) : sBondType = "TRIPLE"; break;
        case (Bond::BondType::QUADRUPLE) : sBondType = "QUADRUPLE"; break;
        case (Bond::BondType::QUINTUPLE) : sBondType = "QUINTUPLE"; break;
        case (Bond::BondType::HEXTUPLE) : sBondType = "HEXTUPLE"; break;
        case (Bond::BondType::ONEANDAHALF) : sBondType = "ONEANDAHALF"; break;
        case (Bond::BondType::TWOANDAHALF) : sBondType = "TWOANDAHALF"; break;
        case (Bond::BondType::THREEANDAHALF) : sBondType = "THREEANDAHALF"; break;
        case (Bond::BondType::FOURANDAHALF) : sBondType = "FOURANDAHALF"; break;
        case (Bond::BondType::FIVEANDAHALF) : sBondType = "FIVEANDAHALF"; break;
        case (Bond::BondType::AROMATIC) : sBondType = "AROMATIC"; break;
        case (Bond::BondType::IONIC) : sBondType = "IONIC"; break;
        case (Bond::BondType::HYDROGEN) : sBondType = "HYDROGEN"; break;
        case (Bond::BondType::THREECENTER) : sBondType = "THREECENTER"; break;
        case (Bond::BondType::DATIVEONE) : sBondType = "DATIVEONE"; break;
        case (Bond::BondType::DATIVE) : sBondType = "DATIVE"; break;
        case (Bond::BondType::DATIVEL) : sBondType = "DATIVEL"; break;
        case (Bond::BondType::DATIVER) : sBondType = "DATIVER"; break;
        case (Bond::BondType::OTHER) : sBondType = "OTHER"; break;
        case (Bond::BondType::ZERO) : sBondType = "ZERO"; break;
    }
    
    return sBondType;
}
/*
GML* Database::MolGraph2GML(const MolGraph &curMolGraph)
//curMolGraph: input MolGraph object; return pointer to a transferred GML object
{
    GML* pCurGML;
    Node* pCurNode;
    Edge* pCurEdge;
    string sEdgeLabel;
    int i;
    list<boost::list_edge<unsigned long, boost::shared_ptr<Bond> > >::const_iterator itEdge;
    
    pCurGML = new GML();
    pCurGML->setDirected(UNDIRECTED);
    for (i = 0; i < (int)curMolGraph.m_vertices.size(); i++)
    {
        pCurNode = new Node();
        pCurNode->iNodeId = curMolGraph.m_vertices.at(i).m_property->getIdx();
        pCurNode->sLabel = curMolGraph.m_vertices.at(i).m_property->getSymbol();
        pCurGML->mNodeIdPos.insert(pair<int, size_t>(pCurNode->iNodeId, pCurGML->vpNodes.size()));
        pCurGML->vpNodes.push_back(pCurNode);
    }
    for (itEdge = curMolGraph.m_edges.begin(); itEdge != curMolGraph.m_edges.end(); ++itEdge)
    {
        pCurEdge = new Edge();
        pCurEdge->setDirected(false);
        sEdgeLabel = "";
        pCurEdge->iEdgeId = itEdge->m_property->getIdx();
        sEdgeLabel += ((queryIsBondInRing(itEdge->m_property.get())) ? "ring|" : "linear|" );
        sEdgeLabel += ((itEdge->m_property->getIsAromatic()) ? "t|" : "f|");
        sEdgeLabel += ((itEdge->m_property->getIsConjugated()) ? "t|" : "f|");
        sEdgeLabel += BondType2Str(itEdge->m_property->getBondType());
        pCurEdge->sLabel = sEdgeLabel;
        pCurEdge->iSource = itEdge->m_property->getBeginAtomIdx();
        pCurEdge->iTarget = itEdge->m_property->getEndAtomIdx();
        pCurEdge->pSourceNode = pCurGML->vpNodes.at(pCurGML->mNodeIdPos.find(pCurEdge->iSource)->second);
        pCurEdge->pTargetNode = pCurGML->vpNodes.at(pCurGML->mNodeIdPos.find(pCurEdge->iTarget)->second);
        pCurEdge->pSourceNode->mpUndirectedEdges.insert(pair<int, Edge*>(pCurEdge->iTarget, pCurEdge));
        pCurEdge->pTargetNode->mpUndirectedEdges.insert(pair<int, Edge*>(pCurEdge->iSource, pCurEdge));
        pCurGML->vpEdges.push_back(pCurEdge);
    }
    
    return pCurGML;
}*/

GML* Database::ROMol2GML(const ROMol *pCurrentROMol)
//pCurrentROMol: pointer to the current ROMol; return pointer to a transferred GML object
{
    GML* pCurGML;
    Node* pCurNode;
    Edge* pCurEdge;
    string sEdgeLabel;
    ROMol::EDGE_ITER firstB,lastB;
    ROMol::VERTEX_ITER atBegin,atEnd;
    
    pCurGML = new GML();
    pCurGML->setDirected(UNDIRECTED);
    boost::tie(atBegin,atEnd) = pCurrentROMol->getVertices();
    while(atBegin!=atEnd)
    {
        boost::shared_ptr<Atom> atom = (*pCurrentROMol)[*atBegin];
        pCurNode = new Node();
        pCurNode->iNodeId = atom->getIdx();
        pCurNode->sLabel = atom->getSymbol();
        pCurGML->mNodeIdPos.insert(pair<int, size_t>(pCurNode->iNodeId, pCurGML->vpNodes.size()));
        pCurGML->vpNodes.push_back(pCurNode);
        ++atBegin;
    }
    boost::tie(firstB,lastB) = pCurrentROMol->getEdges();
    while(firstB!=lastB)
    {
        boost::shared_ptr<Bond> bond = (*pCurrentROMol)[*firstB];
        pCurEdge = new Edge();
        pCurEdge->setDirected(false);
        sEdgeLabel = "";
        pCurEdge->iEdgeId = bond->getIdx();
        sEdgeLabel += ((queryIsBondInRing(bond.get())) ? "ring|" : "linear|" );
        sEdgeLabel += ((bond->getIsAromatic()) ? "t|" : "f|");
        sEdgeLabel += ((bond->getIsConjugated()) ? "t|" : "f|");
        sEdgeLabel += BondType2Str(bond->getBondType());
        pCurEdge->sLabel = sEdgeLabel;
        pCurEdge->iSource = bond->getBeginAtomIdx();
        pCurEdge->iTarget = bond->getEndAtomIdx();
        pCurEdge->pSourceNode = pCurGML->vpNodes.at(pCurGML->mNodeIdPos.find(pCurEdge->iSource)->second);
        pCurEdge->pTargetNode = pCurGML->vpNodes.at(pCurGML->mNodeIdPos.find(pCurEdge->iTarget)->second);
        pCurEdge->pSourceNode->mpUndirectedEdges.insert(pair<int, Edge*>(pCurEdge->iTarget, pCurEdge));
        pCurEdge->pTargetNode->mpUndirectedEdges.insert(pair<int, Edge*>(pCurEdge->iSource, pCurEdge));
        pCurGML->vpEdges.push_back(pCurEdge);
        ++firstB;
    }
    return pCurGML;
}

void Database::readOfficialDatabaseFile(ifstream &dbfile_stream, string & sFirstLine)
//dbfile_stream: stream of databasefile; //sFirstLine: the first line of database
{
    string sCurrentLine;
    dbRecordType currentDBRecord;
    vector<string> vsItems;
    vector<string> vsValues;
    size_t i, item_length;
    ExtraInchiReturnValues currentExtraInchiReturnValues;
    vector<int> viFragmentMapping;
    int iFragmentNum;
    
    boost::algorithm::to_lower(sFirstLine);
    boost::split(vsItems, sFirstLine, boost::is_any_of("\t"), boost::token_compress_on);
    item_length = vsItems.size();
    while (!dbfile_stream.eof())
    {
        sCurrentLine.clear();
        getline(dbfile_stream, sCurrentLine);
        boost::algorithm::trim(sCurrentLine);
        if (sCurrentLine.length() > 0)
        {
            vsValues.clear();
            currentExtraInchiReturnValues = {};
            iFragmentNum = 0;
            currentDBRecord.dMass = 0;
            currentDBRecord.pCompound = nullptr;
            currentDBRecord.pCompound = new compoundType;
            boost::split(vsValues, sCurrentLine, boost::is_any_of("\t"), boost::token_compress_on);
            for (i = 0; i < item_length; i++)
            {
                if (vsItems.at(i) == "identifier")
                    currentDBRecord.pCompound->sIdentifier = vsValues.at(i);
                if (vsItems.at(i) == "name")
                    currentDBRecord.pCompound->sName = vsValues.at(i);
                if (vsItems.at(i) == "inchi")
                {
                    currentDBRecord.pCompound->sInchi = vsValues.at(i);
                    currentDBRecord.pCompound->pRWMol = InchiToMol(vsValues.at(i), currentExtraInchiReturnValues);
                    iFragmentNum = (int)(MolOps::getMolFrags(*(currentDBRecord.pCompound->pRWMol), viFragmentMapping));
                    currentDBRecord.pCompound->molTopology = currentDBRecord.pCompound->pRWMol->getTopology();
                }
                if (vsItems.at(i) == "links")
                    currentDBRecord.pCompound->sLinks = vsValues.at(i);
            }
            if (currentDBRecord.pCompound->dMass == 0)
                currentDBRecord.pCompound->dMass = Descriptors::calcExactMW(*(currentDBRecord.pCompound->pRWMol));
            if (iFragmentNum == 1)
            {
                currentDBRecord.dMass = currentDBRecord.pCompound->dMass;
                vCompoundDB.push_back(currentDBRecord);
            } else
                delete currentDBRecord.pCompound;
        }
    }
}

void Database::readInternalDatabaseFile(ifstream &dbfile_stream, string &sFirstLine)
{
    string sCurrentLine;
    dbRecordType currentDBRecord;
    vector<string> vsValues;
    ExtraInchiReturnValues currentExtraInchiReturnValues;
    vector<int> viFragmentMapping;
    int iFragmentNum;
    
    sCurrentLine = sFirstLine;
    do
    {
        boost::algorithm::trim(sCurrentLine);
        if (sCurrentLine.length() > 0)
        {
            vsValues.clear();
            currentExtraInchiReturnValues = {};
            iFragmentNum = 0;
            currentDBRecord.dMass = 0;
            currentDBRecord.pCompound = nullptr;
            currentDBRecord.pCompound = new compoundType;
            boost::split(vsValues, sCurrentLine, boost::is_any_of("\t"), boost::token_compress_on);
            if (vsValues.size() == 5)
            {
                currentDBRecord.pCompound->sIdentifier = vsValues.at(0);
                currentDBRecord.pCompound->sInchi = vsValues.at(1);
                currentDBRecord.pCompound->dMass = boost::lexical_cast<double>(vsValues.at(2));
                currentDBRecord.pCompound->sName = vsValues.at(3);
                currentDBRecord.pCompound->sLinks = vsValues.at(4);
                currentDBRecord.pCompound->pRWMol = InchiToMol(vsValues.at(1), currentExtraInchiReturnValues);
                viFragmentMapping.clear();
                iFragmentNum = (int)(MolOps::getMolFrags(*(currentDBRecord.pCompound->pRWMol), viFragmentMapping));
                currentDBRecord.pCompound->molTopology = currentDBRecord.pCompound->pRWMol->getTopology();
            }else
            {
                cerr<<"wrong database format, internal format requires 5 columns"<<endl;
                exit(1);
            }
            if (iFragmentNum == 1)
            {
                currentDBRecord.dMass = currentDBRecord.pCompound->dMass;
                vCompoundDB.push_back(currentDBRecord);
            } else
                delete currentDBRecord.pCompound;
        }

        sCurrentLine.clear();
        getline(dbfile_stream, sCurrentLine);
    }while (!dbfile_stream.eof());
}

void Database::readDatabaseFile(ifstream &dbfile_stream)
//dbfile_stream: stream of databasefile
{
    string sFirstLine;
    
    if (!dbfile_stream.eof())
    {
        sFirstLine.clear();
        getline(dbfile_stream, sFirstLine);
        boost::algorithm::trim(sFirstLine);
        if (boost::algorithm::starts_with(boost::algorithm::to_lower_copy(sFirstLine),
                                          "identifier	name	inchi	links"))
        //official standard database
        {
            readOfficialDatabaseFile(dbfile_stream, sFirstLine);
        }
        else
        //internal database
        {
            readInternalDatabaseFile(dbfile_stream, sFirstLine);
        }
    }
}

void Database::checkBondPatternsByGrammar()
{
    size_t i, iCompoundNum;
    
    iCompoundNum = vCompoundDB.size();
    
    //use openmp
    #pragma omp parallel for \
    private(i) \
    schedule(guided)
    for (i = 0; i < iCompoundNum; i++)
    {
        //cout<<"database compound "<<i<<" out of "<<iCompoundNum<<" compounds"<<endl;
        checkOneBondPatternsByGrammar(i);
    }
}

void Database::checkOneBondPatternsByGrammar(size_t iCompoundIdx)
//iCompoundIdx: index of compound
{
    size_t i;
    bool bPatternIdentified;
    compoundType *pCurCompound;
    ParseRegularGraph ParseG(true);//bPartialParse = true
    
    pCurCompound = vCompoundDB.at(iCompoundIdx).pCompound;
    pCurCompound->vbGrammarPattern.clear();
    for (i = 0; i < pSettingInfo->vpGeneralGrammars.size(); i++)
    {
        bPatternIdentified = ParseG.parseGraph(pCurCompound->rgTopology, *(pSettingInfo->vpGeneralGrammars.at(i)));
        pCurCompound->vbGrammarPattern.push_back(bPatternIdentified);
    }
    
}

void Database::loadDatabase(const string &sDatabasePath)
//sDatabasePath: the path of database file
{
    ifstream dbfile_stream;
    bool bDBFileExist;
    
    dbfile_stream.open(sDatabasePath.c_str(), ifstream::in);
    bDBFileExist = dbfile_stream.is_open();
    
    if (bDBFileExist)
    {
        readDatabaseFile(dbfile_stream);
        sort(vCompoundDB.begin(), vCompoundDB.end(), lighter);
        //cout<<vCompoundDB.size()<<endl;
        //if (pSettingInfo->bGraphGrammar) //comment for testing
        readTopologies();
        if (pSettingInfo->bGraphGrammar)
            checkBondPatternsByGrammar();
    }else
    {
        cerr<<"we are at"<<endl;
        system("pwd");
        cerr<<"Can't open file "<<sDatabasePath<<" !"<<endl;
        exit(1);
    }
    
    dbfile_stream.clear();
    dbfile_stream.close();
}

bool Database::lighter(dbRecordType record_i, dbRecordType record_j)
//record_i: record of the first compound; record_j: record of the second compound
{
    return (record_i.dMass < record_j.dMass);
}

















