//
//  Database.h
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/18/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#ifndef __midas_cpp__Database__
#define __midas_cpp__Database__

#include <stdio.h>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include <INCHI-API/inchi.h>
#include <GraphMol/Descriptors/MolDescriptors.h>
#include <GraphMol/QueryOps.h>

#include "RegularGraph.h"
#include "SettingInfo.h"
#include "ParseRegularGraph.h"

using namespace RDKit;
using namespace std;

struct compoundType
{
    double dMass = 0;
    string sIdentifier = "";
    string sName = "";
    string sInchi = "";
    string sLinks = "";
    RWMol* pRWMol = nullptr;
    MolGraph molTopology;
    RegularGraph rgTopology;
    vector<bool> vbGrammarPattern; //element is true if the corresponding general bond pattern is identified in the compound 
};

struct dbRecordType
{
    double dMass = 0; //mass of compound
    compoundType *pCompound; // point to each compound record
};

class Database
{
    const SettingInfo * pSettingInfo; //pointer to setting information
    
    void readTopologies(); // read topologies of all compound entries
    void readOneTopology(int iEntryID); // read topology of a compound entry
    static string BondType2Str(Bond::BondType bondtype); // convert bondtype to string
    
public:
    vector<dbRecordType> vCompoundDB; // storing all compounds from the the database
    
    Database(const SettingInfo * pSettingInfo);
    ~Database();
    
    //Read compound info from database to vCompoundDB; sort in ascending according to mass 
    void loadDatabase(const string & sDatabasePath);
    void readDatabaseFile(ifstream & dbfile_stream);//Read database file
    void readOfficialDatabaseFile(ifstream & dbfile_stream, string & sFirstLine);//Read official standard database file
    void readInternalDatabaseFile(ifstream & dbfile_stream, string & sFirstLine);//Read internal database file
    void clear(); // remove all records
    static bool lighter(dbRecordType record_i, dbRecordType record_j); //return true if record_i is lighter than record_j.
    //static GML* MolGraph2GML(const MolGraph & curMolGraph); // transfer topology (MolGraph) to GML object
    static GML* ROMol2GML(const ROMol *pCurrentROMol); // transfer ROMol object to GML object
    void checkBondPatternsByGrammar();//check whether each bond pattern is available in all compounds using graph grammar gules
    void checkOneBondPatternsByGrammar(size_t iCompoundIdx);//check bond patterns for each compound by graph grammar rules
};

#endif /* defined(__midas_cpp__Database__) */
