//
//  MidasCore.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/13/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include "MidasCore.h"

MidasCore::MidasCore()
{
    
}

MidasCore::~MidasCore()
{
    delete pSettingInfo;
}

MidasCore::MidasCore(map<string, string> mArguments)
{
    pSettingInfo = new SettingInfo();
    ParseConfigure configureParser;
    
    pSettingInfo->iMaxNumSpectra = MAXNUMSPECTRA;
    pSettingInfo->dProtonMass = PROTONMASS;
    pSettingInfo->iHighestRank = HIGHESTRANK;
    InitializeArguments(mArguments);
    configureParser.ReadConfigureFile(pSettingInfo);
}

string MidasCore::AddPathSeparator(const string &sPath)
//sPath: input path
{
    string sProcessedPath = sPath;
    
    if (sProcessedPath == "")
        cerr<<"Warning: receive empty file path"<<endl;
    if (sProcessedPath.at(sProcessedPath.length()-1) != this->cPathSeparator)
        sProcessedPath += cPathSeparator;
    
    return sProcessedPath;
}

void MidasCore::InitializeArguments(const map<string, string> &mArguments)
//mArguments: input arguments
{
    map<string, string >::const_iterator mssIterator;
    
    mssIterator = mArguments.find("working directory");
    if (mssIterator != mArguments.end())
        pSettingInfo->sWorkingDirectory = AddPathSeparator(mssIterator->second);
    else
        pSettingInfo->sWorkingDirectory = AddPathSeparator(".");
    mssIterator = mArguments.find("output directory");
    if (mssIterator != mArguments.end())
        pSettingInfo->sOutputDirectory = AddPathSeparator(mssIterator->second);
    else
        pSettingInfo->sOutputDirectory = pSettingInfo->sWorkingDirectory;
    mssIterator = mArguments.find("spectra file path");
    if (mssIterator != mArguments.end())
        pSettingInfo->sSpectraFilePath = mssIterator->second;
    mssIterator = mArguments.find("configure file path");
    if (mssIterator != mArguments.end())
        pSettingInfo->sConfigureFilePath = mssIterator->second;
    mssIterator = mArguments.find("running mode");
    if (mssIterator != mArguments.end())
        pSettingInfo->sMode = mssIterator->second;
    mssIterator = mArguments.find("annotation file name");
    if (mssIterator != mArguments.end())
        pSettingInfo->sAnnotationFileName = mssIterator->second;
    mssIterator = mArguments.find("results file name");
    if (mssIterator != mArguments.end())
        pSettingInfo->sAnnotationFileName = mssIterator->second;
    
    if ((pSettingInfo->sSpectraFilePath=="")||(pSettingInfo->sConfigureFilePath=="")||(pSettingInfo->sMode==""))
    {
        cerr<<"-s -c -m should be specified!"<<endl;
        exit(1);
    }
    boost::filesystem::path pathInfo(pSettingInfo->sSpectraFilePath);
    if (boost::algorithm::to_lower_copy(pSettingInfo->sMode) == "internal")
        pSettingInfo->bInternal = true;
    else
    {
        if (pSettingInfo->sResultsFileName == "")
            pSettingInfo->sResultsFileName = pathInfo.filename().string() + ".meb";
        if (pSettingInfo->sAnnotationFileName == "")
            pSettingInfo->sAnnotationFileName = pathInfo.filename().string() + ".AFT2";
    }
    
}

void MidasCore::OutputResults_internal(const vector<SpectrumUnit *> &vpSpectra)
//vpSpectra: vector of pointers to processed spectra
{
    size_t i;
    
    for (i = 0; i < vpSpectra.size(); i++)
    {
        vpSpectra.at(i)->OutputResults_internal();
    }
}

void MidasCore::OutputResults_official(const vector<SpectrumUnit *> &vpSpectra)
//vpSpectra: vector of pointers to processed spectra
{
    string sMebFileName, sAFT2FileName, sSpectrumInfo, sCompoundScoreInfo;
    ofstream mebFileStream, AFT2FileStream;
    vector<SpectrumUnit *>::const_iterator itpSpectrumUnit;
    vector<scoreInfo>::const_iterator itScoreInfo;
    vector<peakFragmentInfo>::const_iterator itPeakScoreInfo;
    int i, j, iCurrentRank = 0;
    double dOldScore = 0;

    sMebFileName = pSettingInfo->sOutputDirectory + pSettingInfo->sResultsFileName + ".meb";
    sAFT2FileName = pSettingInfo->sOutputDirectory + pSettingInfo->sResultsFileName + ".AFT2";
    mebFileStream.open(sMebFileName, ios::app);
    AFT2FileStream.open(sAFT2FileName, ios::app);
    if ((mebFileStream.is_open()) && (AFT2FileStream.is_open()))
    {
        for (itpSpectrumUnit = vpSpectra.begin(); itpSpectrumUnit != vpSpectra.end(); ++itpSpectrumUnit)
        {
            sSpectrumInfo=(*itpSpectrumUnit)->sFileName+"\t"+(*itpSpectrumUnit)->sScanNumber+"\t"
                +to_string((*itpSpectrumUnit)->dPrecursorMZ)+"\t"+(*itpSpectrumUnit)->sRetentionTime+"\t"
                +(*itpSpectrumUnit)->sScanType+"\t";
            i = 0;
            for(itScoreInfo=(*itpSpectrumUnit)->vScoreMatches.begin();itScoreInfo!=(*itpSpectrumUnit)->vScoreMatches.end(); ++itScoreInfo)
            {
                if (i==0)
                    iCurrentRank = 1;
                else
                    if (dOldScore > itScoreInfo->dScore)
                        iCurrentRank = i+1;
                if ((pSettingInfo->iHighestRank > 0) && (iCurrentRank > pSettingInfo->iHighestRank))
                    break;
                dOldScore = itScoreInfo->dScore;
                sCompoundScoreInfo=to_string(iCurrentRank)+"\t"+to_string(itScoreInfo->dParentMassError)+"\t"
                    +itScoreInfo->sExplainedPeaks+"\t"+to_string(itScoreInfo->dScore)+"\t"
                    +itScoreInfo->sCompoundIdentifier+"\t"+itScoreInfo->sCompoundName+"\t"+itScoreInfo->sInchi+"\t"
                    +itScoreInfo->sInchi+"\t"+itScoreInfo->sModifications+"\t"+itScoreInfo->sCompoundLinks;
                mebFileStream<<sSpectrumInfo<<sCompoundScoreInfo<<"\n";
                AFT2FileStream<<"M\t"<<sSpectrumInfo<<sCompoundScoreInfo<<"\n";
                j=0;
                for (itPeakScoreInfo=itScoreInfo->vPeakFragInfo.begin();itPeakScoreInfo!=itScoreInfo->vPeakFragInfo.end();
                     ++itPeakScoreInfo)
                {
                    AFT2FileStream<<(*itpSpectrumUnit)->vPeaks.at(j).dMZ<<"\t"
                        <<(*itpSpectrumUnit)->vPeaks.at(j).dIntensity<<"\t"<<(*itpSpectrumUnit)->vPeaks.at(j).dResolution
                        <<"\t"<<(*itpSpectrumUnit)->vPeaks.at(j).dBaseline<<"\t"<<(*itpSpectrumUnit)->vPeaks.at(j).dNoise
                        <<"\t"<<(*itpSpectrumUnit)->vPeaks.at(j).iPeakCharge<<"\t";
                    if (itPeakScoreInfo->sSmiles == "")
                        AFT2FileStream<<"NA\tNA\tNA\tNA\n";
                    else
                        AFT2FileStream<<itPeakScoreInfo->sSmiles<<"\t"<<to_string(itPeakScoreInfo->iProtonOffset)<<"\t"
                            <<to_string(itPeakScoreInfo->dMZError)<<"\t"<<to_string(itPeakScoreInfo->iFragmentationLevel)
                            <<"\n";
                    j++;
                }
                i++;
            }
        }
    }else
    {
        cerr<<"cannot open files "<<sMebFileName<<" or "<<sAFT2FileName<<endl;
        exit(1);
    }
    mebFileStream.close();
    AFT2FileStream.close();
}

void MidasCore::OutputResults(const vector<SpectrumUnit *> &vpSpectra)
//vpSpectra: vector of pointers to processed spectra
{
    if (pSettingInfo->bInternal)
        OutputResults_internal(vpSpectra);
    else
        OutputResults_official(vpSpectra);
}

void MidasCore::OutputHead()
{
    string sMebFileName, sAFT2FileName, sCurrentLine;
    ifstream configureFileStream;
    ofstream mebFileStream, AFT2FileStream;
    
    sMebFileName = pSettingInfo->sOutputDirectory + pSettingInfo->sResultsFileName + ".meb";
    sAFT2FileName = pSettingInfo->sOutputDirectory + pSettingInfo->sResultsFileName + ".AFT2";

    configureFileStream.open(pSettingInfo->sConfigureFilePath, ifstream::in);
    mebFileStream.open(sMebFileName, ios::out);
    if ((configureFileStream.is_open()) && (mebFileStream.is_open()))
    {
        while (!configureFileStream.eof())
        {
            sCurrentLine.clear();
            getline(configureFileStream, sCurrentLine);
            mebFileStream<<"#\t"<<sCurrentLine<<"\n";
        }
        mebFileStream<<"#\n#\tFilename: Filename of the input MS2 file\n";
        mebFileStream<<"#\tScanNumber: Scan number(s). Merged scans will have multiple numbers\n";
        mebFileStream<<"#\tPrecursorMZ: precursor m/z\n#\tRetentionTime: retention time\n";
        mebFileStream<<"#\tScanType: Scan types, e.g. positive/negative ion mode\n#\tRank: Rank of this identification\n";
        mebFileStream<<"#\tParentMassError: mass error of the parent ion\n";
        mebFileStream<<"#\tExplainedPeaks: e.g. 8>10 means, out of the 10 peaks, 8 are explained by this compound\n";
        mebFileStream<<"#\tScore: Score of this identification\n#\tIdentifier: Identifier from the database\n";
        mebFileStream<<"#\tName: Name from the database\n#\tOriginalInChI: Original InChI from the database\n";
        mebFileStream<<"#\tIdentifiedInChI: Identified InChI with potential modifications\n";
        mebFileStream<<"#\tModifications: Modifications added to the metabolite\n#\tLinks: Links from the database\n#\n";
        mebFileStream<<"Filename\tScanNumber\tPrecursorMZ\tRetentionTime\tScanType\tRank\tParentMassError\t";
        mebFileStream<<"ExplainedPeaks\tScore\tIdentifier\tName\tOriginalInChI\tIdentifiedInChI\tModifications\tLinks\n";
    }else
    {
        cerr<<"cannot open files "<<pSettingInfo->sConfigureFilePath<<" or "<<sMebFileName<<endl;
        exit(1);
    }
    configureFileStream.close();
    mebFileStream.close();
    AFT2FileStream.open(sAFT2FileName, ios::out);
    if (AFT2FileStream.is_open())
    {
        AFT2FileStream<<"Filename\tScanNumber\tPrecursorMZ\tRetentionTime\tScanType\tRank\tParentMassError\t";
        AFT2FileStream<<"ExplainedPeaks\tScore\tIdentifier\tName\tOriginalInChI\tIdentifiedInChI\tModifications\tLinks\n";
        AFT2FileStream<<"m/z\tIntensity\tResolution\tBaseline\tNoise\tCharge\tSMILES\tProtonOffset\tm/zError\t";
        AFT2FileStream<<"FragmentationLevel\n";
    }else
    {
        cerr<<"cannot open file "<<sAFT2FileName<<endl;
        exit(1);
    }
    AFT2FileStream.close();
}

void MidasCore::Process()
{
    Database compoundDB(pSettingInfo);
    ParseSpectraFile *pSpectraFileParser;
    vector<SpectrumUnit *> vpSpectra;
    size_t i, iSepectraNum;
    
    compoundDB.loadDatabase(pSettingInfo->sMetaboliteDatabasePath);
    
    pSpectraFileParser = new ParseSpectraFile(pSettingInfo);
    ClearSpectraVector(vpSpectra);
    pSpectraFileParser->loadSpectraFile(pSettingInfo->sSpectraFilePath);
    if (!(pSettingInfo->bInternal))
        OutputHead();
    while (pSpectraFileParser->nextBatchSpectra(pSettingInfo->iMaxNumSpectra, vpSpectra))
    {
        iSepectraNum = vpSpectra.size();
        //use openmp
        #pragma omp parallel for \
        shared(compoundDB) private(i) \
        schedule(guided)
        for (i=0; i<iSepectraNum; i++)
        {
            vpSpectra.at(i)->Process(compoundDB.vCompoundDB);
        }
        OutputResults(vpSpectra);
        ClearSpectraVector(vpSpectra);
    }
    vector<SpectrumUnit *>().swap(vpSpectra);
    delete pSpectraFileParser;
}

void MidasCore::ClearSpectraVector(vector<SpectrumUnit *> & vpSpectra)
//vpSpectra: vector of points to spectra
{
    vector<SpectrumUnit *>::iterator itpSpectrumUnit;
    
    for (itpSpectrumUnit = vpSpectra.begin(); itpSpectrumUnit < vpSpectra.end(); itpSpectrumUnit++)
    {
        delete *itpSpectrumUnit;
    }
    vpSpectra.clear();
}



