//
//  ParseConfigure.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/13/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include "ParseConfigure.h"

ParseConfigure::ParseConfigure()
{
    
}

ParseConfigure::~ParseConfigure()
{
    
}

void ParseConfigure::ReadConfigureFile(SettingInfo *pSetting)
//pSetting: the point to the object storing all setting information
{
    string sConfigureFilePath;
    ifstream configureFile_stream;
    bool bConfigureFileExist;
    
    pSetting->mGrammarPath.clear();
    pSetting->mGrammarWeight.clear();
    sConfigureFilePath = pSetting->sConfigureFilePath;
    configureFile_stream.open(sConfigureFilePath.c_str(), ifstream::in);
    bConfigureFileExist = configureFile_stream.is_open();
    
    if (bConfigureFileExist)
    {
        ParseConfigureInfo(pSetting, configureFile_stream);
    }else
    {
        cerr<<"we are at"<<endl;
        system("pwd");
        cerr<<"Can't open file "<<sConfigureFilePath<<" !"<<endl;
        exit(1);
    }
    configureFile_stream.clear();
    configureFile_stream.close();
    
    if (pSetting->bGraphGrammar)
        ProcessGraphGrammar(pSetting);
}

void ParseConfigure::ParseConfigureInfo(SettingInfo *pSetting, ifstream &currentConfigureFile_stream)
//pSetting: point to the object storing all setting information; currentConfigureFile_stream: stream of configure file
{
    string sCurrentLine;
    size_t poundPos, equalPos;
    string sCurrentSection="";
    
    while (!currentConfigureFile_stream.eof())
    {
        sCurrentLine.clear();
        getline(currentConfigureFile_stream, sCurrentLine);
        poundPos = sCurrentLine.find("#");
        if (poundPos != string::npos)
            sCurrentLine.erase(poundPos);
        boost::algorithm::trim(sCurrentLine);
        if (sCurrentLine.length() > 0)
        {
            if ((sCurrentLine.at(0) == '[') && (sCurrentLine.at(sCurrentLine.length()-1) == ']'))
            {
                sCurrentSection = boost::algorithm::to_lower_copy(sCurrentLine);
            }
            else
            {
                equalPos = sCurrentLine.find("=");
                if (equalPos != string::npos)
                {
                    ParseLine(pSetting, sCurrentLine.substr(0, equalPos),sCurrentLine.substr(equalPos+1),sCurrentSection);
                }else
                {
                    cerr<<"wrong format, can't find = : "<<sCurrentLine<<endl;
                }
            }
        }
    }
}

void ParseConfigure::ParseLine(SettingInfo *pSetting, string sItemName, string sItemValue, const string &sSectionName)
//pSetting: point to the object storing all setting information; sItemName: item name;
//sItemValue: string item value; sSectionName: name of the current section
{
    boost::algorithm::trim(sItemName);
    boost::algorithm::trim(sItemValue);
    boost::algorithm::to_lower(sItemName);
    if (sSectionName == "[metabolite_identification]")
    {
        if (sItemName == "metabolite_database")
            pSetting->sMetaboliteDatabasePath = sItemValue;
        if (sItemName == "default_polarity")
        {
            if (boost::algorithm::to_lower_copy(sItemValue) == "positive")
                pSetting->bPositiveDefaultPolarity = true;
            else
                pSetting->bPositiveDefaultPolarity = false;
        }
        if (sItemName == "default_charge_state")
            pSetting->iDefaultChargeState = boost::lexical_cast<int>(sItemValue);
        if (sItemName == "parent_mass_windows")
            pSetting->viParentMassWindows = NumSequenceParse<int>(sItemValue, ",");
        if (sItemName == "positive_ion_fragment_mass_windows")
            pSetting->viPositiveIonFragmentMassWindows = NumSequenceParse<int>(sItemValue, ",");
        if (sItemName == "negative_ion_fragment_mass_windows")
            pSetting->viNegativeIonFragmentMassWindows = NumSequenceParse<int>(sItemValue, ",");
        if (sItemName == "mass_tolerance_parent_ion")
            pSetting->dMassToleranceParentIon = boost::lexical_cast<double>(sItemValue);
        if (sItemName == "mass_tolerance_fragment_ions")
            pSetting->dMassToleranceFragmentIons = boost::lexical_cast<double>(sItemValue);
        if (sItemName == "break_rings")
        {
            if (boost::algorithm::to_lower_copy(sItemValue) == "true")
                pSetting->bBreakRings = true;
            else
                pSetting->bBreakRings = false;
        }
        if (sItemName == "fragmentation_depth")
        {
            pSetting->iFragmentationDepth = boost::lexical_cast<int>(sItemValue);
            if ((pSetting->iFragmentationDepth < 1) || (pSetting->iNumberofProcesses > 5))
            {
                cerr<<"Fragmentation_Depth should be between 1 and 5"<<endl;
                exit(1);
            }
        }
        if (sItemName == "number_of_processes")
            pSetting->iNumberofProcesses = boost::lexical_cast<int>(sItemValue);
        if (sItemName == "number_of_spectra_processed_together")
            pSetting->iMaxNumSpectra = boost::lexical_cast<int>(sItemValue);
        if (sItemName == "hightest_output_rank")
            pSetting->iHighestRank = boost::lexical_cast<int>(sItemValue);
        if (sItemName == "graph_grammar")
        {
            if (boost::algorithm::to_lower_copy(sItemValue) == "true")
                pSetting->bGraphGrammar = true;
            else
                pSetting->bGraphGrammar = false;
        }
        if (boost::starts_with(sItemName, "graph_grammar_"))
            ParseGraphGrammarInfo(pSetting, sItemName, sItemValue);
    }
}

void ParseConfigure::ParseGraphGrammarInfo(SettingInfo *pSetting, const string &sItemName, const string &sItemValue)
//pSetting: point to the object storing all setting information; sItemName: item name; sItemValue: string item value;
{
    string sGrammarName;
    size_t iGrammarNameLen;
    if (boost::starts_with(sItemName, "graph_grammar_path{"))
    {
        iGrammarNameLen = sItemName.length() - 20;
        sGrammarName = sItemName.substr(19, iGrammarNameLen);
        pSetting->mGrammarPath[sGrammarName] = sItemValue;
    }
    if (boost::starts_with(sItemName, "graph_grammar_weight{"))
    {
        iGrammarNameLen = sItemName.length() - 22;
        sGrammarName = sItemName.substr(21, iGrammarNameLen);
        pSetting->mGrammarWeight[sGrammarName] = boost::lexical_cast<double>(sItemValue);
    }
}

void ParseConfigure::ProcessGraphGrammar(SettingInfo *pSetting)
//pSetting: point to the object storing all setting information;
{
    map<string, string>::const_iterator mssIt;
    map<string, double>::const_iterator msdIt;
    ReadGrammar readG;
    GraphGrammar *pGeneralGrammar;
    GraphGrammar *pSpecialGrammar;
    pSetting->vpGeneralGrammars.clear();
    pSetting->vpSpecialGrammars.clear();
    pSetting->vdGrammarWeights.clear();
    
    for (mssIt = pSetting->mGrammarPath.begin(); mssIt != pSetting->mGrammarPath.end(); ++mssIt)
    {
        
        pGeneralGrammar = new GraphGrammar();
        readG.ParseGrammarFile(mssIt->second, *pGeneralGrammar);
        pSetting->vpGeneralGrammars.push_back(pGeneralGrammar);
        pSpecialGrammar = new GraphGrammar(*pGeneralGrammar);
        pSpecialGrammar->AddAnnotation("focus");
        pSetting->vpSpecialGrammars.push_back(pSpecialGrammar);
        msdIt = pSetting->mGrammarWeight.find((mssIt->first));
        if (msdIt == pSetting->mGrammarWeight.end())
        {
            cerr<<"cannot find grammar "<<mssIt->first<<endl;
            exit(1);
        }else
        {
            pSetting->vdGrammarWeights.push_back(msdIt->second);
        }
    }
}







