//
//  ParseConfigure.h
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/13/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#ifndef __midas_cpp__ParseConfigure__
#define __midas_cpp__ParseConfigure__

#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include "SettingInfo.h"


class ParseConfigure
{
public:
    ParseConfigure();
    ~ParseConfigure();
    void ReadConfigureFile(SettingInfo *pSetting); //read configure file
    //parse configure file information
    void ParseConfigureInfo(SettingInfo *pSetting, ifstream &currentConfigureFile_stream);
    //parse one line of configure file
    void ParseLine(SettingInfo *pSetting, string sItemName, string sItemValue, const string & sSectionName);
    //parse graph grammar information
    void ParseGraphGrammarInfo(SettingInfo *pSetting, const string &sItemName, const string &sItemValue);
    //read and process graph grammars
    void ProcessGraphGrammar(SettingInfo *pSetting);
    //split a string of sequence and translate each item as a nubmer
    template <typename NumberType>
    static vector<NumberType> NumSequenceParse(const string & sInputStr, const string & sSep)
    //sInputStr: the input string, e.g. 1,2,3 or 1.2, 3.6, 3.4; sSep: separator for separating values in the sequence
    //return: the vector of tansferred value, e.g. int, float, or double
    {
        vector<NumberType> vNumValues;
        vector<string> vsValue;
        size_t i;
        
        boost::split(vsValue, sInputStr, boost::is_any_of(sSep), boost::token_compress_on);
        for (i = 0; i < vsValue.size(); i++)
        {
            boost::algorithm::trim(vsValue[i]);
            if (vsValue[i] != "")
                vNumValues.push_back(boost::lexical_cast<NumberType>(vsValue[i]));
        }
        
        return vNumValues;
    }
};
#endif /* defined(__midas_cpp__ParseConfigure__) */
