//
//  ParseSpectraFile.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/20/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include "ParseSpectraFile.h"

ParseSpectraFile::ParseSpectraFile(const SettingInfo * pSettingInfo)
//pSettingInfo: point to all setting information
{
    this->pSettingInfo = pSettingInfo;
    if (pSettingInfo->bPositiveDefaultPolarity)
        this->sScanType = "PositiveIon";
    else
        this->sScanType = "NegativeIon";
    this->iDefaultPrecursorChargeState = pSettingInfo->iDefaultChargeState;
}

ParseSpectraFile::~ParseSpectraFile()
{
    
}

bool ParseSpectraFile::nextBatchSpectra(int iMaxNumSpectra, vector<SpectrumUnit *> &vpSpectra)
//iMaxNumSpectra: maximal number of spectra fed to vpSpectra; vpSpectra: vector of points to spectra
//if no spectrum is available, return false; otherwise even if just one spectrum is fed to vpSpectra, return true;
{
    bool bReVal = true;
    SpectrumUnit *pCurrentSpectrum;
    int SpectraNum = 0;
    
    while (SpectraNum < iMaxNumSpectra)
    {
        pCurrentSpectrum = new SpectrumUnit(this->sInputFileName, this->sScanType, this->iDefaultPrecursorChargeState,
                                            this->pSettingInfo);
//        cout<<pCurrentSpectrum->iPrecursorCharge<<endl;
        if (nextSpectrum(pCurrentSpectrum))
        {
            if (pCurrentSpectrum->vPeaks.size() > 0)
            {
                vpSpectra.push_back(pCurrentSpectrum);
                SpectraNum += 1;
            }else
                delete pCurrentSpectrum;            
        }else
        {
            delete pCurrentSpectrum;
            spectrafile_stream.clear();
            spectrafile_stream.close();
            break;
        }
    }
    if (SpectraNum == 0)
        bReVal = false;
    
    return bReVal;
}

bool ParseSpectraFile::nextSpectrum_internal(SpectrumUnit *pNextSpectrum)
//pNextSpectrum: point to the new sprectrum
{
    bool bReVal = true;
    string sTitleLine = "";
    string sCurrentLine = "";
    vector<string> vsValues;
    peakType currentPeak;
    
    if (!spectrafile_stream.eof())
    {
        while (!spectrafile_stream.eof())
        {
            getline(spectrafile_stream, sCurrentLine);
            boost::algorithm::trim(sCurrentLine);
            if (boost::algorithm::starts_with(sCurrentLine, "//")) //end line of a spectrum
                break;
            if (boost::algorithm::starts_with(sCurrentLine, "*")) //begin line of a spectrum
            {
                vsValues.clear();
                sTitleLine = sCurrentLine;
                boost::split(vsValues, sCurrentLine, boost::is_any_of("\t"), boost::token_compress_on);
                pNextSpectrum->sIdentifier = vsValues.at(1);
                pNextSpectrum->sScanNumber = vsValues.at(2);
                //vsValues.at(3) is actually precursor neutral mass
                if (pSettingInfo->bPositiveDefaultPolarity)
                    pNextSpectrum->dPrecursorMZ = (boost::lexical_cast<double>(vsValues.at(3)) +
                    (this->iDefaultPrecursorChargeState)*pSettingInfo->dProtonMass)/(this->iDefaultPrecursorChargeState);
                else
                    pNextSpectrum->dPrecursorMZ = (boost::lexical_cast<double>(vsValues.at(3)) -
                    (this->iDefaultPrecursorChargeState)*pSettingInfo->dProtonMass)/(this->iDefaultPrecursorChargeState);
                
                pNextSpectrum->sInchi = vsValues.at(4);
            }
            if (boost::algorithm::starts_with(sCurrentLine, "+")) // line for peak information
            {
                sCurrentLine = sCurrentLine.substr(1);
                vsValues.clear();
                currentPeak = {};
                currentPeak.dResolution = 0;
                currentPeak.dBaseline = 0;
                currentPeak.dNoise = 0;
                currentPeak.iPeakCharge = this->iDefaultIonChargeState;
                boost::split(vsValues, sCurrentLine, boost::is_any_of(" "), boost::token_compress_on);
                currentPeak.dMZ = boost::lexical_cast<double>(vsValues.at(0));
                currentPeak.dIntensity = boost::lexical_cast<double>(vsValues.at(1));
                pNextSpectrum->vPeaks.push_back(currentPeak);
            }
        }
    }else
    {
        bReVal = false;
    }
    if (sTitleLine == "")
        bReVal =false;
    
    return bReVal;
}

void ParseSpectraFile::ParseDataLine_official(SpectrumUnit *pNextSpectrum, string &sCurrentLine)
//pNextSpectrum: point to the new sprectrum; sCurrentLine: string of the current line
{
    vector<string> vsValues;
    peakType currentPeak;
    
    boost::split(vsValues, sCurrentLine, boost::is_any_of("\t"), boost::token_compress_on);
    if (vsValues.size() == 6)
    {
        currentPeak.dMZ = boost::lexical_cast<double>(vsValues.at(0));
        currentPeak.dIntensity = boost::lexical_cast<double>(vsValues.at(1));
        currentPeak.dResolution = boost::lexical_cast<double>(vsValues.at(2));
        currentPeak.dBaseline = boost::lexical_cast<double>(vsValues.at(3));
        currentPeak.dNoise = boost::lexical_cast<double>(vsValues.at(4));
        currentPeak.iPeakCharge = boost::lexical_cast<int>(vsValues.at(5));
        if (currentPeak.iPeakCharge == 0)
            currentPeak.iPeakCharge = this->iDefaultIonChargeState;
        pNextSpectrum->vPeaks.push_back(currentPeak);
    }
}

void ParseSpectraFile::ParseSettingLine_official(SpectrumUnit *pNextSpectrum, string &sCurrentLine)
//pNextSpectrum: point to the new sprectrum; sCurrentLine: string of the current line
{
    vector<string> vsValues;
    string sPrecursorCharge;
    
    boost::split(vsValues, sCurrentLine, boost::is_any_of("\t"), boost::token_compress_on);
    if ((vsValues.at(0) == "S") && (vsValues.size() == 4))
    {
        pNextSpectrum->sScanNumber  = vsValues.at(1);
        pNextSpectrum->dPrecursorMZ = boost::lexical_cast<double>(vsValues.at(3));
        
    }
    if ((vsValues.at(0) == "Z") && (vsValues.size() == 3))
    {
        sPrecursorCharge = vsValues.at(1);
        if (sPrecursorCharge == "+0")
            pNextSpectrum->iPrecursorCharge = this->iDefaultPrecursorChargeState;
        else
            if (sPrecursorCharge == "-0")
                pNextSpectrum->iPrecursorCharge = (-1) * this->iDefaultPrecursorChargeState;
            else
                if (sPrecursorCharge != "0") // =="0": iPrecursorCharge has no value change
                    pNextSpectrum->iPrecursorCharge = boost::lexical_cast<int>(sPrecursorCharge);
    }
    if ((vsValues.at(0) == "I") && (vsValues.at(1) == "RetentionTime") && (vsValues.size() >= 3))
        pNextSpectrum->sRetentionTime = vsValues.at(2);
}

bool ParseSpectraFile::nextSpectrum_official(SpectrumUnit *pNextSpectrum)
//pNextSpectrum: point to the new sprectrum
{
    bool bReVal = true;
    bool bDataBegins = false;
    string sCurrentLine = "";
    
    if (sFirstLine != "")
    {
        sCurrentLine = this->sFirstLine;
        sFirstLine = "";
        while (true)//(!spectrafile_stream.eof())
        {
            boost::algorithm::trim(sCurrentLine);
            if (sCurrentLine != "")
            {
                if (isdigit(sCurrentLine.at(0)))
                {
                    bDataBegins = true;
                    ParseDataLine_official(pNextSpectrum, sCurrentLine);
                }else
                {
                    if (bDataBegins) //end of the spectrum, this is the first line of the next spectrum
                    {
                        sFirstLine = sCurrentLine;
                        break;
                    }else
                        ParseSettingLine_official(pNextSpectrum, sCurrentLine);
                }
            }
            if (!spectrafile_stream.eof())
                getline(spectrafile_stream, sCurrentLine);
            else // end of file
                break;
        }
    }else
        bReVal = false;
    return bReVal;
}

bool ParseSpectraFile::nextSpectrum(SpectrumUnit *pNextSpectrum)
//pNextSpectrum: point to the new sprectrum
//if no spectrum available, return false, otherwise, return true
{
    bool bReVal;
    
    if (pSettingInfo->bInternal)
        bReVal = nextSpectrum_internal(pNextSpectrum);
    else
        bReVal = nextSpectrum_official(pNextSpectrum);
    
    return bReVal;
}

void ParseSpectraFile::loadSpectraFile(const string &sSpectraFilePath)
//sSpectraFilePath is the path of the spectra file
{
    bool bSpectraFileExist;
    
    boost::filesystem::path pathInfo(sSpectraFilePath);
    sInputFileName = pathInfo.filename().string();
    spectrafile_stream.open(sSpectraFilePath.c_str(), ifstream::in);
    bSpectraFileExist = spectrafile_stream.is_open();
    if (bSpectraFileExist)
    {
        if (!(pSettingInfo->bInternal)) // offical mode: the spectra file is with FT2 format
            while (sFirstLine == "")
            {
                if (!spectrafile_stream.eof())
                {
                    getline(spectrafile_stream, sFirstLine);
                    boost::algorithm::trim(sFirstLine);
                }else
                    break;
            }
    }else
    {
        cerr<<"we are at"<<endl;
        system("pwd");
        cerr<<"Can't open file "<<sSpectraFilePath<<" !"<<endl;
        exit(1);
    }
    
}


