//
//  ParseSpectraFile.h
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/20/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#ifndef __midas_cpp__ParseSpectraFile__
#define __midas_cpp__ParseSpectraFile__

#include <stdio.h>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include "SpectrumUnit.h"
#include "SettingInfo.h"

using namespace std;

class ParseSpectraFile
{
    const SettingInfo *pSettingInfo;
    //bool bInternal; //true for internal mode; false for official mode
    string sInputFileName; //file name of the spectra file
    string sScanType; //spectrum type: PositiveIon or NegativeIon
    ifstream spectrafile_stream; // stream of the file
    string sFirstLine = ""; // first line of the spectrum, just for offical mode
    int iDefaultIonChargeState = 1; //default ion (or peak) charge state, also used in SpectrumUnit
    int iDefaultPrecursorChargeState = 0; //default chage state of precursor 
    // read the next spectrum
    bool nextSpectrum(SpectrumUnit *pNextSpectrum);
    // read the next spectrum under internal mode
    bool nextSpectrum_internal(SpectrumUnit *pNextSpectrum);
    // read the next spectrum under official mode
    bool nextSpectrum_official(SpectrumUnit *pNextSpectrum);
    //parse line of numbers in FT2 file
    void ParseDataLine_official(SpectrumUnit *pNextSpectrum, string & sCurrentLine);
    //parse line of settings in FT2 file
    void ParseSettingLine_official(SpectrumUnit *pNextSpectrum, string & sCurrentLine);

public:
    ParseSpectraFile(const SettingInfo * pSettingInfo);
    ~ParseSpectraFile();
    
    void loadSpectraFile(const string &sSpectraFilePath);// open the spectra file
    bool nextBatchSpectra(int iMaxNumSpectra, vector<SpectrumUnit *> &vpSpectra);// read spectra to vpSpectra
};

#endif /* defined(__midas_cpp__ParseSpectraFile__) */
