//
//  SettingInfo.h
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/13/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#ifndef __midas_cpp__SettingInfo__
#define __midas_cpp__SettingInfo__

#include <stdio.h>
#include <string>
#include <vector>

#include "ReadGrammar.h"

using namespace std;

class SettingInfo
{
public:
    int iMaxNumSpectra; //maximal spectra read from the spectra file
    double dProtonMass; //Mass of Proton;
    int iHighestRank; //highest rank printed in the results; < 0 means no limitation
    
    string sWorkingDirectory; // path of the working directory
    string sConfigureFilePath; // path of the configure file
    string sOutputDirectory; // path of the output directory
    string sMode; //running mode
    string sSpectraFilePath; // path of the spectra file
    string sResultsFileName; // name of the results file (just for official mode)
    string sAnnotationFileName; // name of the annotation file (just for official mode)
    bool bInternal; //true: internal mode; false: official mode

    string sMetaboliteDatabasePath;//path of metabolite database file
    bool bPositiveDefaultPolarity; //Presumed polarity if not provided in the input file.
    int iDefaultChargeState; //Presumed charge state for unknown charge state
    vector<int> viParentMassWindows; //Mass Windows to be open around parent ion.
    vector<int> viPositiveIonFragmentMassWindows; //positive ion fragment mass windows
    vector<int> viNegativeIonFragmentMassWindows; //negative ion fragment mass windows
    double dMassToleranceParentIon; //Parent mass tolerance in Da
    double dMassToleranceFragmentIons; //Fragment mass tolerance in Da
    bool bBreakRings; //Break ring structures in a metabolite or not; true: break rings
    int iFragmentationDepth; // Depth of fragmentation pathways. The valid range is from 1 to 5
    int iNumberofProcesses; //Number of processes the search should initiate (not useful in C++)
    
    bool bGraphGrammar; // true: use graph grammar;
    map<string, string> mGrammarPath; //map of grammar path
    map<string, double> mGrammarWeight;//map of grammar weight
    vector<GraphGrammar *> vpGeneralGrammars; //vector of pointers referring to genernal grammars
    vector<GraphGrammar *> vpSpecialGrammars; //vector of pointers referring to grammars with specific bonds highlighted
    vector<double> vdGrammarWeights; //vector of grammar weights
    
    SettingInfo();
    ~SettingInfo();
};
#endif /* defined(__midas_cpp__SettingInfo__) */
