//
//  main.cpp
//  midas_cpp
//
//  Created by Yingfeng Wang on 3/8/16.
//  Copyright (c) 2016 Yingfeng Wang. All rights reserved.
//

#include <iostream>
 
#include <INCHI-API/inchi.h>
#include <GraphMol/RDKitBase.h>
#include <RDGeneral/types.h>
#include <RDGeneral/RDLog.h>
#include <GraphMol/SmilesParse/SmilesParse.h>
#include <GraphMol/MolOps.h>

#include "MidasCore.h"

using namespace RDKit;
using namespace std;

void ParseArguments(int argc, const char * argv[], map<string, string> & mArguments)
{
    vector<string> vsArguments;
    size_t i;
    
    while(argc--)
        vsArguments.push_back(*argv++);
    for (i = 0; i < vsArguments.size(); i++)
    {
        if ((vsArguments[i] == "-s") || (vsArguments[i] == "--spectra"))
            mArguments.insert(pair<string, string>("spectra file path", vsArguments[++i]));
        else if ((vsArguments[i] == "-w") || (vsArguments[i] == "--working_directory"))
            mArguments.insert(pair<string, string>("working directory", vsArguments[++i]));
        else if ((vsArguments[i] == "-c") || (vsArguments[i] == "--configure"))
            mArguments.insert(pair<string, string>("configure file path", vsArguments[++i]));
        else if ((vsArguments[i] == "-o") || (vsArguments[i] == "--output_directory"))
            mArguments.insert(pair<string, string>("output directory", vsArguments[++i]));
        else if ((vsArguments[i] == "-m") || (vsArguments[i] == "--mode"))
            mArguments.insert(pair<string, string>("running mode", vsArguments[++i]));
        else if ((vsArguments[i] == "-a") || (vsArguments[i] == "--annotation_file_name"))
            mArguments.insert(pair<string, string>("annotation file name", vsArguments[++i]));
        else if ((vsArguments[i] == "-r") || (vsArguments[i] == "--results_file_name"))
            mArguments.insert(pair<string, string>("results file name", vsArguments[++i]));
        else if ((vsArguments[i] == "-h") || (vsArguments[i] == "--help"))
        {
            cout<<"Usage: -s spectra_file_path -w working_directory -c configure_file_path"<<endl;
            cout<<"-o output_directory -m mode (internal or official) -a annotation_file_name(optional)"<<endl;
            cout<<"-r results_file_name"<<endl;
            cout<<"-r and -a are just for official mode."<<endl;
            cout<<"If -r is not specified, the default full output file name is spectra_file_name.meb"<<endl;
            cout<<"If -a is not specified, the default full annotation file name is spectra_file_name.AFT2"<<endl;
            cout<<"If working director is not specified, the current directory will be used as working directory."<<endl;
            cout<<"If output directory is not specified, the working directory is the output directory."<<endl;
            cout<<"the current version doesn't search subdirectory"<<endl;
            exit(0);
        }
    }
}


int main(int argc, const char * argv[])
{
    map<string, string> mArgumentsInfo;
    ParseArguments(argc, argv, mArgumentsInfo);
    
    MidasCore mc(mArgumentsInfo);
    cout<<"Processing..."<<endl;
    
    mc.Process();
    
    cout<<"Done!"<<endl;
/*
    //testing code begin
    RWMol* pCurMol;
    ExtraInchiReturnValues tmp;
    string curInchi = "InChI=1S/C10H9N3O/c1-7-11-10(14)9(13-12-7)8-5-3-2-4-6-8/h2-6H,1H3,(H,11,12,14)";
    string sSmiles = "Oc1ncnnc1-c1ccccc1";//"c.ccccc";
    MolGraph curTopology;
    vector<vector<int> > viiRingBonds;
    list<boost::list_edge<unsigned long, boost::shared_ptr<Bond> > >::const_iterator itEdge;
    vector<boost::shared_ptr<ROMol> > vROMol;
    
//    pCurMol = InchiToMol(curInchi, tmp);
    pCurMol = SmilesToMol(sSmiles, 0, false);
    MolOps::findSSSR(*(pCurMol));
    vROMol = MolOps::getMolFrags(*(pCurMol), false);
    curTopology = pCurMol->getTopology();
    viiRingBonds = pCurMol->getRingInfo()->bondRings();
    //cout<<viiRingBonds.size()<<endl;
    //for (itEdge = curTopology.m_edges.begin(); itEdge != curTopology.m_edges.end(); ++itEdge)
    //    if (queryIsBondInRing(itEdge->m_property.get()))
    //        cout<<"get one" <<endl;
//    cout << InchiToInchiKey(curInchi) <<endl;
//    boostexample(3);
//    cout << "Hello, World!\n";
    //testing code end
*/
    return 0;
}
